/*************************************************************************\
 * Copyright (c) 2002 The University of Chicago, as Operator of Argonne
 * National Laboratory.
 * Copyright (c) 2002 The Regents of the University of California, as
 * Operator of Los Alamos National Laboratory.
 * This file is distributed subject to a Software License Agreement found
 * in the file LICENSE that is included with this distribution. 
\*************************************************************************/

/* program: sddsuwrap
 * unwrap data through 
 */
#include "mdb.h"
#include "SDDS.h"
#include "SDDSutils.h"
#include "scan.h"
#include <ctype.h>

static char *USAGE = "sddsunwrap [<input>] [<output>]\n\
[-pipe=[input][,output]]  -col=<time-col>,<data-column> [-inputVoltage=<value|@<parameter>]] [-z0=<value>]  \n \
-column           provide data column name.\n\
-inputVoltage     the input voltage in volts of TDR (time domain reflectometry.\n\n\
-z0               line impedance (default 50 oms).\n\
sddstdrpeeling    process the TRD data with recursive algorithm to find out the impedance of nonuniform transmission line\n\
Program by Hairong Shang.  (" __DATE__ " " __TIME__ ", SVN revision: " SVN_VERSION ")";

#define CLO_PIPE 0
#define CLO_COLUMN 1
#define CLO_INPUT_VOLTAGE 2
#define CLO_Z0 3
#define CLO_MAJOR_ORDER 4
#define CLO_OPTIONS 5
static char *option[CLO_OPTIONS] = {
  "pipe", "column", "inputVoltage", "z0", "majorOrder"};

int main(int argc, char **argv) {
  double *measData, inputVoltage, Z0;
  char *input = NULL, *output = NULL, *dataColumn, *inputVolPar;
  long iArg, k;
  double *left, *right, *gamma, *Zline, Gproduct, vr_temp, Dincrese;

  SDDS_DATASET SDDSin, SDDSout;
  SCANNED_ARG *scanned;
  unsigned long pipeFlags = 0, majorOrderFlag = 0;
  int64_t i, rows;
  short columnMajorOrder = -1;

  SDDS_RegisterProgramName(argv[0]);

  argc = scanargs(&scanned, argc, argv);
  if (argc == 1)
    bomb(NULL, USAGE);

  input = output = dataColumn = inputVolPar = NULL;
  measData = NULL;
  Z0 = 50;            /*default impedance 50 oms */
  inputVoltage = 0.2; /*default input  voltage */
  left = right = gamma = Zline = NULL;

  for (iArg = 1; iArg < argc; iArg++) {
    if (scanned[iArg].arg_type == OPTION) {
      /* process options here */
      switch (match_string(scanned[iArg].list[0], option, CLO_OPTIONS, 0)) {
      case CLO_MAJOR_ORDER:
        majorOrderFlag = 0;
        scanned[iArg].n_items--;
        if (scanned[iArg].n_items > 0 &&
            (!scanItemList(&majorOrderFlag, scanned[iArg].list + 1, &scanned[iArg].n_items, 0,
                           "row", -1, NULL, 0, SDDS_ROW_MAJOR_ORDER,
                           "column", -1, NULL, 0, SDDS_COLUMN_MAJOR_ORDER, NULL)))
          SDDS_Bomb("invalid -majorOrder syntax/values");
        if (majorOrderFlag & SDDS_COLUMN_MAJOR_ORDER)
          columnMajorOrder = 1;
        else if (majorOrderFlag & SDDS_ROW_MAJOR_ORDER)
          columnMajorOrder = 0;
        break;
      case CLO_INPUT_VOLTAGE:
        if (scanned[iArg].n_items != 2)
          SDDS_Bomb("invalid -inputVoltage syntax");
        if (scanned[iArg].list[1][0] == '@') {
          cp_str(&inputVolPar, scanned[iArg].list[1] + 1);
        } else {
          if (!get_double(&inputVoltage, scanned[iArg].list[1]))
            SDDS_Bomb("invalid -threshold value given");
        }
        break;
      case CLO_COLUMN:
        if (scanned[iArg].n_items != 2)
          SDDS_Bomb("invalid -column syntax");
        dataColumn = scanned[iArg].list[1];
        break;
      case CLO_Z0:
        if (scanned[iArg].n_items != 2)
          SDDS_Bomb("invalid -z0 syntax");
        if (!get_double(&Z0, scanned[iArg].list[1])) {
          SDDS_Bomb("invalid -z0 value given");
        }
        break;
      default:
        fprintf(stderr, "Unknown option %s provided\n", scanned[iArg].list[0]);
        exit(1);
        break;
      }
    } else {
      if (!input)
        input = scanned[iArg].list[0];
      else if (!output)
        output = scanned[iArg].list[0];
      else
        SDDS_Bomb("too many filenames");
    }
  }
  processFilenames("sddstdrpeeling", &input, &output, pipeFlags, 0, NULL);

  if (!SDDS_InitializeInput(&SDDSin, input))
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);

  if (!SDDS_InitializeCopy(&SDDSout, &SDDSin, output, "w"))
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);

  if (columnMajorOrder != -1)
    SDDSout.layout.data_mode.column_major = columnMajorOrder;
  else
    SDDSout.layout.data_mode.column_major = SDDSin.layout.data_mode.column_major;

  if (!SDDS_DefineSimpleColumn(&SDDSout, "PeeledImpedance", "oms", SDDS_DOUBLE))
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
  if (!SDDS_WriteLayout(&SDDSout))
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);

  while (SDDS_ReadPage(&SDDSin) > 0) {
    if ((rows = SDDS_CountRowsOfInterest(&SDDSin)) <= 0)
      continue;
    if (!SDDS_StartPage(&SDDSout, rows) || !SDDS_CopyPage(&SDDSout, &SDDSin))
      SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
    if (inputVolPar) {
      if (!SDDS_GetParameterAsDouble(&SDDSin, inputVolPar, &inputVoltage))
        SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
    }
    if (!(measData = SDDS_GetColumnInDoubles(&SDDSin, dataColumn)))
      SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
    for (i = 0; i < rows; i++)
      measData[i] /= inputVoltage;

    left = calloc(sizeof(*left), rows + 1);
    right = calloc(sizeof(*right), rows + 1);
    gamma = calloc(sizeof(*gamma), rows + 1);
    Zline = calloc(sizeof(*Zline), rows + 1);
    Gproduct = 1;
    vr_temp = 0;
    if (rows >= 1) {
      gamma[1] = measData[0];
      /*left[1] = measData[1] - inputVoltage; */
      left[1] = 1;
      Gproduct = Gproduct * (1 - gamma[1] * gamma[1]);
    }
    if (rows >= 2) {
      vr_temp = (1 - gamma[1]) * right[1] + gamma[1] * left[1];
      gamma[2] = (measData[1] - vr_temp) / Gproduct;
      left[2] = left[1] * (1 + gamma[1]);
      right[1] = left[2] * gamma[2];
      Gproduct = Gproduct * (1 - gamma[2] * gamma[2]);
    }
    for (i = 3; i <= rows; i++) {
      left[i] = left[i - 1] * (1 + gamma[i - 1]);
      left[i - 1] = left[i - 2] * (1 + gamma[i - 2]) - right[i - 2] * gamma[i - 2];
      for (k = i - 2; k > 2; k--) {
        right[k] = gamma[k + 1] * left[k + 1] + (1 - gamma[k + 1] * right[k + 1]);
        left[k] = (1 + gamma[k - 1]) * left[k - 1] - gamma[k - 1] * right[k - 1];
      }
      right[1] = left[2] * gamma[2] + right[2] * (1 - gamma[2]);
      vr_temp = (1 - gamma[1]) * right[1] + gamma[1] * left[1];
      gamma[i] = (measData[i - 1] - vr_temp) / Gproduct;
      Gproduct = Gproduct * (1 - gamma[i] * gamma[i]);
      Dincrese = left[i] * gamma[i];
      right[i - 1] = right[i - 1] + Dincrese;
      for (k = i - 2; k > 1; k--) {
        Dincrese = Dincrese * (1 - gamma[k + 1]);
        right[k] = right[k] + Dincrese;
      }
    }
    Zline[0] = Z0;
    for (i = 1; i <= rows; i++) {
      Zline[i] = (1 + gamma[i]) / (1 - gamma[i]) * Zline[i - 1];
    }
    if (!SDDS_SetColumnFromDoubles(&SDDSout, SDDS_BY_NAME, Zline + 1, rows, "PeeledImpedance") ||
        !SDDS_WritePage(&SDDSout))
      SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors | SDDS_EXIT_PrintErrors);
    free(left);
    free(right);
    free(gamma);
    free(Zline);
    free(measData);
    measData = NULL;
  }
  if (!SDDS_Terminate(&SDDSin) || !SDDS_Terminate(&SDDSout)) {
    SDDS_PrintErrors(stderr, SDDS_VERBOSE_PrintErrors);
    exit(1);
  }

  free_scanargs(&scanned, argc);
  return (0);
}
