/**
 * @file SDDS_ascii.c
 * @brief SDDS ascii data input and output routines
 *
 * This file contains the implementation of the SDDS ascii data input and output routines.
 * It provides functions for writing void pointer data to an ASCII file stream.
 * The functions support different data types such as short, unsigned short, long, unsigned long,
 * long long, unsigned long long, float, double, long double, string, and character.
 *
 *
 * @copyright Copyright (c) 2002 The University of Chicago, as Operator of Argonne
 *            National Laboratory. Copyright (c) 2002 The Regents of the University
 *            of California, as Operator of Los Alamos National Laboratory.
 * @license This file is distributed subject to a Software License Agreement found
 *          in the file LICENSE that is included with this distribution.
 *
 * @author M. Borland, C. Saunders, R. Soliday, H. Shang
 */

#include "SDDS.h"
#include "SDDS_internal.h"
#include "mdb.h"
#include <ctype.h>

#undef DEBUG

#define COMMENT_POSITION 40

#if SDDS_VERSION != 5
#  error "SDDS_VERSION does not match the version number of this file"
#endif

/* Define a dynamically expanded text buffer for use by ReadAsciiParameters,
   ReadAsciiArrays, and ReadAsciiPage.  Since these routines never call each
   other, it is okay that they share this buffer.
*/
#define INITIAL_BIG_BUFFER_SIZE SDDS_MAXLINE

/**
 * Write void pointer data to ASCII file stream
 *
 * @param data Void pointer containing value to be printed as ASCII value
 * @param index Array index of data to be printed, use 0 if not an array
 * @param type The SDDS data type contained in the data variable
 * @param format The optional printf format to use, use NULL to use default format
 * @param fp The FILE pointer to write the data to
 *
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_WriteTypedValue(void *data, int64_t index, int32_t type, char *format, FILE *fp) {
  char c, *s;
  short hasWhitespace;

  if (!data) {
    SDDS_SetError("Unable to write value--data pointer is NULL (SDDS_WriteTypedValue)");
    return (0);
  }
  if (!fp) {
    SDDS_SetError("Unable to print value--file pointer is NULL (SDDS_WriteTypedValue)");
    return (0);
  }
  switch (type) {
  case SDDS_SHORT:
    fprintf(fp, format ? format : "%hd", *((short *)data + index));
    break;
  case SDDS_USHORT:
    fprintf(fp, format ? format : "%hu", *((unsigned short *)data + index));
    break;
  case SDDS_LONG:
    fprintf(fp, format ? format : "%" PRId32, *((int32_t *)data + index));
    break;
  case SDDS_ULONG:
    fprintf(fp, format ? format : "%" PRIu32, *((uint32_t *)data + index));
    break;
  case SDDS_LONG64:
    fprintf(fp, format ? format : "%" PRId64, *((int64_t *)data + index));
    break;
  case SDDS_ULONG64:
    fprintf(fp, format ? format : "%" PRIu64, *((uint64_t *)data + index));
    break;
  case SDDS_FLOAT:
    fprintf(fp, format ? format : "%15.8e", *((float *)data + index));
    break;
  case SDDS_DOUBLE:
    fprintf(fp, format ? format : "%22.15e", *((double *)data + index));
    break;
  case SDDS_LONGDOUBLE:
    if (LDBL_DIG == 18) {
      fprintf(fp, format ? format : "%22.18Le", *((long double *)data + index));
    } else {
      fprintf(fp, format ? format : "%22.15Le", *((long double *)data + index));
    }
    break;
  case SDDS_STRING:
    /* ignores format string */
    s = *((char **)data + index);
    hasWhitespace = 0;
    if (SDDS_HasWhitespace(s) || SDDS_StringIsBlank(s)) {
      fputc('"', fp);
      hasWhitespace = 1;
    }
    while (s && *s) {
      c = *s++;
      if (c == '!')
        fputs("\\!", fp);
      else if (c == '\\')
        fputs("\\\\", fp);
      else if (c == '"')
        fputs("\\\"", fp);
      else if (c == ' ')
        fputc(' ', fp); /* don't escape plain spaces */
      else if (isspace(c) || !isprint(c))
        fprintf(fp, "\\%03o", c);
      else
        fputc(c, fp);
    }
    if (hasWhitespace)
      fputc('"', fp);
    break;
  case SDDS_CHARACTER:
    /* ignores format string */
    c = *((char *)data + index);
    if (c == '!')
      fputs("\\!", fp);
    else if (c == '\\')
      fputs("\\\\", fp);
    else if (c == '"')
      fputs("\\\"", fp);
    else if (!c || isspace(c) || !isprint(c))
      fprintf(fp, "\\%03o", c);
    else
      fputc(c, fp);
    break;
  default:
    SDDS_SetError("Unable to write value--unknown data type (SDDS_WriteTypedValue)");
    return (0);
  }
  return (1);
}

/**
 * Write void pointer data to ASCII file stream with LZMA compression
 *
 * @param data Void pointer containing value to be printed as ASCII value
 * @param index Array index of data to be printed, use 0 if not an array
 * @param type The SDDS data type contained in the data variable
 * @param format The optional printf format to use, use NULL to use default format
 * @param lzmafp The LZMA file pointer to write the data to
 *
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_LZMAWriteTypedValue(void *data, int64_t index, int32_t type, char *format, struct lzmafile *lzmafp) {
  char c, *s;
  short hasWhitespace;

  if (!data) {
    SDDS_SetError("Unable to write value--data pointer is NULL (SDDS_LZMAWriteTypedValue)");
    return (0);
  }
  if (!lzmafp) {
    SDDS_SetError("Unable to print value--file pointer is NULL (SDDS_LZMAWriteTypedValue)");
    return (0);
  }
  switch (type) {
  case SDDS_SHORT:
    lzma_printf(lzmafp, format ? format : "%hd", *((short *)data + index));
    break;
  case SDDS_USHORT:
    lzma_printf(lzmafp, format ? format : "%hu", *((unsigned short *)data + index));
    break;
  case SDDS_LONG:
    lzma_printf(lzmafp, format ? format : "%" PRId32, *((int32_t *)data + index));
    break;
  case SDDS_ULONG:
    lzma_printf(lzmafp, format ? format : "%" PRIu32, *((uint32_t *)data + index));
    break;
  case SDDS_LONG64:
    lzma_printf(lzmafp, format ? format : "%" PRId64, *((int64_t *)data + index));
    break;
  case SDDS_ULONG64:
    lzma_printf(lzmafp, format ? format : "%" PRIu64, *((uint64_t *)data + index));
    break;
  case SDDS_FLOAT:
    lzma_printf(lzmafp, format ? format : "%15.8e", *((float *)data + index));
    break;
  case SDDS_DOUBLE:
    lzma_printf(lzmafp, format ? format : "%22.15e", *((double *)data + index));
    break;
  case SDDS_LONGDOUBLE:
    if (LDBL_DIG == 18) {
      lzma_printf(lzmafp, format ? format : "%22.18Le", *((long double *)data + index));
    } else {
      lzma_printf(lzmafp, format ? format : "%22.15Le", *((long double *)data + index));
    }
    break;
  case SDDS_STRING:
    /* ignores format string */
    s = *((char **)data + index);
    hasWhitespace = 0;
    if (SDDS_HasWhitespace(s) || SDDS_StringIsBlank(s)) {
      lzma_putc('"', lzmafp);
      hasWhitespace = 1;
    }
    while (s && *s) {
      c = *s++;
      if (c == '!')
        lzma_puts("\\!", lzmafp);
      else if (c == '\\')
        lzma_puts("\\\\", lzmafp);
      else if (c == '"')
        lzma_puts("\\\"", lzmafp);
      else if (c == ' ')
        lzma_putc(' ', lzmafp); /* don't escape plain spaces */
      else if (isspace(c) || !isprint(c))
        lzma_printf(lzmafp, "\\%03o", c);
      else
        lzma_putc(c, lzmafp);
    }
    if (hasWhitespace)
      lzma_putc('"', lzmafp);
    break;
  case SDDS_CHARACTER:
    /* ignores format string */
    c = *((char *)data + index);
    if (c == '!')
      lzma_puts("\\!", lzmafp);
    else if (c == '\\')
      lzma_puts("\\\\", lzmafp);
    else if (c == '"')
      lzma_puts("\\\"", lzmafp);
    else if (!c || isspace(c) || !isprint(c))
      lzma_printf(lzmafp, "\\%03o", c);
    else
      lzma_putc(c, lzmafp);
    break;
  default:
    SDDS_SetError("Unable to write value--unknown data type (SDDS_LZMAWriteTypedValue)");
    return (0);
  }
  return (1);
}

#if defined(zLib)
/**
 * Write void pointer data to ASCII file stream with GZIP compression
 *
 * @param data Void pointer containing value to be printed as ASCII value
 * @param index Array index of data to be printed, use 0 if not an array
 * @param type The SDDS data type contained in the data variable
 * @param format The optional printf format to use, use NULL to use default format
 * @param gzfp The GZIP file pointer to write the data to
 *
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_GZipWriteTypedValue(void *data, int64_t index, int32_t type, char *format, gzFile gzfp) {
  char c, *s;
  short hasWhitespace;

  if (!data) {
    SDDS_SetError("Unable to write value--data pointer is NULL (SDDS_GZipWriteTypedValue)");
    return (0);
  }
  if (!gzfp) {
    SDDS_SetError("Unable to print value--file pointer is NULL (SDDS_GZipWriteTypedValue)");
    return (0);
  }
  switch (type) {
  case SDDS_SHORT:
    gzprintf(gzfp, format ? format : "%hd", *((short *)data + index));
    break;
  case SDDS_USHORT:
    gzprintf(gzfp, format ? format : "%hu", *((unsigned short *)data + index));
    break;
  case SDDS_LONG:
    gzprintf(gzfp, format ? format : "%" PRId32, *((int32_t *)data + index));
    break;
  case SDDS_ULONG:
    gzprintf(gzfp, format ? format : "%" PRIu32, *((uint32_t *)data + index));
    break;
  case SDDS_LONG64:
    gzprintf(gzfp, format ? format : "%" PRId64, *((int64_t *)data + index));
    break;
  case SDDS_ULONG64:
    gzprintf(gzfp, format ? format : "%" PRIu64, *((uint64_t *)data + index));
    break;
  case SDDS_FLOAT:
    gzprintf(gzfp, format ? format : "%15.8e", *((float *)data + index));
    break;
  case SDDS_DOUBLE:
    gzprintf(gzfp, format ? format : "%22.15e", *((double *)data + index));
    break;
  case SDDS_LONGDOUBLE:
    if (LDBL_DIG == 18) {
      gzprintf(gzfp, format ? format : "%22.18Le", *((long double *)data + index));
    } else {
      gzprintf(gzfp, format ? format : "%22.15Le", *((long double *)data + index));
    }
    break;
  case SDDS_STRING:
    /* ignores format string */
    s = *((char **)data + index);
    hasWhitespace = 0;
    if (SDDS_HasWhitespace(s) || SDDS_StringIsBlank(s)) {
      gzputc(gzfp, '"');
      hasWhitespace = 1;
    }
    while (s && *s) {
      c = *s++;
      if (c == '!')
        gzputs(gzfp, "\\!");
      else if (c == '\\')
        gzputs(gzfp, "\\\\");
      else if (c == '"')
        gzputs(gzfp, "\\\"");
      else if (c == ' ')
        gzputc(gzfp, ' '); /* don't escape plain spaces */
      else if (isspace(c) || !isprint(c))
        gzprintf(gzfp, "\\%03o", c);
      else
        gzputc(gzfp, c);
    }
    if (hasWhitespace)
      gzputc(gzfp, '"');
    break;
  case SDDS_CHARACTER:
    /* ignores format string */
    c = *((char *)data + index);
    if (c == '!')
      gzputs(gzfp, "\\!");
    else if (c == '\\')
      gzputs(gzfp, "\\\\");
    else if (c == '"')
      gzputs(gzfp, "\\\"");
    else if (!c || isspace(c) || !isprint(c))
      gzprintf(gzfp, "\\%03o", c);
    else
      gzputc(gzfp, c);
    break;
  default:
    SDDS_SetError("Unable to write value--unknown data type (SDDS_GZipWriteTypedValue)");
    return (0);
  }
  return (1);
}
#endif

/**
 * Writes a page of data in ASCII format to the SDDS dataset.
 *
 * @param SDDS_dataset The SDDS dataset to write the page to.
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_WriteAsciiPage(SDDS_DATASET *SDDS_dataset) {
#if defined(zLib)
  gzFile gzfp;
#endif
  FILE *fp;
  struct lzmafile *lzmafp;
  int64_t i, rows;

  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_WriteAsciiPage"))
    return (0);
#if defined(zLib)
  if (SDDS_dataset->layout.gzipFile) {
    if (!(gzfp = SDDS_dataset->layout.gzfp)) {
      SDDS_SetError("Unable to write page--file pointer is NULL (SDDS_WriteAsciiPage)");
      return (0);
    }
    if (SDDS_dataset->layout.data_mode.no_row_counts && (SDDS_dataset->page_number > 1 || SDDS_dataset->file_had_data))
      gzputc(gzfp, '\n');
    gzprintf(gzfp, "! page number %" PRId32 "\n", SDDS_dataset->page_number);

    if (!SDDS_GZipWriteAsciiParameters(SDDS_dataset, gzfp) || !SDDS_GZipWriteAsciiArrays(SDDS_dataset, gzfp))
      return 0;
    rows = 0;
    if (SDDS_dataset->layout.n_columns) {
      rows = SDDS_CountRowsOfInterest(SDDS_dataset);
      if (!SDDS_dataset->layout.data_mode.no_row_counts) {
        SDDS_dataset->rowcount_offset = gztell(gzfp);
        if (SDDS_dataset->layout.data_mode.fixed_row_count) {
          gzprintf(gzfp, "%20" PRId64 "\n", ((rows / SDDS_dataset->layout.data_mode.fixed_row_increment) + 2) * SDDS_dataset->layout.data_mode.fixed_row_increment);
        } else
          gzprintf(gzfp, "%20" PRId64 "\n", rows);
      }
      for (i = 0; i < SDDS_dataset->n_rows; i++)
        if (SDDS_dataset->row_flag[i] && !SDDS_GZipWriteAsciiRow(SDDS_dataset, i, gzfp))
          return 0;
    }
    SDDS_dataset->last_row_written = SDDS_dataset->n_rows - 1;
    SDDS_dataset->n_rows_written = rows;
    SDDS_dataset->writing_page = 1;
    /*gzflush(gzfp, Z_FULL_FLUSH); */
  } else {
#endif
    if (SDDS_dataset->layout.lzmaFile) {
      if (!(lzmafp = SDDS_dataset->layout.lzmafp)) {
        SDDS_SetError("Unable to write page--file pointer is NULL (SDDS_WriteAsciiPage)");
        return (0);
      }
      if (SDDS_dataset->layout.data_mode.no_row_counts && (SDDS_dataset->page_number > 1 || SDDS_dataset->file_had_data))
        lzma_putc('\n', lzmafp);
      lzma_printf(lzmafp, "! page number %" PRId32 "\n", SDDS_dataset->page_number);

      if (!SDDS_LZMAWriteAsciiParameters(SDDS_dataset, lzmafp) || !SDDS_LZMAWriteAsciiArrays(SDDS_dataset, lzmafp))
        return 0;
      rows = 0;
      if (SDDS_dataset->layout.n_columns) {
        rows = SDDS_CountRowsOfInterest(SDDS_dataset);
        if (!SDDS_dataset->layout.data_mode.no_row_counts) {
          SDDS_dataset->rowcount_offset = lzma_tell(lzmafp);
          if (SDDS_dataset->layout.data_mode.fixed_row_count)
            lzma_printf(lzmafp, "%20" PRId64 "\n", ((rows / SDDS_dataset->layout.data_mode.fixed_row_increment) + 2) * SDDS_dataset->layout.data_mode.fixed_row_increment);
          else
            lzma_printf(lzmafp, "%20" PRId64 "\n", rows);
        }
        for (i = 0; i < SDDS_dataset->n_rows; i++)
          if (SDDS_dataset->row_flag[i] && !SDDS_LZMAWriteAsciiRow(SDDS_dataset, i, lzmafp))
            return 0;
      }
      SDDS_dataset->last_row_written = SDDS_dataset->n_rows - 1;
      SDDS_dataset->n_rows_written = rows;
      SDDS_dataset->writing_page = 1;
    } else {
      if (!(fp = SDDS_dataset->layout.fp)) {
        SDDS_SetError("Unable to write page--file pointer is NULL (SDDS_WriteAsciiPage)");
        return (0);
      }
      if (SDDS_dataset->layout.data_mode.no_row_counts && (SDDS_dataset->page_number > 1 || SDDS_dataset->file_had_data))
        fputc('\n', fp);
      fprintf(fp, "! page number %" PRId32 "\n", SDDS_dataset->page_number);

      if (!SDDS_WriteAsciiParameters(SDDS_dataset, fp) || !SDDS_WriteAsciiArrays(SDDS_dataset, fp))
        return 0;
      rows = 0;
      if (SDDS_dataset->layout.n_columns) {
        rows = SDDS_CountRowsOfInterest(SDDS_dataset);
        if (!SDDS_dataset->layout.data_mode.no_row_counts) {
          SDDS_dataset->rowcount_offset = ftell(fp);
          if (SDDS_dataset->layout.data_mode.fixed_row_count)
            fprintf(fp, "%20" PRId64 "\n", ((rows / SDDS_dataset->layout.data_mode.fixed_row_increment) + 2) * SDDS_dataset->layout.data_mode.fixed_row_increment);
          else
            fprintf(fp, "%20" PRId64 "\n", rows);
        }
        for (i = 0; i < SDDS_dataset->n_rows; i++)
          if (SDDS_dataset->row_flag[i] && !SDDS_WriteAsciiRow(SDDS_dataset, i, fp))
            return 0;
      }
      SDDS_dataset->last_row_written = SDDS_dataset->n_rows - 1;
      SDDS_dataset->n_rows_written = rows;
      SDDS_dataset->writing_page = 1;
      fflush(fp);
    }
#if defined(zLib)
  }
#endif

  return (1);
}

/**
 * Writes the parameter data of an SDDS dataset in ASCII format to a file.
 *
 * @param SDDS_dataset The SDDS dataset containing the parameter data.
 * @param fp The file pointer to write the ASCII data to.
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_WriteAsciiParameters(SDDS_DATASET *SDDS_dataset, FILE *fp) {
  int32_t i;
  SDDS_LAYOUT *layout;
  /*  char *predefined_format; */

  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_WriteAsciiParameters"))
    return (0);
  layout = &SDDS_dataset->layout;
  for (i = 0; i < layout->n_parameters; i++) {
    if (layout->parameter_definition[i].fixed_value)
      continue;
    if (!SDDS_WriteTypedValue(SDDS_dataset->parameter[i], 0, layout->parameter_definition[i].type, NULL, fp)) {
      SDDS_SetError("Unable to write ascii parameters (SDDS_WriteAsciiParameters)");
      return 0;
    }
    fputc('\n', fp);
  }
  return (1);
}

/**
 * Writes the parameter data of an SDDS dataset in ASCII format to a file with LZMA compression.
 *
 * @param SDDS_dataset The SDDS dataset containing the parameter data.
 * @param lzmafp The LZMA file pointer to write the ASCII data to.
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_LZMAWriteAsciiParameters(SDDS_DATASET *SDDS_dataset, struct lzmafile *lzmafp) {
  int32_t i;
  SDDS_LAYOUT *layout;
  /*  char *predefined_format; */

  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_LZMAWriteAsciiParameters"))
    return (0);
  layout = &SDDS_dataset->layout;
  for (i = 0; i < layout->n_parameters; i++) {
    if (layout->parameter_definition[i].fixed_value)
      continue;
    if (!SDDS_LZMAWriteTypedValue(SDDS_dataset->parameter[i], 0, layout->parameter_definition[i].type, NULL, lzmafp)) {
      SDDS_SetError("Unable to write ascii parameters (SDDS_LZMAWriteAsciiParameters)");
      return 0;
    }
    lzma_putc('\n', lzmafp);
  }
  return (1);
}

#if defined(zLib)
/**
 * Writes the parameter data of an SDDS dataset in ASCII format to a file with GZIP compression.
 *
 * @param SDDS_dataset The SDDS dataset containing the parameter data.
 * @param gzfp The GZIP file pointer to write the ASCII data to.
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_GZipWriteAsciiParameters(SDDS_DATASET *SDDS_dataset, gzFile gzfp) {
  int32_t i;
  SDDS_LAYOUT *layout;
  /*  char *predefined_format; */

  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_GZipWriteAsciiParameters"))
    return (0);
  layout = &SDDS_dataset->layout;
  for (i = 0; i < layout->n_parameters; i++) {
    if (layout->parameter_definition[i].fixed_value)
      continue;
    if (!SDDS_GZipWriteTypedValue(SDDS_dataset->parameter[i], 0, layout->parameter_definition[i].type, NULL, gzfp)) {
      SDDS_SetError("Unable to write ascii parameters (SDDS_GZipWriteAsciiParameters)");
      return 0;
    }
    gzputc(gzfp, '\n');
  }
  return (1);
}
#endif

/**
 * Writes the arrays in ASCII format to the specified file.
 *
 * @param SDDS_dataset The SDDS dataset containing the arrays to be written.
 * @param fp The file pointer to write the arrays to.
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_WriteAsciiArrays(SDDS_DATASET *SDDS_dataset, FILE *fp) {
  int32_t i, j;
  SDDS_LAYOUT *layout;
  /*  char *predefined_format; */
  ARRAY_DEFINITION *array_definition;
  SDDS_ARRAY *array;

  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_WriteAsciiArrays"))
    return (0);
  layout = &SDDS_dataset->layout;
  for (j = 0; j < layout->n_arrays; j++) {
    array_definition = layout->array_definition + j;
    array = &SDDS_dataset->array[j];
    for (i = 0; i < array_definition->dimensions; i++)
      fprintf(fp, "%" PRId32 " ", array->dimension[i]);
    fprintf(fp, "          ! %" PRId32 "-dimensional array %s:\n", array_definition->dimensions, array_definition->name);
    for (i = 0; i < array->elements;) {
      if (!SDDS_WriteTypedValue(array->data, i, array_definition->type, NULL, fp)) {
        SDDS_SetError("Unable to write array--couldn't write ASCII data (SDDS_WriteAsciiArrays)");
        return (0);
      }
      i++;
      if (i % 6 == 0 || i == array->elements)
        fputc('\n', fp);
      else
        fputc(' ', fp);
    }
  }
  return (1);
}

/**
 * Writes the arrays in ASCII format to the specified file in LZMA compressed format.
 *
 * @param SDDS_dataset The SDDS dataset containing the arrays to be written.
 * @param lzmafp The LZMA file pointer to write the arrays to.
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_LZMAWriteAsciiArrays(SDDS_DATASET *SDDS_dataset, struct lzmafile *lzmafp) {
  int32_t i, j;
  SDDS_LAYOUT *layout;
  /*  char *predefined_format; */
  ARRAY_DEFINITION *array_definition;
  SDDS_ARRAY *array;

  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_LZMAWriteAsciiArrays"))
    return (0);
  layout = &SDDS_dataset->layout;
  for (j = 0; j < layout->n_arrays; j++) {
    array_definition = layout->array_definition + j;
    array = &SDDS_dataset->array[j];
    for (i = 0; i < array_definition->dimensions; i++)
      lzma_printf(lzmafp, "%" PRId32 " ", array->dimension[i]);
    lzma_printf(lzmafp, "          ! %" PRId32 "-dimensional array %s:\n", array_definition->dimensions, array_definition->name);
    for (i = 0; i < array->elements;) {
      if (!SDDS_LZMAWriteTypedValue(array->data, i, array_definition->type, NULL, lzmafp)) {
        SDDS_SetError("Unable to write array--couldn't write ASCII data (SDDS_LZMAWriteAsciiArrays)");
        return (0);
      }
      i++;
      if (i % 6 == 0 || i == array->elements)
        lzma_putc('\n', lzmafp);
      else
        lzma_putc(' ', lzmafp);
    }
  }
  return (1);
}

#if defined(zLib)
/**
 * Writes the arrays in ASCII format to the specified file in GZIP compressed format.
 *
 * @param SDDS_dataset The SDDS dataset containing the arrays to be written.
 * @param gzfp The GZIP file pointer to write the arrays to.
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_GZipWriteAsciiArrays(SDDS_DATASET *SDDS_dataset, gzFile gzfp) {
  int32_t i, j;
  SDDS_LAYOUT *layout;
  /*  char *predefined_format; */
  ARRAY_DEFINITION *array_definition;
  SDDS_ARRAY *array;

  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_GZipWriteAsciiArrays"))
    return (0);
  layout = &SDDS_dataset->layout;
  for (j = 0; j < layout->n_arrays; j++) {
    array_definition = layout->array_definition + j;
    array = &SDDS_dataset->array[j];
    for (i = 0; i < array_definition->dimensions; i++)
      gzprintf(gzfp, "%" PRId32 " ", array->dimension[i]);
    gzprintf(gzfp, "          ! %" PRId32 "-dimensional array %s:\n", array_definition->dimensions, array_definition->name);
    for (i = 0; i < array->elements;) {
      if (!SDDS_GZipWriteTypedValue(array->data, i, array_definition->type, NULL, gzfp)) {
        SDDS_SetError("Unable to write array--couldn't write ASCII data (SDDS_GZipWriteAsciiArrays)");
        return (0);
      }
      i++;
      if (i % 6 == 0 || i == array->elements)
        gzputc(gzfp, '\n');
      else
        gzputc(gzfp, ' ');
    }
  }
  return (1);
}
#endif

/**
 * Writes a row of data in ASCII format to the specified file.
 *
 * @param SDDS_dataset The SDDS dataset containing the data.
 * @param row The row number to write.
 * @param fp The file pointer to write the data to.
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_WriteAsciiRow(SDDS_DATASET *SDDS_dataset, int64_t row, FILE *fp) {
  int32_t newline_needed;
  int64_t i, n_per_line, line;
  /*  int32_t  embedded_quotes_present; */
  SDDS_LAYOUT *layout;
  /*  char *predefined_format, *s; */

  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_WriteAsciiRow"))
    return (0);
  layout = &SDDS_dataset->layout;
  if (SDDS_dataset->layout.data_mode.lines_per_row <= 0)
    SDDS_dataset->layout.data_mode.lines_per_row = 1;
  n_per_line = SDDS_dataset->layout.n_columns / SDDS_dataset->layout.data_mode.lines_per_row;
  line = 1;
  newline_needed = 0;
  for (i = 0; i < layout->n_columns; i++) {
    if (!SDDS_WriteTypedValue(SDDS_dataset->data[i], row, layout->column_definition[i].type, NULL, fp)) {
      SDDS_SetError("Unable to write ascii row (SDDS_WriteAsciiRow)");
      return 0;
    }
    if ((i + 1) % n_per_line == 0 && line != SDDS_dataset->layout.data_mode.lines_per_row) {
      newline_needed = 0;
      fputc('\n', fp);
      line++;
    } else {
      fputc(' ', fp);
      newline_needed = 1;
    }
  }
  if (newline_needed)
    fputc('\n', fp);
  return (1);
}

/**
 * Writes a row of data in ASCII format to the specified file in LZMA compressed format.
 *
 * @param SDDS_dataset The SDDS dataset containing the data.
 * @param row The row number to write.
 * @param lzmafp The LZMA file pointer to write the data to.
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_LZMAWriteAsciiRow(SDDS_DATASET *SDDS_dataset, int64_t row, struct lzmafile *lzmafp) {
  int32_t newline_needed;
  int64_t i, n_per_line, line;
  /*  int32_t  embedded_quotes_present; */
  SDDS_LAYOUT *layout;
  /*  char *predefined_format, *s; */

  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_LZMAWriteAsciiRow"))
    return (0);
  layout = &SDDS_dataset->layout;
  if (SDDS_dataset->layout.data_mode.lines_per_row <= 0)
    SDDS_dataset->layout.data_mode.lines_per_row = 1;
  n_per_line = SDDS_dataset->layout.n_columns / SDDS_dataset->layout.data_mode.lines_per_row;
  line = 1;
  newline_needed = 0;
  for (i = 0; i < layout->n_columns; i++) {
    if (!SDDS_LZMAWriteTypedValue(SDDS_dataset->data[i], row, layout->column_definition[i].type, NULL, lzmafp)) {
      SDDS_SetError("Unable to write ascii row (SDDS_LZMAWriteAsciiRow)");
      return 0;
    }
    if ((i + 1) % n_per_line == 0 && line != SDDS_dataset->layout.data_mode.lines_per_row) {
      newline_needed = 0;
      lzma_putc('\n', lzmafp);
      line++;
    } else {
      lzma_putc(' ', lzmafp);
      newline_needed = 1;
    }
  }
  if (newline_needed)
    lzma_putc('\n', lzmafp);
  return (1);
}

#if defined(zLib)
/**
 * Writes a row of data in ASCII format to the specified file in GZIP compressed format.
 *
 * @param SDDS_dataset The SDDS dataset containing the data.
 * @param row The row number to write.
 * @param lzmafp The GZIP file pointer to write the data to.
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_GZipWriteAsciiRow(SDDS_DATASET *SDDS_dataset, int64_t row, gzFile gzfp) {
  int32_t newline_needed;
  int64_t i, n_per_line, line;
  /*  int32_t  embedded_quotes_present; */
  SDDS_LAYOUT *layout;
  /*  char *predefined_format, *s; */

  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_GZipWriteAsciiRow"))
    return (0);
  layout = &SDDS_dataset->layout;
  if (SDDS_dataset->layout.data_mode.lines_per_row <= 0)
    SDDS_dataset->layout.data_mode.lines_per_row = 1;
  n_per_line = SDDS_dataset->layout.n_columns / SDDS_dataset->layout.data_mode.lines_per_row;
  line = 1;
  newline_needed = 0;
  for (i = 0; i < layout->n_columns; i++) {
    if (!SDDS_GZipWriteTypedValue(SDDS_dataset->data[i], row, layout->column_definition[i].type, NULL, gzfp)) {
      SDDS_SetError("Unable to write ascii row (SDDS_GZipWriteAsciiRow)");
      return 0;
    }
    if ((i + 1) % n_per_line == 0 && line != SDDS_dataset->layout.data_mode.lines_per_row) {
      newline_needed = 0;
      gzputc(gzfp, '\n');
      line++;
    } else {
      gzputc(gzfp, ' ');
      newline_needed = 1;
    }
  }
  if (newline_needed)
    gzputc(gzfp, '\n');
  return (1);
}
#endif

/**
 * Reads the parameters from an ASCII file into the SDDS dataset.
 *
 * @param SDDS_dataset The SDDS dataset structure to store the parameters.
 * @return 1 on success,
 *         0 on error,
 *        -1 on EOF
 */
int32_t SDDS_ReadAsciiParameters(SDDS_DATASET *SDDS_dataset) {
  int32_t i, first_read;
  SDDS_LAYOUT *layout;
#if defined(zLib)
  gzFile gzfp;
#endif
  FILE *fp;
  struct lzmafile *lzmafp;
  char *bigBuffer = NULL;
  int32_t bigBufferSize = 0;

  layout = &SDDS_dataset->layout;
  first_read = 1;
  if (layout->n_parameters > 0) {
    if (!(bigBuffer = SDDS_Malloc(sizeof(*bigBuffer) * (bigBufferSize = INITIAL_BIG_BUFFER_SIZE)))) {
      SDDS_SetError("Unable to read parameters--buffer allocation failure (SDDS_ReadAsciiParameters)");
      return (0);
    }
  }
#if defined(zLib)
  if (SDDS_dataset->layout.gzipFile) {
    gzfp = layout->gzfp;
    for (i = 0; i < layout->n_parameters; i++) {
      if (layout->parameter_definition[i].definition_mode & SDDS_WRITEONLY_DEFINITION)
        continue;
      bigBuffer[0] = 0;
      if (!layout->parameter_definition[i].fixed_value) {
        if (!fgetsGZipSkipCommentsResize(SDDS_dataset, &bigBuffer, &bigBufferSize, gzfp, '!')) {
          if (first_read) {
            if (bigBuffer)
              free(bigBuffer);
            return (-1);
          }
          SDDS_SetError("Unable to read parameters--data ends prematurely (SDDS_ReadAsciiParameters)");
          return (0);
        }
        first_read = 0;
        bigBuffer[strlen(bigBuffer) - 1] = 0;
      } else
        strcpy(bigBuffer, layout->parameter_definition[i].fixed_value);
      if (!SDDS_ScanData(bigBuffer, layout->parameter_definition[i].type, 0, SDDS_dataset->parameter[i], 0, 1)) {
        SDDS_SetError("Unable to read page--parameter scanning error (SDDS_ReadAsciiParameters)");
        return (0);
      }
    }
  } else {
#endif
    if (SDDS_dataset->layout.lzmaFile) {
      lzmafp = layout->lzmafp;
      for (i = 0; i < layout->n_parameters; i++) {
        if (layout->parameter_definition[i].definition_mode & SDDS_WRITEONLY_DEFINITION)
          continue;
        bigBuffer[0] = 0;
        if (!layout->parameter_definition[i].fixed_value) {
          if (!fgetsLZMASkipCommentsResize(SDDS_dataset, &bigBuffer, &bigBufferSize, lzmafp, '!')) {
            if (first_read) {
              if (bigBuffer)
                free(bigBuffer);
              return (-1);
            }
            SDDS_SetError("Unable to read parameters--data ends prematurely (SDDS_ReadAsciiParameters)");
            return (0);
          }
          first_read = 0;
          bigBuffer[strlen(bigBuffer) - 1] = 0;
        } else
          strcpy(bigBuffer, layout->parameter_definition[i].fixed_value);
        if (!SDDS_ScanData(bigBuffer, layout->parameter_definition[i].type, 0, SDDS_dataset->parameter[i], 0, 1)) {
          SDDS_SetError("Unable to read page--parameter scanning error (SDDS_ReadAsciiParameters)");
          return (0);
        }
      }
    } else {
      fp = layout->fp;
      for (i = 0; i < layout->n_parameters; i++) {
        if (layout->parameter_definition[i].definition_mode & SDDS_WRITEONLY_DEFINITION)
          continue;
        bigBuffer[0] = 0;
        if (!layout->parameter_definition[i].fixed_value) {
          if (!fgetsSkipCommentsResize(SDDS_dataset, &bigBuffer, &bigBufferSize, fp, '!')) {
            if (first_read) {
              if (bigBuffer)
                free(bigBuffer);
              return (-1);
            }
            SDDS_SetError("Unable to read parameters--data ends prematurely (SDDS_ReadAsciiParameters)");
            return (0);
          }
          first_read = 0;
          bigBuffer[strlen(bigBuffer) - 1] = 0;
        } else
          strcpy(bigBuffer, layout->parameter_definition[i].fixed_value);
        if (!SDDS_ScanData(bigBuffer, layout->parameter_definition[i].type, 0, SDDS_dataset->parameter[i], 0, 1)) {
          SDDS_SetError("Unable to read page--parameter scanning error (SDDS_ReadAsciiParameters)");
          return (0);
        }
      }
    }
#if defined(zLib)
  }
#endif
  if (bigBuffer)
    free(bigBuffer);
  return (1);
}

/**
 * @brief Reads the arrays from an ASCII file into the SDDS dataset.
 *
 * This function reads the arrays from an ASCII file and stores them in the SDDS dataset.
 *
 * @param SDDS_dataset Pointer to the SDDS dataset structure.
 * @return 1 on success,
 *         0 on error,
 *        -1 on EOF
 */
int32_t SDDS_ReadAsciiArrays(SDDS_DATASET *SDDS_dataset) {
  int32_t i, j, length;
  SDDS_LAYOUT *layout;
  char *buffer = NULL;
  int32_t bufferSize = 0;
  char *bigBuffer = NULL;
  int32_t bigBufferSize = 0;
  char *bigBufferCopy;
  int32_t bigBufferCopySize;
  /*  ARRAY_DEFINITION *array_definition; */
  SDDS_ARRAY *array;
#if defined(zLib)
  gzFile gzfp = NULL;
#endif
  FILE *fp = NULL;
  struct lzmafile *lzmafp = NULL;

  layout = &SDDS_dataset->layout;
#if defined(zLib)
  if (SDDS_dataset->layout.gzipFile) {
    gzfp = layout->gzfp;
  } else {
#endif
    if (SDDS_dataset->layout.lzmaFile) {
      lzmafp = layout->lzmafp;
    } else {
      fp = layout->fp;
    }
#if defined(zLib)
  }
#endif
  if (layout->n_arrays > 0) {
    if (!(bigBuffer = SDDS_Malloc(sizeof(*bigBuffer) * (bigBufferSize = INITIAL_BIG_BUFFER_SIZE)))) {
      SDDS_SetError("Unable to read parameters--buffer allocation failure (SDDS_ReadAsciiArrays)");
      return (0);
    }
  }
  for (i = 0; i < layout->n_arrays; i++) {
#if defined(zLib)
    if (SDDS_dataset->layout.gzipFile) {
      if (!fgetsGZipSkipCommentsResize(SDDS_dataset, &bigBuffer, &bigBufferSize, gzfp, '!') || SDDS_StringIsBlank(bigBuffer)) {
        if (i == 0)
          return (-1);
        SDDS_SetError("Unable to read array--dimensions missing (SDDS_ReadAsciiArrays)");
        return (0);
      }
    } else {
#endif
      if (SDDS_dataset->layout.lzmaFile) {
        if (!fgetsLZMASkipCommentsResize(SDDS_dataset, &bigBuffer, &bigBufferSize, lzmafp, '!') || SDDS_StringIsBlank(bigBuffer)) {
          if (i == 0)
            return (-1);
          SDDS_SetError("Unable to read array--dimensions missing (SDDS_ReadAsciiArrays)");
          return (0);
        }
      } else {
        if (!fgetsSkipCommentsResize(SDDS_dataset, &bigBuffer, &bigBufferSize, fp, '!') || SDDS_StringIsBlank(bigBuffer)) {
          if (i == 0)
            return (-1);
          SDDS_SetError("Unable to read array--dimensions missing (SDDS_ReadAsciiArrays)");
          return (0);
        }
      }
#if defined(zLib)
    }
#endif
    if (!(array = SDDS_dataset->array + i)) {
      SDDS_SetError("Unable to read array--pointer to structure storage area is NULL (SDDS_ReadAsciiArrays)");
      return (0);
    }
    if (array->definition && !SDDS_FreeArrayDefinition(array->definition)) {
      SDDS_SetError("Unable to get array--array definition corrupted (SDDS_ReadAsciiArrays)");
      return (0);
    }
    if (!SDDS_CopyArrayDefinition(&array->definition, layout->array_definition + i)) {
      SDDS_SetError("Unable to read array--definition copy failed (SDDS_ReadAsciiArrays)");
      return (0);
    }
    /*if (array->dimension) free(array->dimension); */
    if (!(array->dimension = SDDS_Realloc(array->dimension, sizeof(*array->dimension) * array->definition->dimensions))) {
      SDDS_SetError("Unable to read array--allocation failure (SDDS_ReadAsciiArrays)");
      return (0);
    }
    array->elements = 1;
    if ((length = strlen(bigBuffer)) >= bufferSize) {
      if (!(buffer = SDDS_Realloc(buffer, sizeof(*buffer) * (bufferSize = 2 * length)))) {
        SDDS_SetError("Unable to scan data--allocation failure (SDDS_ReadAsciiArrays");
        return (0);
      }
    }
    for (j = 0; j < array->definition->dimensions; j++) {
      if (SDDS_GetToken(bigBuffer, buffer, SDDS_MAXLINE) <= 0 || sscanf(buffer, "%" SCNd32, array->dimension + j) != 1 || array->dimension[j] < 0) {
        SDDS_SetError("Unable to read array--dimensions missing or negative (SDDS_ReadAsciiArrays)");
        return (0);
      }
      array->elements *= array->dimension[j];
    }
    if (!array->elements)
      continue;
    if (array->data)
      free(array->data);
    array->data = array->pointer = NULL;
    if (!(array->data = SDDS_Realloc(array->data, array->elements * SDDS_type_size[array->definition->type - 1]))) {
      SDDS_SetError("Unable to read array--allocation failure (SDDS_ReadAsciiArrays)");
      return (0);
    }
    SDDS_ZeroMemory(array->data, array->elements * SDDS_type_size[array->definition->type - 1]);
    j = 0;
    bigBuffer[0] = 0;
    do {
      if (SDDS_StringIsBlank(bigBuffer)) {
        bigBuffer[0] = 0;
#if defined(zLib)
        if (SDDS_dataset->layout.gzipFile) {
          if (!fgetsGZipSkipCommentsResize(SDDS_dataset, &bigBuffer, &bigBufferSize, gzfp, '!') || SDDS_StringIsBlank(bigBuffer)) {
            SDDS_SetError("Unable to read array--dimensions missing (SDDS_ReadAsciiArrays)");
            return (0);
          }
        } else {
#endif
          if (SDDS_dataset->layout.lzmaFile) {
            if (!fgetsLZMASkipCommentsResize(SDDS_dataset, &bigBuffer, &bigBufferSize, lzmafp, '!') || SDDS_StringIsBlank(bigBuffer)) {
              SDDS_SetError("Unable to read array--dimensions missing (SDDS_ReadAsciiArrays)");
              return (0);
            }
          } else {
            if (!fgetsSkipCommentsResize(SDDS_dataset, &bigBuffer, &bigBufferSize, fp, '!') || SDDS_StringIsBlank(bigBuffer)) {
              SDDS_SetError("Unable to read array--dimensions missing (SDDS_ReadAsciiArrays)");
              return (0);
            }
          }
#if defined(zLib)
        }
#endif
      }
      /* copy the bigBuffer because SDDS_ScanData2 will change the bigBufferCopy pointer that SDDS_ScanData
         cannot do becuase bigBuffer is a static variable. This change was implemented to greatly improve
         the speed of reading a very large line of array elements. The previous version just did a lot of
         strcpy commands to these huge lines which was really slow. */
      bigBufferCopy = bigBuffer;
      bigBufferCopySize = strlen(bigBufferCopy);
      do {
        if (!SDDS_ScanData2(bigBufferCopy, &bigBufferCopy, &bigBufferCopySize, array->definition->type, array->definition->field_length, array->data, j, 0)) {
          SDDS_SetError("Unable to read array--error scanning data element (SDDS_ReadAsciiArrays)");
          return (0);
        }
      } while (++j < array->elements && !SDDS_StringIsBlank(bigBufferCopy));
      bigBuffer[0] = 0;
    } while (j < array->elements);
  }
  if (buffer)
    free(buffer);
  if (bigBuffer)
    free(bigBuffer);
  return (1);
}

/**
 * Reads then next SDDS ASCII page into memory.
 *
 * @param SDDS_dataset A pointer to an SDDS dataset
 * @param sparse_interval The column data can be sparsed over row intervals if this is greater than 1
 * @param sparse_offset This is used to skip the initial rows of the column data
 *
 * @return Page number on success,
 *         -1 if it is the end-of-file,
 *         0 on error
 */
int32_t SDDS_ReadAsciiPage(SDDS_DATASET *SDDS_dataset, int64_t sparse_interval, int64_t sparse_offset) {
  return SDDS_ReadAsciiPageDetailed(SDDS_dataset, sparse_interval, sparse_offset, 0);
}
/**
 * Reads the last specified number of rows from an ASCII page of an SDDS dataset.
 *
 * @param SDDS_dataset The pointer to the SDDS dataset structure.
 * @param last_rows The number of rows to read from the SDDS_dataset.
 * @return Page number on success,
 *         -1 if it is the end-of-file,
 *         0 on error
 */
int32_t SDDS_ReadAsciiPageLastRows(SDDS_DATASET *SDDS_dataset, int64_t last_rows) {
  return SDDS_ReadAsciiPageDetailed(SDDS_dataset, 1, 0, last_rows);
}

/**
 * @brief Reads a page of data from an ASCII file into an SDDS dataset.
 *
 * This function reads a page of data from an ASCII file into an SDDS dataset.
 * It allows specifying a sparse interval, sparse offset, and the number of last rows to read.
 *
 * @param SDDS_dataset The SDDS dataset to read the data into.
 * @param sparse_interval The interval between rows to read. Use 1 to read every row.
 * @param sparse_offset The offset from the first row to start reading.
 * @param last_rows The number of last rows to read. Use 0 to read all rows. This overrides sparse_interval and sparse_offset.
 *
 * @return Page number on success,
 *         -1 if it is the end-of-file,
 *         0 on error
 */
int32_t SDDS_ReadAsciiPageDetailed(SDDS_DATASET *SDDS_dataset, int64_t sparse_interval, int64_t sparse_offset, int64_t last_rows) {
  SDDS_LAYOUT *layout;
  int32_t no_row_counts, end_of_data, retval, lineCount;
  int64_t n_rows, i, j, k, rows_to_store;
  /*  int32_t page_number; */
  char s[SDDS_MAXLINE];
  char *dataRead, *bigBufferCopy;
  int32_t bigBufferCopySize;
  char *bigBuffer = NULL;
  int32_t bigBufferSize = 0;
  /*  char *ptr; */
#if defined(zLib)
  gzFile gzfp = NULL;
#endif
  FILE *fp = NULL;
  struct lzmafile *lzmafp = NULL;
  /*  double value; */

  if (SDDS_dataset->autoRecovered)
    return -1;

  SDDS_SetReadRecoveryMode(SDDS_dataset, 0);

#if defined(zLib)
  if (SDDS_dataset->layout.gzipFile) {
    gzfp = SDDS_dataset->layout.gzfp;
  } else {
#endif
    if (SDDS_dataset->layout.lzmaFile) {
      lzmafp = SDDS_dataset->layout.lzmafp;
    } else {
      fp = SDDS_dataset->layout.fp;
    }
#if defined(zLib)
  }
#endif
  if (SDDS_dataset->page_number == -1)
    /* end of file already hit */
    return -1;
#if defined(zLib)
  if (SDDS_dataset->layout.gzipFile) {
    if (gzeof(gzfp) && SDDS_dataset->page_number > 0)
      /* end of file and not first page about to be read */
      return SDDS_dataset->page_number = -1;
  } else {
#endif
    if (SDDS_dataset->layout.lzmaFile) {
      if (lzma_eof(lzmafp) && SDDS_dataset->page_number > 0)
        /* end of file and not first page about to be read */
        return SDDS_dataset->page_number = -1;
    } else {
      if (feof(fp) && SDDS_dataset->page_number > 0)
        /* end of file and not first page about to be read */
        return SDDS_dataset->page_number = -1;
    }
#if defined(zLib)
  }
#endif
  if (!SDDS_AsciiDataExpected(SDDS_dataset) && SDDS_dataset->page_number != 0)
    /* if no columns or arrays and only fixed value parameters, then only one "page" allowed */
    return (SDDS_dataset->page_number = -1);

#if defined(zLib)
  if (SDDS_dataset->layout.gzipFile) {
    if (SDDS_dataset->page_number == 0)
      for (i = 0; i < SDDS_dataset->layout.data_mode.additional_header_lines; i++) {
        if (!fgetsGZipSkipComments(SDDS_dataset, s, SDDS_MAXLINE, gzfp, '!'))
          return (SDDS_dataset->page_number = -1); /* indicates end of data */
      }
  } else {
#endif
    if (SDDS_dataset->layout.lzmaFile) {
      if (SDDS_dataset->page_number == 0)
        for (i = 0; i < SDDS_dataset->layout.data_mode.additional_header_lines; i++) {
          if (!fgetsLZMASkipComments(SDDS_dataset, s, SDDS_MAXLINE, lzmafp, '!'))
            return (SDDS_dataset->page_number = -1); /* indicates end of data */
        }
    } else {
      if (SDDS_dataset->page_number == 0)
        for (i = 0; i < SDDS_dataset->layout.data_mode.additional_header_lines; i++) {
          if (!fgetsSkipComments(SDDS_dataset, s, SDDS_MAXLINE, fp, '!'))
            return (SDDS_dataset->page_number = -1); /* indicates end of data */
        }
    }
#if defined(zLib)
  }
#endif

  /* the next call increments the page number */
  if (!SDDS_StartPage(SDDS_dataset, 0)) {
    SDDS_SetError("Unable to read page--couldn't start page (SDDS_ReadAsciiPage)");
    return (0);
  }

  layout = &SDDS_dataset->layout;

  if ((retval = SDDS_ReadAsciiParameters(SDDS_dataset)) < 1) {
    if (retval)
      return (SDDS_dataset->page_number = retval);
    SDDS_SetError("Unable to read page--couldn't read parameters (SDDS_ReadAsciiPage)");
    return (0);
  }
  if ((retval = SDDS_ReadAsciiArrays(SDDS_dataset)) < 1) {
    if (retval)
      return (SDDS_dataset->page_number = retval);
    SDDS_SetError("Unable to read page--couldn't read arrays (SDDS_ReadAsciiPage)");
    return (0);
  }

  if (last_rows < 0)
    last_rows = 0;
  if (sparse_interval <= 0)
    sparse_interval = 1;
  if (sparse_offset < 0)
    sparse_offset = 0;

  SDDS_dataset->rowcount_offset = -1;
  if (layout->n_columns) {
    if (!(bigBuffer = SDDS_Malloc(sizeof(*bigBuffer) * (bigBufferSize = INITIAL_BIG_BUFFER_SIZE)))) {
      SDDS_SetError("Unable to read parameters--buffer allocation failure (SDDS_ReadAsciiPage)");
      return (0);
    }

    n_rows = 0;
    no_row_counts = 0;
    if (!SDDS_dataset->layout.data_mode.no_row_counts) {
      /* read the number of rows */
#if defined(zLib)
      if (SDDS_dataset->layout.gzipFile) {
        if (!fgetsGZipSkipComments(SDDS_dataset, s, SDDS_MAXLINE, gzfp, '!')) {
          if (bigBuffer)
            free(bigBuffer);
          return (SDDS_dataset->page_number = -1); /* indicates end of data */
        }
      } else {
#endif
        if (SDDS_dataset->layout.lzmaFile) {
          if (!fgetsLZMASkipComments(SDDS_dataset, s, SDDS_MAXLINE, lzmafp, '!')) {
            if (bigBuffer)
              free(bigBuffer);
            return (SDDS_dataset->page_number = -1); /* indicates end of data */
          }
        } else {
          do {
            SDDS_dataset->rowcount_offset = ftell(fp);
            if (!fgets(s, SDDS_MAXLINE, fp)) {
              if (bigBuffer)
                free(bigBuffer);
              return (SDDS_dataset->page_number = -1); /* indicates end of data */
            }
          } while (s[0] == '!');
        }
#if defined(zLib)
      }
#endif
      if (sscanf(s, "%" SCNd64, &n_rows) != 1 || n_rows < 0) {
        SDDS_SetError("Unable to read page--file has no (valid) number-of-rows entry (SDDS_ReadAsciiPage)");
        return (0);
      }
      if (n_rows > SDDS_GetRowLimit()) {
        /* number of rows is "unreasonably" large---treat like end-of-file */
        if (bigBuffer)
          free(bigBuffer);
        return (SDDS_dataset->page_number = -1);
      }
      if (last_rows) {
        sparse_interval = 1;
        sparse_offset = n_rows - last_rows;
        if (sparse_offset < 0)
          sparse_offset = 0;
      }
      rows_to_store = (n_rows - sparse_offset) / sparse_interval + 2;
    } else {
      // fix last_rows when no row count is available
      no_row_counts = 1;
      n_rows = TABLE_LENGTH_INCREMENT;
      rows_to_store = n_rows;
    }

    if (rows_to_store >= SDDS_dataset->n_rows_allocated) {
      /* lengthen the page */
      if (!SDDS_LengthenTable(SDDS_dataset, rows_to_store - SDDS_dataset->n_rows_allocated)) {
        SDDS_SetError("Unable to read page--couldn't lengthen data page (SDDS_ReadAsciiPage)");
        return (0);
      }
    }

    /* read the page values */
    j = end_of_data = k = 0;
    s[0] = 0;
    if (!no_row_counts && n_rows == 0) {
      SDDS_dataset->n_rows = 0;
      if (bigBuffer)
        free(bigBuffer);
      return (SDDS_dataset->page_number);
    }
    bigBuffer[0] = 0;
    bigBufferCopy = bigBuffer;
    do {
      if (j >= SDDS_dataset->n_rows_allocated) {
        /* lengthen the page */
        if (!SDDS_LengthenTable(SDDS_dataset, TABLE_LENGTH_INCREMENT)) {
          SDDS_SetError("Unable to read page--couldn't lengthen data page (SDDS_ReadAsciiPage)");
          return (0);
        }
      }
      lineCount = 0;
      dataRead = NULL;
      for (i = 0; i < SDDS_dataset->layout.n_columns; i++) {
        if (layout->column_definition[i].definition_mode & SDDS_WRITEONLY_DEFINITION)
          continue;
        if (SDDS_StringIsBlank(bigBufferCopy)) {
          bigBuffer[0] = 0;
          bigBufferCopy = bigBuffer;
          dataRead = NULL;
#if defined(zLib)
          if (SDDS_dataset->layout.gzipFile) {
            if (!(dataRead = fgetsGZipSkipCommentsResize(SDDS_dataset, &bigBuffer, &bigBufferSize, gzfp, '!')) || SDDS_StringIsBlank(bigBuffer)) {
              SDDS_dataset->n_rows = j;
              if (no_row_counts) {
                /* legitmate end of data set */
                end_of_data = 1;
                break;
              }
              /* error, but may be recoverable */
              gzseek(gzfp, 0L, SEEK_END);
              if (SDDS_dataset->autoRecover) {
                SDDS_dataset->autoRecovered = 1;
                SDDS_ClearErrors();
                if (bigBuffer)
                  free(bigBuffer);
                return (SDDS_dataset->page_number);
              }
              SDDS_SetError("Unable to read page (SDDS_ReadAsciiPage)");
              SDDS_SetReadRecoveryMode(SDDS_dataset, 1);
              return (0);
            }
          } else {
#endif
            if (SDDS_dataset->layout.lzmaFile) {
              if (!(dataRead = fgetsLZMASkipCommentsResize(SDDS_dataset, &bigBuffer, &bigBufferSize, lzmafp, '!')) || SDDS_StringIsBlank(bigBuffer)) {
                SDDS_dataset->n_rows = j;
                if (no_row_counts) {
                  /* legitmate end of data set */
                  end_of_data = 1;
                  break;
                }
                /* error, but may be recoverable */
                lzma_seek(lzmafp, 0L, SEEK_END);
                if (SDDS_dataset->autoRecover) {
                  SDDS_dataset->autoRecovered = 1;
                  SDDS_ClearErrors();
                  if (bigBuffer)
                    free(bigBuffer);
                  return (SDDS_dataset->page_number);
                }
                SDDS_SetError("Unable to read page (SDDS_ReadAsciiPage)");
                SDDS_SetReadRecoveryMode(SDDS_dataset, 1);
                return (0);
              }
            } else {
              if (!(dataRead = fgetsSkipCommentsResize(SDDS_dataset, &bigBuffer, &bigBufferSize, fp, '!')) || SDDS_StringIsBlank(bigBuffer)) {
                SDDS_dataset->n_rows = j;
                if (no_row_counts) {
                  /* legitmate end of data set */
                  end_of_data = 1;
                  break;
                }
                /* error, but may be recoverable */
                fseek(fp, 0L, SEEK_END);
                if (SDDS_dataset->autoRecover) {
                  SDDS_dataset->autoRecovered = 1;
                  SDDS_ClearErrors();
                  if (bigBuffer)
                    free(bigBuffer);
                  return (SDDS_dataset->page_number);
                }
                SDDS_SetError("Unable to read page (SDDS_ReadAsciiPage)");
                SDDS_SetReadRecoveryMode(SDDS_dataset, 1);
                return (0);
              }
            }
#if defined(zLib)
          }
#endif
          lineCount++;
          bigBufferCopy = bigBuffer;
          bigBufferCopySize = strlen(bigBufferCopy);
        }
        if (!SDDS_ScanData2(bigBufferCopy, &bigBufferCopy, &bigBufferCopySize, layout->column_definition[i].type, layout->column_definition[i].field_length, SDDS_dataset->data[i], j, 0)) {
          /* error, but may be recoverable */
          SDDS_dataset->n_rows = j;
#if defined(zLib)
          if (SDDS_dataset->layout.gzipFile) {
            gzseek(gzfp, 0L, SEEK_END);
          } else {
#endif
            if (SDDS_dataset->layout.lzmaFile) {
              lzma_seek(lzmafp, 0L, SEEK_END);
            } else {
              fseek(fp, 0L, SEEK_END);
            }
#if defined(zLib)
          }
#endif
          if (SDDS_dataset->autoRecover) {
            SDDS_dataset->autoRecovered = 1;
            SDDS_ClearErrors();
            if (bigBuffer)
              free(bigBuffer);
            return (SDDS_dataset->page_number);
          }
          SDDS_SetReadRecoveryMode(SDDS_dataset, 1);
          SDDS_SetError("Unable to read page--scanning error (SDDS_ReadAsciiPage)");
          return (0);
        }
#if defined(DEBUG)
        fprintf(stderr, "line remaining = %s\n", bigBuffer);
#endif
      }
      if (end_of_data)
        /* ran out of data for no_row_counts=1 */
        break;
      if (layout->data_mode.lines_per_row != 0 && lineCount != layout->data_mode.lines_per_row) {
        sprintf(s, "Unable to read page--line layout error at line %" PRId64 " of page %" PRId32 " (SDDS_ReadAsciiPage)", j + 1, SDDS_dataset->page_number);
        SDDS_SetError(s);
        /* data ends prematurely, which is an error that may be recoverable */
#if defined(zLib)
        if (SDDS_dataset->layout.gzipFile) {
          gzseek(gzfp, 0L, SEEK_END);
        } else {
#endif
          if (SDDS_dataset->layout.lzmaFile) {
            lzma_seek(lzmafp, 0L, SEEK_END);
          } else {
            fseek(fp, 0L, SEEK_END);
          }
#if defined(zLib)
        }
#endif
        if (SDDS_dataset->autoRecover) {
          SDDS_dataset->autoRecovered = 1;
          SDDS_ClearErrors();
          if (bigBuffer)
            free(bigBuffer);
          return (SDDS_dataset->page_number);
        }
        SDDS_SetReadRecoveryMode(SDDS_dataset, 1);
        SDDS_dataset->n_rows = j;
        return (0);
      }
      if (layout->data_mode.lines_per_row != 0) {
        bigBuffer[0] = 0;
        bigBufferCopy = bigBuffer;
      }
      if (--sparse_offset < 0 && k % sparse_interval == 0)
        j++;
      k++;
    } while (k < n_rows || no_row_counts);
    if (end_of_data && !(SDDS_dataset->page_number == 1) && j == 0 && !dataRead) {
      /* For end of data in no_row_counts=1 mode for any page other than the first,
       * an end-of-file is not a valid way to end an empty page (only an incomplete line is)
       */
      if (bigBuffer)
        free(bigBuffer);
      return SDDS_dataset->page_number = -1;
    }
    SDDS_dataset->n_rows = j;
  }
  if (bigBuffer)
    free(bigBuffer);

  return (SDDS_dataset->page_number);
}

/**
 * Scan string data and save the values into a data pointer
 *
 * @param string The data to be scanned
 * @param type The SDDS data type to store the values as
 * @param field_length Set to 0
 * @param data Void pointer to save the value to. Must already have memory allocated to it.
 * @param index Array index of data variable to save the value to
 * @param is_parameter Set to 1 if the data is from an SDDS parameter
 *
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_ScanData(char *string, int32_t type, int32_t field_length, void *data, int64_t index, int32_t is_parameter) {
  char *buffer = NULL;
  int32_t abs_field_length, length;
  int32_t bufferSize = 0;

  abs_field_length = abs(field_length);
  if (!string) {
    SDDS_SetError("Unable to scan data--input string is NULL (SDDS_ScanData)");
    return (0);
  }
  if (!data) {
    SDDS_SetError("Unable to scan data--data pointer is NULL (SDDS_ScanData)");
    return (0);
  }
  if (!(buffer = SDDS_Malloc(sizeof(*buffer) * (bufferSize = SDDS_MAXLINE)))) {
    SDDS_SetError("Unable to scan data--allocation failure (SDDS_ScanData)");
    return (0);
  }
  if ((length = strlen(string)) < abs_field_length)
    length = abs_field_length;
  if (bufferSize <= length) {
    if (!(buffer = SDDS_Realloc(buffer, sizeof(*buffer) * (bufferSize = 2 * length)))) {
      /* I allocate 2*length in the hopes that I won't need to realloc too often if I do this */
      SDDS_SetError("Unable to scan data--allocation failure (SDDS_ScanData)");
      return (0);
    }
  }
  if (type != SDDS_STRING) {
    /* for non-string data, fill buffer with string to be scanned */
    if (field_length) {
      /* fill with fixed number of characters */
      if (abs_field_length > (int32_t)strlen(string)) {
        strcpy(buffer, string);
        *string = 0;
      } else {
        strncpy(buffer, string, abs_field_length);
        buffer[field_length] = 0;
        strcpy(string, string + abs_field_length);
      }
    } else if (SDDS_GetToken(string, buffer, bufferSize) < 0) {
      SDDS_SetError("Unable to scan data--tokenizing error (SDDS_ScanData)");
      return (0);
    }
  }
  switch (type) {
  case SDDS_SHORT:
    if (sscanf(buffer, "%hd", ((short *)data) + index) == 1) {
      if (buffer)
        free(buffer);
      return (1);
    }
    break;
  case SDDS_USHORT:
    if (sscanf(buffer, "%hu", ((unsigned short *)data) + index) == 1) {
      if (buffer)
        free(buffer);
      return (1);
    }
    break;
  case SDDS_LONG:
    if (sscanf(buffer, "%" SCNd32, ((int32_t *)data) + index) == 1) {
      if (buffer)
        free(buffer);
      return (1);
    }
    break;
  case SDDS_ULONG:
    if (sscanf(buffer, "%" SCNu32, ((uint32_t *)data) + index) == 1) {
      if (buffer)
        free(buffer);
      return (1);
    }
    break;
  case SDDS_LONG64:
    if (sscanf(buffer, "%" SCNd64, ((int64_t *)data) + index) == 1) {
      if (buffer)
        free(buffer);
      return (1);
    }
    break;
  case SDDS_ULONG64:
    if (sscanf(buffer, "%" SCNu64, ((uint64_t *)data) + index) == 1) {
      if (buffer)
        free(buffer);
      return (1);
    }
    break;
  case SDDS_FLOAT:
    if (sscanf(buffer, "%f", ((float *)data) + index) == 1) {
      if (buffer)
        free(buffer);
      return (1);
    }
    break;
  case SDDS_DOUBLE:
    if (sscanf(buffer, "%lf", ((double *)data) + index) == 1) {
      if (buffer)
        free(buffer);
      return (1);
    }
    break;
  case SDDS_LONGDOUBLE:
    if (sscanf(buffer, "%Lf", ((long double *)data) + index) == 1) {
      if (buffer)
        free(buffer);
      return (1);
    }
    break;
  case SDDS_STRING:
    if (is_parameter) {
      int32_t len;
      if (((char **)data)[index]) {
        free(((char **)data)[index]);
        ((char **)data)[index] = NULL;
      }
      if ((len = strlen(string)) > 0) {
        if (string[len - 1] == '\r')
          string[len - 1] = 0;
      }
      if (string[0] == '"')
        SDDS_GetToken(string, buffer, bufferSize);
      else
        strcpy(buffer, string);
      SDDS_InterpretEscapes(buffer);
      if (SDDS_CopyString(((char **)data) + index, buffer)) {
        if (buffer)
          free(buffer);
        return (1);
      }
    } else {
      if (field_length) {
        if (abs_field_length > (int32_t)strlen(string)) {
          strcpy(buffer, string);
          *string = 0;
        } else {
          strncpy(buffer, string, abs_field_length);
          buffer[abs_field_length] = 0;
          strcpy(string, string + abs_field_length);
        }
        if (field_length < 0)
          SDDS_RemovePadding(buffer);
      } else if (SDDS_GetToken(string, buffer, bufferSize) < 0)
        break;
      if (((char **)data)[index]) {
        free(((char **)data)[index]);
        ((char **)data)[index] = NULL;
      }
      SDDS_InterpretEscapes(buffer);
      if (SDDS_CopyString(((char **)data) + index, buffer)) {
        if (buffer)
          free(buffer);
        return (1);
      }
    }
    break;
  case SDDS_CHARACTER:
    SDDS_InterpretEscapes(buffer);
    *(((char *)data) + index) = buffer[0];
    if (buffer)
      free(buffer);
    return (1);
  default:
    SDDS_SetError("Unknown data type encountered (SDDS_ScanData)");
    return (0);
  }
  SDDS_SetError("Unable to scan data--scanning or allocation error (SDDS_ScanData)");
  return (0);
}

/**
 * Scan string data and save the values into a data pointer.
 * Faster than SDDS_ScanData for very long strings but results in the string variable to be shortened after each call.
 *
 * @param string The data to be scanned
 * @param pstring The pointer to the data to be scanned
 * @param strlength The string length of the data to be scanned
 * @param type The SDDS data type to store the values as
 * @param field_length Set to 0
 * @param data Void pointer to save the value to. Must already have memory allocated to it.
 * @param index Array index of data variable to save the value to
 * @param is_parameter Set to 1 if the data is from an SDDS parameter
 *
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_ScanData2(char *string, char **pstring, int32_t *strlength, int32_t type, int32_t field_length, void *data, int64_t index, int32_t is_parameter) {
  char *buffer = NULL;
  int32_t abs_field_length, length;
  int32_t bufferSize = 0;

  abs_field_length = abs(field_length);
  if (!string) {
    SDDS_SetError("Unable to scan data--input string is NULL (SDDS_ScanData2)");
    return (0);
  }
  if (!data) {
    SDDS_SetError("Unable to scan data--data pointer is NULL (SDDS_ScanData2)");
    return (0);
  }
  if (!(buffer = SDDS_Malloc(sizeof(*buffer) * (bufferSize = SDDS_MAXLINE)))) {
    SDDS_SetError("Unable to scan data--allocation failure (SDDS_ScanData2)");
    return (0);
  }
  length = *strlength;
  if (length < abs_field_length)
    length = abs_field_length;
  if (bufferSize <= length) {
    if (!(buffer = SDDS_Realloc(buffer, sizeof(*buffer) * (bufferSize = 2 * length)))) {
      /* I allocate 2*length in the hopes that I won't need to realloc too often if I do this */
      SDDS_SetError("Unable to scan data--allocation failure (SDDS_ScanData2)");
      return (0);
    }
  }
  if (type != SDDS_STRING) {
    /* for non-string data, fill buffer with string to be scanned */
    if (field_length) {
      /* fill with fixed number of characters */
      if (abs_field_length > *strlength) {
        strcpy(buffer, string);
        **pstring = 0;
        *strlength = 0;
      } else {
        strncpy(buffer, string, abs_field_length);
        buffer[abs_field_length] = 0;
        *pstring += abs_field_length;
        *strlength -= abs_field_length;
      }
    } else if (SDDS_GetToken2(string, pstring, strlength, buffer, bufferSize) < 0) {
      SDDS_SetError("Unable to scan data--tokenizing error (SDDS_ScanData2)");
      return (0);
    }
  }
  switch (type) {
  case SDDS_SHORT:
    if (sscanf(buffer, "%hd", ((short *)data) + index) == 1) {
      if (buffer)
        free(buffer);
      return (1);
    }
    break;
  case SDDS_USHORT:
    if (sscanf(buffer, "%hu", ((unsigned short *)data) + index) == 1) {
      if (buffer)
        free(buffer);
      return (1);
    }
    break;
  case SDDS_LONG:
    if (sscanf(buffer, "%" SCNd32, ((int32_t *)data) + index) == 1) {
      if (buffer)
        free(buffer);
      return (1);
    }
    break;
  case SDDS_ULONG:
    if (sscanf(buffer, "%" SCNu32, ((uint32_t *)data) + index) == 1) {
      if (buffer)
        free(buffer);
      return (1);
    }
    break;
  case SDDS_LONG64:
    if (sscanf(buffer, "%" SCNd64, ((int64_t *)data) + index) == 1) {
      if (buffer)
        free(buffer);
      return (1);
    }
    break;
  case SDDS_ULONG64:
    if (sscanf(buffer, "%" SCNu64, ((uint64_t *)data) + index) == 1) {
      if (buffer)
        free(buffer);
      return (1);
    }
    break;
  case SDDS_FLOAT:
    if (sscanf(buffer, "%f", ((float *)data) + index) == 1) {
      if (buffer)
        free(buffer);
      return (1);
    }
    break;
  case SDDS_DOUBLE:
    if (sscanf(buffer, "%lf", ((double *)data) + index) == 1) {
      if (buffer)
        free(buffer);
      return (1);
    }
    break;
  case SDDS_LONGDOUBLE:
    if (sscanf(buffer, "%Lf", ((long double *)data) + index) == 1) {
      if (buffer)
        free(buffer);
      return (1);
    }
    break;
  case SDDS_STRING:
    if (is_parameter) {
      int32_t len;
      if (((char **)data)[index]) {
        free(((char **)data)[index]);
        ((char **)data)[index] = NULL;
      }
      if ((len = *strlength) > 0) {
        if (*pstring[len - 1] == '\r') {
          *pstring[len - 1] = 0;
          *strlength -= 1;
        }
      }
      if (*pstring[0] == '"')
        SDDS_GetToken2(*pstring, pstring, strlength, buffer, bufferSize);
      else
        strcpy(buffer, string);
      SDDS_InterpretEscapes(buffer);
      if (SDDS_CopyString(((char **)data) + index, buffer)) {
        if (buffer)
          free(buffer);
        return (1);
      }
    } else {
      if (field_length) {
        if (abs_field_length > *strlength) {
          strcpy(buffer, string);
          **pstring = 0;
          *strlength = 0;
        } else {
          strncpy(buffer, string, abs_field_length);
          buffer[abs_field_length] = 0;
          *pstring += abs_field_length;
          *strlength -= abs_field_length;
        }
        if (field_length < 0)
          SDDS_RemovePadding(buffer);
      } else if (SDDS_GetToken2(string, pstring, strlength, buffer, bufferSize) < 0)
        break;
      if (((char **)data)[index]) {
        free(((char **)data)[index]);
        ((char **)data)[index] = NULL;
      }
      SDDS_InterpretEscapes(buffer);
      if (SDDS_CopyString(((char **)data) + index, buffer)) {
        if (buffer)
          free(buffer);
        return (1);
      }
    }
    break;
  case SDDS_CHARACTER:
    SDDS_InterpretEscapes(buffer);
    *(((char *)data) + index) = buffer[0];
    if (buffer)
      free(buffer);
    return 1;
  default:
    SDDS_SetError("Unknown data type encountered (SDDS_ScanData2)");
    return (0);
  }
  SDDS_SetError("Unable to scan data--scanning or allocation error (SDDS_ScanData2)");
  return (0);
}

/**
 * Determines if the SDDS dataset expects ASCII data. This is almost always true unless there are only fixed value parameters.
 *
 * @param SDDS_dataset The SDDS dataset to check.
 * @return 1 if ASCII data is expected,
 *         0 if ASCII data is not expected
 */
int32_t SDDS_AsciiDataExpected(SDDS_DATASET *SDDS_dataset) {
  int32_t i;
  if (SDDS_dataset->layout.n_columns || SDDS_dataset->layout.n_arrays)
    return (1);
  for (i = 0; i < SDDS_dataset->layout.n_parameters; i++)
    if (!SDDS_dataset->layout.parameter_definition[i].fixed_value)
      return (1);
  return (0);
}

/**
 * @brief Updates the ASCII page of an SDDS dataset.
 *
 * This function updates the ASCII page of the specified SDDS dataset based on the given mode. It is called by SDDS_UpdatePage().
 *
 * @param SDDS_dataset The pointer to the SDDS dataset.
 * @param mode The mode indicating how the ASCII page should be updated.
 *
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_UpdateAsciiPage(SDDS_DATASET *SDDS_dataset, uint32_t mode) {
  FILE *fp;
  int32_t code;
  int64_t i, rows, offset;
  SDDS_FILEBUFFER *fBuffer;

#ifdef DEBUG
  fprintf(stderr, "%" PRId64 " virtual rows present, first=%" PRId32 "\n", SDDS_CountRowsOfInterest(SDDS_dataset), SDDS_dataset->first_row_in_mem);
#endif
  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_UpdateAsciiPage"))
    return (0);
#if defined(zLib)
  if (SDDS_dataset->layout.gzipFile) {
    SDDS_SetError("Unable to perform page updates on a gzip file (SDDS_UpdateAsciiPage)");
    return 0;
  }
#endif
  if (SDDS_dataset->layout.lzmaFile) {
    SDDS_SetError("Unable to perform page updates on an .lzma or .xz file (SDDS_UpdateAsciiPage)");
    return 0;
  }
  if (!SDDS_dataset->writing_page) {
    if (!(code = SDDS_WriteAsciiPage(SDDS_dataset)))
      return 0;
    if (mode & FLUSH_TABLE) {
      SDDS_FreeTableStrings(SDDS_dataset);
      SDDS_dataset->first_row_in_mem = SDDS_CountRowsOfInterest(SDDS_dataset);
      SDDS_dataset->last_row_written = -1;
      SDDS_dataset->n_rows = 0;
    }
    return code;
  }
  if (!(fp = SDDS_dataset->layout.fp)) {
    SDDS_SetError("Unable to update page--file pointer is NULL (SDDS_UpdateAsciiPage)");
    return (0);
  }
  fBuffer = &SDDS_dataset->fBuffer;
  if (!SDDS_FlushBuffer(fp, fBuffer)) {
    SDDS_SetError("Unable to write page--buffer flushing problem (SDDS_UpdateAsciiPage)");
    return 0;
  }
  offset = ftell(fp);

  rows = SDDS_CountRowsOfInterest(SDDS_dataset) + SDDS_dataset->first_row_in_mem;
  if (rows == SDDS_dataset->n_rows_written)
    return (1);
  if (rows < SDDS_dataset->n_rows_written) {
    SDDS_SetError("Unable to update page--new number of rows less than previous number (SDDS_UpdateAsciiPage)");
    return (0);
  }
  if ((!SDDS_dataset->layout.data_mode.fixed_row_count) || (((rows + rows - SDDS_dataset->n_rows_written) / SDDS_dataset->layout.data_mode.fixed_row_increment) != (rows / SDDS_dataset->layout.data_mode.fixed_row_increment))) {
    if (!SDDS_dataset->layout.data_mode.no_row_counts) {
      if (SDDS_fseek(fp, SDDS_dataset->rowcount_offset, SEEK_SET) == -1) {
        SDDS_SetError("Unable to update page--failure doing fseek (SDDS_UpdateAsciiPage)");
        return (0);
      }
      /* overwrite the existing row count */
      if (SDDS_dataset->layout.data_mode.fixed_row_count) {
        if ((rows - SDDS_dataset->n_rows_written) + 1 > SDDS_dataset->layout.data_mode.fixed_row_increment) {
          SDDS_dataset->layout.data_mode.fixed_row_increment = (rows - SDDS_dataset->n_rows_written) + 1;
        }
        fprintf(fp, "%20" PRId64 "\n", ((rows / SDDS_dataset->layout.data_mode.fixed_row_increment) + 2) * SDDS_dataset->layout.data_mode.fixed_row_increment);
      } else
        fprintf(fp, "%20" PRId64 "\n", rows);
      if (SDDS_fseek(fp, offset, SEEK_SET) == -1) {
        SDDS_SetError("Unable to update page--failure doing fseek to end of page (SDDS_UpdateAsciiPage)");
        return (0);
      }
    }
  }
  for (i = SDDS_dataset->last_row_written + 1; i < SDDS_dataset->n_rows; i++)
    if (SDDS_dataset->row_flag[i])
      SDDS_WriteAsciiRow(SDDS_dataset, i, fp);
  if (!SDDS_FlushBuffer(fp, fBuffer)) {
    SDDS_SetError("Unable to write page--buffer flushing problem (SDDS_UpdateAsciiPage)");
    return 0;
  }
  SDDS_dataset->last_row_written = SDDS_dataset->n_rows - 1;
  SDDS_dataset->n_rows_written = rows;
  if (mode & FLUSH_TABLE) {
    SDDS_FreeTableStrings(SDDS_dataset);
    SDDS_dataset->first_row_in_mem = rows;
    SDDS_dataset->last_row_written = -1;
    SDDS_dataset->n_rows = 0;
  }
  return (1);
}
