/*************************************************************************\
 * @file SDDS_output.c
 * @brief This file contains the implementation of the SDDS_output module.
 *
 * The SDDS_output module provides functions for outputting data in the
 * Self-Describing Data Sets (SDDS) format. It includes functions for
 * creating and writing SDDS files, as well as functions for defining
 * and appending data to the SDDS files.
 *
 * @copyright Copyright (c) 2002 The University of Chicago, as Operator of Argonne
 *            National Laboratory. Copyright (c) 2002 The Regents of the University
 *            of California, as Operator of Los Alamos National Laboratory.
 * @license This file is distributed subject to a Software License Agreement found
 *          in the file LICENSE that is included with this distribution.
 *
 * @author M. Borland, C. Saunders, R. Soliday, H. Shang
\*************************************************************************/

#include "SDDS.h"
#include "SDDS_internal.h"
#include "mdb.h"
#include <ctype.h>

#if defined(_WIN32)
#  include <fcntl.h>
#  include <io.h>
#  if defined(__BORLANDC__)
#    define _setmode(handle, amode) setmode(handle, amode)
#  endif
#else
#  include <unistd.h>
#endif

#if SDDS_VERSION != 5
#  error "SDDS_VERSION does not match version of this file"
#endif

#undef DEBUG

/* Allows "temporarily" closing a file.  Use SDDS_ReconnectFile() to open it
 * again in the same position.  Updates the present page and flushes the
 * table.
 */

#if SDDS_MPI_IO
int32_t SDDS_MPI_DisconnectFile(SDDS_DATASET *SDDS_dataset);
int32_t SDDS_MPI_ReconnectFile(SDDS_DATASET *SDDS_dataset);
#endif

/**
 * @brief Disconnects the SDDS dataset from the file.
 *
 * This function disconnects the SDDS dataset from the file it is currently associated with.
 *
 * @param SDDS_dataset The pointer to the SDDS dataset structure.
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_DisconnectFile(SDDS_DATASET *SDDS_dataset) {
#if SDDS_MPI_IO
  if (SDDS_dataset->parallel_io)
    return SDDS_MPI_DisconnectFile(SDDS_dataset);
#endif
  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_DisconnectFile"))
    return 0;
  if (!SDDS_dataset->layout.filename) {
    SDDS_SetError("Can't disconnect file. No filename given. (SDDS_DisconnectFile)");
    return 0;
  }
  if (SDDS_dataset->layout.gzipFile) {
    SDDS_SetError("Can't disconnect file because it is a gzip file. (SDDS_DisconnectFile)");
    return 0;
  }
  if (SDDS_dataset->layout.lzmaFile) {
    SDDS_SetError("Can't disconnect file because it is a lzma or xz file. (SDDS_DisconnectFile)");
    return 0;
  }
  if (SDDS_dataset->layout.disconnected) {
    SDDS_SetError("Can't disconnect file.  Already disconnected. (SDDS_DisconnectFile)");
    return 0;
  }
  if (SDDS_dataset->page_started && !SDDS_UpdatePage(SDDS_dataset, FLUSH_TABLE)) {
    SDDS_SetError("Can't disconnect file.  Problem updating page. (SDDS_DisconnectFile)");
    return 0;
  }
  if (fclose(SDDS_dataset->layout.fp)) {
    SDDS_SetError("Can't disconnect file.  Problem closing file. (SDDS_DisconnectFile)");
    return 0;
  }
  SDDS_dataset->layout.disconnected = 1;
  return 1;
}

/**
 * @brief Reconnects the SDDS dataset file.
 *
 * This function is used to reconnect the SDDS dataset file after it has been closed.
 *
 * @param SDDS_dataset Pointer to the SDDS_DATASET structure.
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_ReconnectFile(SDDS_DATASET *SDDS_dataset) {
#if SDDS_MPI_IO
  if (SDDS_dataset->parallel_io)
    return SDDS_MPI_ReconnectFile(SDDS_dataset);
#endif
  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_ReconnectFile"))
    return 0;
  if (!SDDS_dataset->layout.disconnected || !SDDS_dataset->layout.filename) {
    SDDS_SetError("Can't reconnect file.  Not disconnected or missing filename. (SDDS_ReconnectFile)");
    return 0;
  }
  if (!(SDDS_dataset->layout.fp = fopen(SDDS_dataset->layout.filename, FOPEN_READ_AND_WRITE_MODE))) {
    char s[1024];
    sprintf(s, "Unable to open file %s (SDDS_ReconnectFile)", SDDS_dataset->layout.filename);
    SDDS_SetError(s);
    return 0;
  }
  if (fseek(SDDS_dataset->layout.fp, 0, 2) == -1) {
    SDDS_SetError("Can't reconnect file.  Fseek failed. (SDDS_ReconnectFile)");
    return 0;
  }
  SDDS_dataset->original_layout.fp = SDDS_dataset->layout.fp;
  SDDS_dataset->layout.disconnected = 0;
  return 1;
}

/**
 * @brief Disconnects the input file associated with the SDDS dataset.
 *
 * This function disconnects the input file associated with the SDDS dataset,
 * allowing it to be closed and released. After calling this function, the
 * SDDS dataset will no longer be able to read data from the input file.
 *
 * @param SDDS_dataset The SDDS dataset structure.
 *
 * @return Returns the file position on success,
 *         -1 on failure
 */
long SDDS_DisconnectInputFile(SDDS_DATASET *SDDS_dataset) {
  long position;
#if SDDS_MPI_IO
  if (SDDS_dataset->parallel_io) {
    SDDS_SetError("Error: MPI mode not supported yet in SDDS_DisconnectInputFile");
    return -1;
  }
#endif
  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_DisconnectInputFile"))
    return -1;
  if (!SDDS_dataset->layout.filename) {
    SDDS_SetError("Can't disconnect file. No filename given. (SDDS_DisconnectInputFile)");
    return -1;
  }
  if (SDDS_dataset->layout.gzipFile) {
    SDDS_SetError("Can't disconnect file because it is a gzip file. (SDDS_DisconnectInputFile)");
    return -1;
  }
  if (SDDS_dataset->layout.lzmaFile) {
    SDDS_SetError("Can't disconnect file because it is a lzma or xz file. (SDDS_DisconnectInputFile)");
    return -1;
  }
  if (SDDS_dataset->layout.disconnected) {
    SDDS_SetError("Can't disconnect file.  Already disconnected. (SDDS_DisconnectInputFile)");
    return -1;
  }
  position = ftell(SDDS_dataset->layout.fp);
  if (fclose(SDDS_dataset->layout.fp)) {
    SDDS_SetError("Can't disconnect file.  Problem closing file. (SDDS_DisconnectInputFile)");
    return -1;
  }
  SDDS_dataset->layout.disconnected = 1;
  return position;
}

/**
 * @brief Reconnects the input file for the SDDS dataset at the specified position.
 *
 * This function allows the user to reconnect the input file for the SDDS dataset at a specific position.
 *
 * @param SDDS_dataset A pointer to the SDDS_DATASET structure representing the dataset.
 * @param position The position of the input file to reconnect.
 *
 * @return Returns 1 on success, 0 on failure.
 */
int32_t SDDS_ReconnectInputFile(SDDS_DATASET *SDDS_dataset, long position) {
#if SDDS_MPI_IO
  if (SDDS_dataset->parallel_io) {
    SDDS_SetError("Error: MPI mode not supported yet in SDDS_ReconnectInputFile");
    return 0;
  }
#endif
  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_ReconnectInputFile"))
    return 0;
  if (!SDDS_dataset->layout.disconnected || !SDDS_dataset->layout.filename) {
    SDDS_SetError("Can't reconnect file.  Not disconnected or missing filename. (SDDS_ReconnectInputFile)");
    return 0;
  }
  if (!(SDDS_dataset->layout.fp = fopen(SDDS_dataset->layout.filename, FOPEN_READ_MODE))) {
    char s[1024];
    sprintf(s, "Unable to open file %s (SDDS_ReconnectInputFile)", SDDS_dataset->layout.filename);
    SDDS_SetError(s);
    return 0;
  }
  if (fseek(SDDS_dataset->layout.fp, position, SEEK_SET) == -1) {
    SDDS_SetError("Can't reconnect file.  Fseek failed. (SDDS_ReconnectInputFile)");
    return 0;
  }
  SDDS_dataset->original_layout.fp = SDDS_dataset->layout.fp;
  SDDS_dataset->layout.disconnected = 0;
  return 1;
}

/* appends to a file by adding a new page */

/**
 * @brief Initializes the SDDS dataset for appending data to an existing file by adding a new page.
 *
 * This function initializes the SDDS dataset structure for appending data to an existing file by adding a new page.
 *
 * @param SDDS_dataset A pointer to the SDDS_DATASET structure.
 * @param filename The name of the file to which data will be appended.
 *
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_InitializeAppend(SDDS_DATASET *SDDS_dataset, const char *filename) {
  /*  char *ptr, *datafile, *headerfile; */
  char s[SDDS_MAXLINE];
  int64_t endOfLayoutOffset, endOfFileOffset;
  char *extension;

  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_InitializeAppend"))
    return 0;
  if (!SDDS_ZeroMemory((void *)SDDS_dataset, sizeof(SDDS_DATASET))) {
    sprintf(s, "Unable to initialize input for file %s--can't zero SDDS_DATASET structure (SDDS_InitializeAppend)", filename);
    SDDS_SetError(s);
    return 0;
  }
  SDDS_dataset->layout.popenUsed = SDDS_dataset->layout.gzipFile = SDDS_dataset->layout.lzmaFile = SDDS_dataset->layout.disconnected = 0;
  SDDS_dataset->layout.depth = SDDS_dataset->layout.data_command_seen = SDDS_dataset->layout.commentFlags = SDDS_dataset->deferSavingLayout = 0;
  if (!filename)
    SDDS_dataset->layout.filename = NULL;
  else if (!SDDS_CopyString(&SDDS_dataset->layout.filename, filename)) {
    sprintf(s, "Memory allocation failure initializing file %s (SDDS_InitializeAppend)", filename);
    SDDS_SetError(s);
    return 0;
  } else if ((extension = strrchr(filename, '.')) && ((strcmp(extension, ".gz") == 0) || (strcmp(extension, ".lzma") == 0) || (strcmp(extension, ".xz") == 0))) {
    sprintf(s, "Cannot append to a compressed file %s (SDDS_InitializeAppend)", filename);
    SDDS_SetError(s);
    return 0;
  }

  SDDS_dataset->layout.popenUsed = 0;
  if (!filename) {
#if defined(_WIN32)
    if (_setmode(_fileno(stdin), _O_BINARY) == -1) {
      sprintf(s, "unable to set stdin to binary mode");
      SDDS_SetError(s);
      return 0;
    }
#endif
    SDDS_dataset->layout.fp = stdin;
  } else {
    if (SDDS_FileIsLocked(filename)) {
      sprintf(s, "unable to open file %s for appending--file is locked (SDDS_InitializeAppend)", filename);
      SDDS_SetError(s);
      return 0;
    }
    if (!(SDDS_dataset->layout.fp = fopen(filename, FOPEN_READ_AND_WRITE_MODE))) {
      sprintf(s, "Unable to open file %s for appending (SDDS_InitializeAppend)", filename);
      SDDS_SetError(s);
      return 0;
    }
    if (!SDDS_LockFile(SDDS_dataset->layout.fp, filename, "SDDS_InitializeAppend"))
      return 0;
  }

  if (!SDDS_ReadLayout(SDDS_dataset, SDDS_dataset->layout.fp))
    return 0;
  endOfLayoutOffset = ftell(SDDS_dataset->layout.fp);
  if (SDDS_dataset->layout.n_columns &&
      (!(SDDS_dataset->column_flag = (int32_t *)SDDS_Malloc(sizeof(int32_t) * SDDS_dataset->layout.n_columns)) ||
       !(SDDS_dataset->column_order = (int32_t *)SDDS_Malloc(sizeof(int32_t) * SDDS_dataset->layout.n_columns)) ||
       !SDDS_SetMemory(SDDS_dataset->column_flag, SDDS_dataset->layout.n_columns, SDDS_LONG, (int32_t)1, (int32_t)0) ||
       !SDDS_SetMemory(SDDS_dataset->column_order, SDDS_dataset->layout.n_columns, SDDS_LONG, (int32_t)0, (int32_t)1))) {
    SDDS_SetError("Unable to initialize input--memory allocation failure (SDDS_InitializeAppend)");
    return 0;
  }
  if (fseek(SDDS_dataset->layout.fp, 0, 2) == -1) {
    SDDS_SetError("Unable to initialize append--seek failure (SDDS_InitializeAppend)");
    return 0;
  }
  endOfFileOffset = ftell(SDDS_dataset->layout.fp);
  if (endOfFileOffset == endOfLayoutOffset)
    SDDS_dataset->file_had_data = 0; /* appending to empty file */
  else
    SDDS_dataset->file_had_data = 1;       /* appending to nonempty file */
  SDDS_dataset->layout.layout_written = 1; /* its already in the file */
  SDDS_dataset->mode = SDDS_WRITEMODE;     /*writing */
  return 1;
}

/**
 * @brief Initializes the SDDS dataset for appending data to the last page in the file.
 *
 * This function initializes the SDDS dataset for appending data to the last page in the file.
 *
 * @param SDDS_dataset A pointer to the SDDS dataset structure.
 * @param filename The name of the file to which the data will be appended.
 * @param updateInterval The update interval for appending data. This is the number of rows that will be written to the file before more memory is allocated.
 * @param rowsPresentReturn A pointer to store the number of rows present in the dataset after initialization.
 *
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_InitializeAppendToPage(SDDS_DATASET *SDDS_dataset, const char *filename, int64_t updateInterval, int64_t *rowsPresentReturn) {
  /*  char *ptr, *datafile, *headerfile; */
  char s[SDDS_MAXLINE];
  int64_t endOfLayoutOffset, endOfFileOffset, rowCountOffset, offset;
  int32_t rowsPresent32;
  int64_t rowsPresent;
  char *extension;
  int32_t previousBufferSize;

  *rowsPresentReturn = -1;
  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_InitializeAppendToPage"))
    return 0;
  if (!SDDS_ZeroMemory((void *)SDDS_dataset, sizeof(SDDS_DATASET))) {
    sprintf(s, "Unable to initialize input for file %s--can't zero SDDS_DATASET structure (SDDS_InitializeAppendToPage)", filename);
    SDDS_SetError(s);
    return 0;
  }
  SDDS_dataset->layout.popenUsed = SDDS_dataset->layout.gzipFile = SDDS_dataset->layout.lzmaFile = SDDS_dataset->layout.disconnected = 0;
  SDDS_dataset->layout.depth = SDDS_dataset->layout.data_command_seen = SDDS_dataset->layout.commentFlags = SDDS_dataset->deferSavingLayout = 0;
  if (!filename)
    SDDS_dataset->layout.filename = NULL;
  else if (!SDDS_CopyString(&SDDS_dataset->layout.filename, filename)) {
    sprintf(s, "Memory allocation failure initializing file %s (SDDS_InitializeAppendToPage)", filename);
    SDDS_SetError(s);
    return 0;
  } else if ((extension = strrchr(filename, '.')) && ((strcmp(extension, ".gz") == 0) || (strcmp(extension, ".lzma") == 0) || (strcmp(extension, ".xz") == 0))) {
    sprintf(s, "Cannot append to a compressed file %s (SDDS_InitializeAppendToPage)", filename);
    SDDS_SetError(s);
    return 0;
  }

  if (!filename) {
#if defined(_WIN32)
    if (_setmode(_fileno(stdin), _O_BINARY) == -1) {
      sprintf(s, "unable to set stdin to binary mode");
      SDDS_SetError(s);
      return 0;
    }
#endif
    SDDS_dataset->layout.fp = stdin;
  } else {
    if (SDDS_FileIsLocked(filename)) {
      sprintf(s, "unable to open file %s for appending--file is locked (SDDS_InitializeAppendToPage)", filename);
      SDDS_SetError(s);
      return 0;
    }
    if (!(SDDS_dataset->layout.fp = fopen(filename, FOPEN_READ_AND_WRITE_MODE))) {
      sprintf(s, "Unable to open file %s for appending (SDDS_InitializeAppendToPage)", filename);
      SDDS_SetError(s);
      return 0;
    }
    if (!SDDS_LockFile(SDDS_dataset->layout.fp, filename, "SDDS_InitializeAppendToPage")) {
      return 0;
    }
  }

  if (!SDDS_ReadLayout(SDDS_dataset, SDDS_dataset->layout.fp)) {
    return 0;
  }
  endOfLayoutOffset = ftell(SDDS_dataset->layout.fp);
  if (SDDS_dataset->layout.n_columns &&
      (!(SDDS_dataset->column_flag = (int32_t *)SDDS_Malloc(sizeof(int32_t) * SDDS_dataset->layout.n_columns)) ||
       !(SDDS_dataset->column_order = (int32_t *)SDDS_Malloc(sizeof(int32_t) * SDDS_dataset->layout.n_columns)) ||
       !SDDS_SetMemory(SDDS_dataset->column_flag, SDDS_dataset->layout.n_columns, SDDS_LONG, (int32_t)1, (int32_t)0) ||
       !SDDS_SetMemory(SDDS_dataset->column_order, SDDS_dataset->layout.n_columns, SDDS_LONG, (int32_t)0, (int32_t)1))) {
    SDDS_SetError("Unable to initialize input--memory allocation failure (SDDS_InitializeAppendToPage)");
    return 0;
  }
  rowCountOffset = -1;
  rowsPresent = 0;
#ifdef DEBUG
  fprintf(stderr, "Data mode is %s\n", SDDS_data_mode[SDDS_dataset->layout.data_mode.mode - 1]);
#endif
  SDDS_dataset->pagecount_offset = NULL;
  previousBufferSize = SDDS_SetDefaultIOBufferSize(0);
  if (!SDDS_dataset->layout.data_mode.no_row_counts) {
    /* read pages to get to the last page */
    while (SDDS_ReadPageSparse(SDDS_dataset, 0, 10000, 0) > 0) {
      rowCountOffset = SDDS_dataset->rowcount_offset;
      offset = ftell(SDDS_dataset->layout.fp);
      fseek(SDDS_dataset->layout.fp, rowCountOffset, 0);

      if (SDDS_dataset->layout.data_mode.mode == SDDS_BINARY) {
        fread(&rowsPresent32, sizeof(rowsPresent32), 1, SDDS_dataset->layout.fp);
        if (SDDS_dataset->swapByteOrder) {
          SDDS_SwapLong(&rowsPresent32);
        }
        if (rowsPresent32 == INT32_MIN) {
          fread(&rowsPresent, sizeof(rowsPresent), 1, SDDS_dataset->layout.fp);
          if (SDDS_dataset->swapByteOrder) {
            SDDS_SwapLong64(&rowsPresent);
          }
        } else {
          rowsPresent = rowsPresent32;
        }
      } else {
        char buffer[30];
        if (!fgets(buffer, 30, SDDS_dataset->layout.fp) || strlen(buffer) != 21 || sscanf(buffer, "%" SCNd64, &rowsPresent) != 1) {
#ifdef DEBUG
          fprintf(stderr, "buffer for row count data: >%s<\n", buffer);
#endif
          SDDS_SetError("Unable to initialize input--row count not present or not correct length (SDDS_InitializeAppendToPage)");
          SDDS_SetDefaultIOBufferSize(previousBufferSize);
          return 0;
        }
      }
      fseek(SDDS_dataset->layout.fp, offset, 0);
#ifdef DEBUG
      fprintf(stderr, "%" PRId64 " rows present\n", rowsPresent);
#endif
    }
    if (rowCountOffset == -1) {
      SDDS_SetDefaultIOBufferSize(previousBufferSize);
      SDDS_SetError("Unable to initialize input--problem finding row count offset (SDDS_InitializeAppendToPage)");
      return 0;
    }
  }
  SDDS_SetDefaultIOBufferSize(previousBufferSize);
  SDDS_dataset->fBuffer.bytesLeft = SDDS_dataset->fBuffer.bufferSize;

#ifdef DEBUG
  fprintf(stderr, "Starting page with %" PRId64 " rows\n", updateInterval);
#endif
  if (!SDDS_StartPage(SDDS_dataset, updateInterval)) {
    SDDS_SetError("Unable to initialize input--problem starting page (SDDS_InitializeAppendToPage)");
    return 0;
  }

  /* seek to the end of the file */
  if (fseek(SDDS_dataset->layout.fp, 0, 2) == -1) {
    SDDS_SetError("Unable to initialize append--seek failure (SDDS_InitializeAppendToPage)");
    return 0;
  }
  endOfFileOffset = ftell(SDDS_dataset->layout.fp);
  if (endOfFileOffset == endOfLayoutOffset)
    SDDS_dataset->file_had_data = 0; /* appending to empty file */
  else {
    SDDS_dataset->file_had_data = 1; /* appending to nonempty file */
    if (rowCountOffset != -1) {
      SDDS_dataset->rowcount_offset = rowCountOffset;
      SDDS_dataset->n_rows_written = rowsPresent;
      SDDS_dataset->first_row_in_mem = rowsPresent;
      SDDS_dataset->last_row_written = -1;
      *rowsPresentReturn = rowsPresent;
      SDDS_dataset->writing_page = 1;
    }
  }
#ifdef DEBUG
  fprintf(stderr, "rowcount_offset = %" PRId64 ", n_rows_written = %" PRId64 ", first_row_in_mem = %" PRId64 ", last_row_written = %" PRId64 "\n", SDDS_dataset->rowcount_offset, SDDS_dataset->n_rows_written, SDDS_dataset->first_row_in_mem, SDDS_dataset->last_row_written);
#endif
  SDDS_dataset->page_number = 1;
  SDDS_dataset->layout.layout_written = 1; /* its already in the file */
  SDDS_dataset->mode = SDDS_WRITEMODE;     /*writing */
  return 1;
}

/**
 * @brief Initializes the SDDS output dataset.
 *
 * This function initializes the SDDS output dataset with the specified parameters.
 *
 * @param SDDS_dataset Pointer to the SDDS_DATASET structure.
 * @param data_mode The data mode for the output dataset. SDDS_ASCII, SDDS_BINARY or SDDS_PARALLEL.
 * @param lines_per_row The number of lines per row in the output dataset. Used only for ASCII output and normally set to 1.
 * @param description The description of the output dataset. NULL if no description is desired.
 * @param contents The contents of the output dataset. NULL if no contents are desired.
 * @param filename The filename of the output dataset.
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_InitializeOutput(SDDS_DATASET *SDDS_dataset, int32_t data_mode, int32_t lines_per_row, const char *description, const char *contents, const char *filename) {
  char s[SDDS_MAXLINE];
  char *extension;
  char *outputEndianess = NULL;

  if (data_mode == SDDS_PARALLEL)
    return SDDS_Parallel_InitializeOutput(SDDS_dataset, description, contents, filename);

  if (sizeof(gzFile) != sizeof(void *)) {
    SDDS_SetError("gzFile is not the same size as void *, possible corruption of the SDDS_LAYOUT structure");
    return (0);
  }
  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_InitializeOutput"))
    return 0;
  if (!SDDS_ZeroMemory((void *)SDDS_dataset, sizeof(SDDS_DATASET))) {
    sprintf(s, "Unable to initialize output for file %s--can't zero SDDS_DATASET structure (SDDS_InitializeOutput)", filename);
    SDDS_SetError(s);
    return 0;
  }
  SDDS_dataset->layout.popenUsed = SDDS_dataset->layout.gzipFile = SDDS_dataset->layout.lzmaFile = SDDS_dataset->layout.disconnected = 0;
  SDDS_dataset->layout.depth = SDDS_dataset->layout.data_command_seen = SDDS_dataset->layout.commentFlags = SDDS_dataset->deferSavingLayout = 0;
  if (!filename) {
#if defined(_WIN32)
    if (_setmode(_fileno(stdout), _O_BINARY) == -1) {
      sprintf(s, "unable to set stdout to binary mode");
      SDDS_SetError(s);
      return 0;
    }
#endif
    SDDS_dataset->layout.fp = stdout;
  } else {
    if (SDDS_FileIsLocked(filename)) {
      sprintf(s, "unable to open file %s for writing--file is locked (SDDS_InitializeOutput)", filename);
      SDDS_SetError(s);
      return 0;
    }
    if ((extension = strrchr(filename, '.')) && ((strcmp(extension, ".xz") == 0) || (strcmp(extension, ".lzma") == 0))) {
      SDDS_dataset->layout.lzmaFile = 1;
      data_mode = SDDS_BINARY; /* force binary mode for output lzma files. The reading of ascii lzma files is flaky because of the lzma_gets command, plus the output files will be much smaller */
      if (!(SDDS_dataset->layout.lzmafp = lzma_open(filename, FOPEN_WRITE_MODE))) {
        sprintf(s, "Unable to open file %s for writing (SDDS_InitializeOutput)", filename);
        SDDS_SetError(s);
        return 0;
      }
      SDDS_dataset->layout.fp = SDDS_dataset->layout.lzmafp->fp;
    } else {
      if (!(SDDS_dataset->layout.fp = fopen(filename, FOPEN_WRITE_MODE))) {
        sprintf(s, "Unable to open file %s for writing (SDDS_InitializeOutput)", filename);
        SDDS_SetError(s);
        return 0;
      }
    }
    if (!SDDS_LockFile(SDDS_dataset->layout.fp, filename, "SDDS_InitializeOutput"))
      return 0;
#if defined(zLib)
    if ((extension = strrchr(filename, '.')) && (strcmp(extension, ".gz") == 0)) {
      SDDS_dataset->layout.gzipFile = 1;
      if ((SDDS_dataset->layout.gzfp = gzdopen(fileno(SDDS_dataset->layout.fp), FOPEN_WRITE_MODE)) == NULL) {
        sprintf(s, "Unable to open compressed file %s for writing (SDDS_InitializeOutput)", filename);
        SDDS_SetError(s);
        return 0;
      }
    }
#endif
  }
  SDDS_dataset->page_number = SDDS_dataset->page_started = 0;
  SDDS_dataset->file_had_data = SDDS_dataset->layout.layout_written = 0;
  if (!filename)
    SDDS_dataset->layout.filename = NULL;
  else if (!SDDS_CopyString(&SDDS_dataset->layout.filename, filename)) {
    sprintf(s, "Memory allocation failure initializing file %s (SDDS_InitializeOutput)", filename);
    SDDS_SetError(s);
    return 0;
  }
  if ((outputEndianess = getenv("SDDS_OUTPUT_ENDIANESS"))) {
    if (strncmp(outputEndianess, "big", 3) == 0)
      SDDS_dataset->layout.byteOrderDeclared = SDDS_BIGENDIAN;
    else if (strncmp(outputEndianess, "little", 6) == 0)
      SDDS_dataset->layout.byteOrderDeclared = SDDS_LITTLEENDIAN;
  } else {
    SDDS_dataset->layout.byteOrderDeclared = SDDS_IsBigEndianMachine() ? SDDS_BIGENDIAN : SDDS_LITTLEENDIAN;
  }

  if (data_mode < 0 || data_mode > SDDS_NUM_DATA_MODES) {
    sprintf(s, "Invalid data mode for file %s (SDDS_InitializeOutput)", filename ? filename : "stdout");
    SDDS_SetError(s);
    return 0;
  }
  if (data_mode == SDDS_ASCII && lines_per_row <= 0) {
    sprintf(s, "Invalid number of lines per row for file %s (SDDS_InitializeOutput)", filename ? filename : "stdout");
    SDDS_SetError(s);
    return 0;
  }
  SDDS_dataset->layout.version = SDDS_VERSION;
  SDDS_dataset->layout.data_mode.mode = data_mode;
  SDDS_dataset->layout.data_mode.lines_per_row = lines_per_row;
  SDDS_dataset->layout.data_mode.no_row_counts = 0;
  SDDS_dataset->layout.data_mode.fixed_row_count = 0;
  SDDS_dataset->layout.data_mode.fsync_data = 0;
  SDDS_dataset->layout.data_mode.column_memory_mode = DEFAULT_COLUMN_MEMORY_MODE;
  /*This is only temporary, soon the default will be column major order */
  SDDS_dataset->layout.data_mode.column_major = 0;
  if (description && !SDDS_CopyString(&SDDS_dataset->layout.description, description)) {
    sprintf(s, "Memory allocation failure initializing file %s (SDDS_InitializeOutput)", filename ? filename : "stdout");
    SDDS_SetError(s);
    return 0;
  }
  if (contents && !SDDS_CopyString(&SDDS_dataset->layout.contents, contents)) {
    sprintf(s, "Memory allocation failure initializing file %s (SDDS_InitializeOutput)", filename ? filename : "stdout");
    SDDS_SetError(s);
    return 0;
  }
  SDDS_dataset->mode = SDDS_WRITEMODE; /*writing */
  SDDS_dataset->pagecount_offset = NULL;
  SDDS_dataset->parallel_io = 0;
  return (1);
}

/**
 * @brief Initializes the SDDS output for parallel processing.
 *
 * This function initializes the SDDS dataset for parallel output. It sets the description, contents, and filename of the dataset.
 * Do not use with compessed files. The output will always be binary.
 *
 * @param SDDS_dataset A pointer to the SDDS_DATASET structure.
 * @param description A string containing the description of the dataset. NULL if no description is desired.
 * @param contents A string containing the contents of the dataset. NULL if no contents are desired.
 * @param filename A string containing the filename of the dataset.
 *
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_Parallel_InitializeOutput(SDDS_DATASET *SDDS_dataset, const char *description, const char *contents, const char *filename) {
  /*  SDDS_DATASET *SDDS_dataset; */
  char s[SDDS_MAXLINE];
  char *outputEndianess = NULL;

  /* SDDS_dataset = &(MPI_dataset->sdds_dataset); */
  if (sizeof(gzFile) != sizeof(void *)) {
    SDDS_SetError("gzFile is not the same size as void *, possible corruption of the SDDS_LAYOUT structure");
    return (0);
  }
  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_InitializeOutput"))
    return 0;
  /* if (!SDDS_ZeroMemory((void *)SDDS_dataset, sizeof(SDDS_DATASET))) {
     sprintf(s,
     "Unable to initialize output for file %s--can't zero SDDS_DATASET structure (SDDS_InitializeOutput)",
     filename);
     SDDS_SetError(s);
     return 0;
     } */
  /*the sdds dataset memory has been zeroed in the SDDS_MPI_Setup */
  SDDS_dataset->layout.popenUsed = SDDS_dataset->layout.gzipFile = SDDS_dataset->layout.lzmaFile = SDDS_dataset->layout.disconnected = 0;
  SDDS_dataset->layout.depth = SDDS_dataset->layout.data_command_seen = SDDS_dataset->layout.commentFlags = SDDS_dataset->deferSavingLayout = 0;
  SDDS_dataset->layout.fp = NULL;

  SDDS_dataset->page_number = SDDS_dataset->page_started = 0;
  SDDS_dataset->file_had_data = SDDS_dataset->layout.layout_written = 0;
  if (!filename)
    SDDS_dataset->layout.filename = NULL;
  else if (!SDDS_CopyString(&SDDS_dataset->layout.filename, filename)) {
    sprintf(s, "Memory allocation failure initializing file %s (SDDS_InitializeOutput)", filename);
    SDDS_SetError(s);
    return 0;
  }
  if ((outputEndianess = getenv("SDDS_OUTPUT_ENDIANESS"))) {
    if (strncmp(outputEndianess, "big", 3) == 0)
      SDDS_dataset->layout.byteOrderDeclared = SDDS_BIGENDIAN;
    else if (strncmp(outputEndianess, "little", 6) == 0)
      SDDS_dataset->layout.byteOrderDeclared = SDDS_LITTLEENDIAN;
  } else {
    SDDS_dataset->layout.byteOrderDeclared = SDDS_IsBigEndianMachine() ? SDDS_BIGENDIAN : SDDS_LITTLEENDIAN;
  }
  /* set big-endian for binary files, since it is the only type of MPI binary file.
     SDDS_dataset->layout.byteOrderDeclared = SDDS_BIGENDIAN; */
  SDDS_dataset->layout.version = SDDS_VERSION;
  /* it turned out that hard to write ascii file in parallel, fixed it as SDDS_BINARY */
  SDDS_dataset->layout.data_mode.mode = SDDS_BINARY;
  SDDS_dataset->layout.data_mode.lines_per_row = 0;
  SDDS_dataset->layout.data_mode.no_row_counts = 0;
  SDDS_dataset->layout.data_mode.fixed_row_count = 0;
  SDDS_dataset->layout.data_mode.fsync_data = 0;
  SDDS_dataset->layout.data_mode.column_memory_mode = DEFAULT_COLUMN_MEMORY_MODE;
  /*This is only temporary, soon the default will be column major order */
  SDDS_dataset->layout.data_mode.column_major = 0;
  if (description && !SDDS_CopyString(&SDDS_dataset->layout.description, description)) {
    sprintf(s, "Memory allocation failure initializing file %s (SDDS_InitializeOutput)", filename ? filename : "stdout");
    SDDS_SetError(s);
    return 0;
  }
  if (contents && !SDDS_CopyString(&SDDS_dataset->layout.contents, contents)) {
    sprintf(s, "Memory allocation failure initializing file %s (SDDS_InitializeOutput)", filename ? filename : "stdout");
    SDDS_SetError(s);
    return 0;
  }
  SDDS_dataset->layout.n_parameters = SDDS_dataset->layout.n_columns = SDDS_dataset->layout.n_arrays = SDDS_dataset->layout.n_associates = 0;
  SDDS_dataset->mode = SDDS_WRITEMODE; /*writing */
  SDDS_dataset->pagecount_offset = NULL;
  SDDS_dataset->parallel_io = 1;
  return (1);
}

/**
 * @brief Sets the flag indicating whether or not to output row counts in the SDDS dataset.
 *
 * @param SDDS_dataset The pointer to the SDDS dataset.
 * @param value The value to set the flag to. A non-zero value indicates that row counts should not be output, while a zero value indicates that row counts should be output.
 * @return int32_t Returns 1 on success, 0 on failure.
 */
int32_t SDDS_SetNoRowCounts(SDDS_DATASET *SDDS_dataset, int32_t value) {
  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_SetNoRowCounts"))
    return 0;
  if (SDDS_dataset->layout.layout_written) {
    SDDS_SetError("Can't change no_row_counts after writing the layout, or for a file you are reading.");
    return 0;
  }
  SDDS_dataset->layout.data_mode.no_row_counts = value ? 1 : 0;
  return 1;
}

/**
 * Writes the SDDS header describing the layout of the data tables that will follow.
 *
 * @param SDDS_dataset Address of the SDDS_DATASET structure for the data set.
 *
 * @return Returns 1 on success. On failure, returns 0 and records an error message.
 *
 */
int32_t SDDS_WriteLayout(SDDS_DATASET *SDDS_dataset) {
  SDDS_LAYOUT *layout;
#if defined(zLib)
  gzFile gzfp;
#endif
  FILE *fp;
  struct lzmafile *lzmafp;
  int64_t i;
  char *outputEndianess = NULL;

#if SDDS_MPI_IO
  if (SDDS_dataset->parallel_io)
    return SDDS_MPI_WriteLayout(SDDS_dataset);
#endif
  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_WriteLayout"))
    return 0;

  if (!SDDS_SaveLayout(SDDS_dataset))
    return 0;

  layout = &SDDS_dataset->layout;

  if (SDDS_dataset->layout.disconnected) {
    SDDS_SetError("Can't write layout--file is disconnected (SDDS_WriteLayout)");
    return 0;
  }

  if (layout->layout_written) {
    SDDS_SetError("Can't write layout--already written to file (SDDS_WriteLayout)");
    return 0;
  }

  if ((outputEndianess = getenv("SDDS_OUTPUT_ENDIANESS"))) {
    if (strncmp(outputEndianess, "big", 3) == 0)
      layout->byteOrderDeclared = SDDS_BIGENDIAN;
    else if (strncmp(outputEndianess, "little", 6) == 0)
      layout->byteOrderDeclared = SDDS_LITTLEENDIAN;
  }

  if (!layout->byteOrderDeclared)
    layout->byteOrderDeclared = SDDS_IsBigEndianMachine() ? SDDS_BIGENDIAN : SDDS_LITTLEENDIAN;

  layout->version = 1;
  for (i = 0; i < layout->n_parameters; i++) {
    if ((layout->parameter_definition[i].type == SDDS_ULONG) || (layout->parameter_definition[i].type == SDDS_USHORT)) {
      layout->version = 2;
      break;
    }
  }
  for (i = 0; i < layout->n_arrays; i++) {
    if ((layout->array_definition[i].type == SDDS_ULONG) || (layout->array_definition[i].type == SDDS_USHORT)) {
      layout->version = 2;
      break;
    }
  }
  for (i = 0; i < layout->n_columns; i++) {
    if ((layout->column_definition[i].type == SDDS_ULONG) || (layout->column_definition[i].type == SDDS_USHORT)) {
      layout->version = 2;
      break;
    }
  }
  if ((layout->data_mode.column_major) && (layout->data_mode.mode == SDDS_BINARY)) {
    layout->version = 3;
  }
  for (i = 0; i < layout->n_parameters; i++) {
    if (layout->parameter_definition[i].type == SDDS_LONGDOUBLE) {
      layout->version = 4;
      break;
    }
  }
  for (i = 0; i < layout->n_arrays; i++) {
    if (layout->array_definition[i].type == SDDS_LONGDOUBLE) {
      layout->version = 4;
      break;
    }
  }
  for (i = 0; i < layout->n_columns; i++) {
    if (layout->column_definition[i].type == SDDS_LONGDOUBLE) {
      layout->version = 4;
      break;
    }
  }
  if ((LDBL_DIG != 18) && (layout->version == 4)) {
    if (getenv("SDDS_LONGDOUBLE_64BITS") == NULL) {
      SDDS_SetError("Error: Operating system does not support 80bit float variables used by SDDS_LONGDOUBLE (SDDS_WriteLayout)\nSet SDDS_LONGDOUBLE_64BITS environment variable to read old files that used 64bit float variables for SDDS_LONGDOUBLE");
      return 0;
    }
  }
  for (i = 0; i < layout->n_parameters; i++) {
    if ((layout->parameter_definition[i].type == SDDS_ULONG64) || (layout->parameter_definition[i].type == SDDS_LONG64)) {
      layout->version = 5;
      break;
    }
  }
  for (i = 0; i < layout->n_arrays; i++) {
    if ((layout->array_definition[i].type == SDDS_ULONG64) || (layout->array_definition[i].type == SDDS_LONG64)) {
      layout->version = 5;
      break;
    }
  }
  for (i = 0; i < layout->n_columns; i++) {
    if ((layout->column_definition[i].type == SDDS_ULONG64) || (layout->column_definition[i].type == SDDS_LONG64)) {
      layout->version = 5;
      break;
    }
  }

  // force layout version 5 because the row and column indexes are now 64bit long integers
  // layout->version = 5;

#if defined(zLib)
  if (SDDS_dataset->layout.gzipFile) {
    if (!(gzfp = layout->gzfp)) {
      SDDS_SetError("Can't write SDDS layout--file pointer is NULL (SDDS_WriteLayout)");
      return 0;
    }

    /* write out the layout data */
    if (!SDDS_GZipWriteVersion(layout->version, gzfp)) {
      SDDS_SetError("Can't write SDDS layout--error writing version (SDDS_WriteLayout)");
      return 0;
    }
    if (layout->version < 3) {
      if (SDDS_dataset->layout.data_mode.mode == SDDS_BINARY) {
        if (layout->byteOrderDeclared == SDDS_BIGENDIAN)
          gzprintf(gzfp, "!# big-endian\n");
        else
          gzprintf(gzfp, "!# little-endian\n");
      }
      if (SDDS_dataset->layout.data_mode.fixed_row_count) {
        gzprintf(gzfp, "!# fixed-rowcount\n");
      }
    }
    if (!SDDS_GZipWriteDescription(layout->description, layout->contents, gzfp)) {
      SDDS_SetError("Can't write SDDS layout--error writing description (SDDS_WriteLayout)");
      return 0;
    }

    for (i = 0; i < layout->n_parameters; i++)
      if (!SDDS_GZipWriteParameterDefinition(layout->parameter_definition + i, gzfp)) {
        SDDS_SetError("Unable to write layout--error writing parameter definition (SDDS_WriteLayout)");
        return 0;
      }

    for (i = 0; i < layout->n_arrays; i++)
      if (!SDDS_GZipWriteArrayDefinition(layout->array_definition + i, gzfp)) {
        SDDS_SetError("Unable to write layout--error writing array definition (SDDS_WriteLayout)");
        return 0;
      }

    for (i = 0; i < layout->n_columns; i++)
      if (!SDDS_GZipWriteColumnDefinition(layout->column_definition + i, gzfp)) {
        SDDS_SetError("Unable to write layout--error writing column definition (SDDS_WriteLayout)");
        return 0;
      }

#  if RW_ASSOCIATES != 0
    for (i = 0; i < layout->n_associates; i++)
      if (!SDDS_GZipWriteAssociateDefinition(layout->associate_definition + i, gzfp)) {
        SDDS_SetError("Unable to write layout--error writing associated file data (SDDS_WriteLayout)");
        return 0;
      }
#  endif

    if (!SDDS_GZipWriteDataMode(layout, gzfp)) {
      SDDS_SetError("Unable to write layout--error writing data mode (SDDS_WriteLayout)");
      return 0;
    }

    layout->layout_written = 1;
    /*gzflush(gzfp, Z_FULL_FLUSH); */
  } else {
#endif
    if (SDDS_dataset->layout.lzmaFile) {
      if (!(lzmafp = layout->lzmafp)) {
        SDDS_SetError("Can't write SDDS layout--file pointer is NULL (SDDS_WriteLayout)");
        return 0;
      }

      /* write out the layout data */
      if (!SDDS_LZMAWriteVersion(layout->version, lzmafp)) {
        SDDS_SetError("Can't write SDDS layout--error writing version (SDDS_WriteLayout)");
        return 0;
      }
      if (layout->version < 3) {
        if (SDDS_dataset->layout.data_mode.mode == SDDS_BINARY) {
          if (layout->byteOrderDeclared == SDDS_BIGENDIAN)
            lzma_printf(lzmafp, "!# big-endian\n");
          else
            lzma_printf(lzmafp, "!# little-endian\n");
        }
        if (SDDS_dataset->layout.data_mode.fixed_row_count) {
          lzma_printf(lzmafp, "!# fixed-rowcount\n");
        }
      }
      if (!SDDS_LZMAWriteDescription(layout->description, layout->contents, lzmafp)) {
        SDDS_SetError("Can't write SDDS layout--error writing description (SDDS_WriteLayout)");
        return 0;
      }
      for (i = 0; i < layout->n_parameters; i++)
        if (!SDDS_LZMAWriteParameterDefinition(layout->parameter_definition + i, lzmafp)) {
          SDDS_SetError("Unable to write layout--error writing parameter definition (SDDS_WriteLayout)");
          return 0;
        }
      for (i = 0; i < layout->n_arrays; i++)
        if (!SDDS_LZMAWriteArrayDefinition(layout->array_definition + i, lzmafp)) {
          SDDS_SetError("Unable to write layout--error writing array definition (SDDS_WriteLayout)");
          return 0;
        }
      for (i = 0; i < layout->n_columns; i++)
        if (!SDDS_LZMAWriteColumnDefinition(layout->column_definition + i, lzmafp)) {
          SDDS_SetError("Unable to write layout--error writing column definition (SDDS_WriteLayout)");
          return 0;
        }

#if RW_ASSOCIATES != 0
      for (i = 0; i < layout->n_associates; i++)
        if (!SDDS_LZMAWriteAssociateDefinition(layout->associate_definition + i, lzmafp)) {
          SDDS_SetError("Unable to write layout--error writing associated file data (SDDS_WriteLayout)");
          return 0;
        }
#endif

      if (!SDDS_LZMAWriteDataMode(layout, lzmafp)) {
        SDDS_SetError("Unable to write layout--error writing data mode (SDDS_WriteLayout)");
        return 0;
      }

      layout->layout_written = 1;
    } else {

      if (!(fp = layout->fp)) {
        SDDS_SetError("Can't write SDDS layout--file pointer is NULL (SDDS_WriteLayout)");
        return 0;
      }

      /* write out the layout data */
      if (!SDDS_WriteVersion(layout->version, fp)) {
        SDDS_SetError("Can't write SDDS layout--error writing version (SDDS_WriteLayout)");
        return 0;
      }
      if (layout->version < 3) {
        if (SDDS_dataset->layout.data_mode.mode == SDDS_BINARY) {
          if (layout->byteOrderDeclared == SDDS_BIGENDIAN)
            fprintf(fp, "!# big-endian\n");
          else
            fprintf(fp, "!# little-endian\n");
        }
        if (SDDS_dataset->layout.data_mode.fixed_row_count) {
          fprintf(fp, "!# fixed-rowcount\n");
        }
      }
      if (!SDDS_WriteDescription(layout->description, layout->contents, fp)) {
        SDDS_SetError("Can't write SDDS layout--error writing description (SDDS_WriteLayout)");
        return 0;
      }

      for (i = 0; i < layout->n_parameters; i++)
        if (!SDDS_WriteParameterDefinition(layout->parameter_definition + i, fp)) {
          SDDS_SetError("Unable to write layout--error writing parameter definition (SDDS_WriteLayout)");
          return 0;
        }

      for (i = 0; i < layout->n_arrays; i++)
        if (!SDDS_WriteArrayDefinition(layout->array_definition + i, fp)) {
          SDDS_SetError("Unable to write layout--error writing array definition (SDDS_WriteLayout)");
          return 0;
        }

      for (i = 0; i < layout->n_columns; i++)
        if (!SDDS_WriteColumnDefinition(layout->column_definition + i, fp)) {
          SDDS_SetError("Unable to write layout--error writing column definition (SDDS_WriteLayout)");
          return 0;
        }

#if RW_ASSOCIATES != 0
      for (i = 0; i < layout->n_associates; i++)
        if (!SDDS_WriteAssociateDefinition(layout->associate_definition + i, fp)) {
          SDDS_SetError("Unable to write layout--error writing associated file data (SDDS_WriteLayout)");
          return 0;
        }
#endif

      if (!SDDS_WriteDataMode(layout, fp)) {
        SDDS_SetError("Unable to write layout--error writing data mode (SDDS_WriteLayout)");
        return 0;
      }

      layout->layout_written = 1;
      fflush(fp);
    }
#if defined(zLib)
  }
#endif
  if (SDDS_SyncDataSet(SDDS_dataset) != 0)
    return 0;
  return (1);
}

/**
 * Writes out the current data table. Must be preceeded by a call to SDDS_WriteLayout at some point.
 *
 * @param SDDS_dataset Address of the SDDS_DATASET structure for the data set.
 *
 * @return Returns 1 on success. On failure, returns 0 and records an error message.
 *
 */
int32_t SDDS_WritePage(SDDS_DATASET *SDDS_dataset) {
  int32_t result;
#if SDDS_MPI_IO
  if (SDDS_dataset->parallel_io)
    return SDDS_MPI_WritePage(SDDS_dataset);
#endif
  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_WritePage"))
    return 0;
  if (!SDDS_dataset->layout.layout_written) {
    SDDS_SetError("Unable to write page--layout not written (SDDS_WritePage)");
    return 0;
  }
  if (SDDS_dataset->layout.disconnected) {
    SDDS_SetError("Can't write page--file is disconnected (SDDS_WritePage)");
    return 0;
  }
  if (SDDS_dataset->layout.data_mode.mode == SDDS_ASCII)
    result = SDDS_WriteAsciiPage(SDDS_dataset);
  else if (SDDS_dataset->layout.data_mode.mode == SDDS_BINARY)
    result = SDDS_WriteBinaryPage(SDDS_dataset);
  else {
    SDDS_SetError("Unable to write page--unknown data mode (SDDS_WritePage)");
    return 0;
  }
  if (result == 1)
    if (SDDS_SyncDataSet(SDDS_dataset) != 0)
      return 0;
  return (result);
}

/**
 * @brief Updates the current page of the SDDS dataset.
 *
 * This function updates the current page of the SDDS dataset based on the specified mode.
 *
 * @param SDDS_dataset The pointer to the SDDS dataset.
 * @param mode The mode is either FLUSH_TABLE or 0. FLUSH_TABLE indicates that the current page is complete and should be written to disk.
 *
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_UpdatePage(SDDS_DATASET *SDDS_dataset, uint32_t mode) {
  int32_t result;
  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_UpdatePage"))
    return 0;
  if (SDDS_dataset->layout.disconnected) {
    SDDS_SetError("Can't write page--file is disconnected (SDDS_UpdatePage)");
    return 0;
  }
  if (SDDS_dataset->page_started == 0) {
    SDDS_SetError("Can't update page--no page started (SDDS_UpdatePage)");
    return 0;
  }
  if (SDDS_dataset->layout.data_mode.mode == SDDS_ASCII)
    result = SDDS_UpdateAsciiPage(SDDS_dataset, mode);
  else if (SDDS_dataset->layout.data_mode.mode == SDDS_BINARY)
    result = SDDS_UpdateBinaryPage(SDDS_dataset, mode);
  else {
    SDDS_SetError("Unable to update page--unknown data mode (SDDS_UpdatePage)");
    return 0;
  }
  if (result == 1)
    if (SDDS_SyncDataSet(SDDS_dataset) != 0)
      return 0;
  return (result);
}

/**
 * @brief Calls fsync to ensure any buffered data is written to disk. NOT IMPLEMENTED.
 *
 * Calls fsync to ensure any buffered data is written to disk.
 *
 * @param SDDS_dataset The SDDS dataset
 * @return Returns 0 on success, or a negative value on failure.
 */
int32_t SDDS_SyncDataSet(SDDS_DATASET *SDDS_dataset) {
#if defined(vxWorks) || defined(_WIN32) || defined(linux) || defined(__APPLE__)
  return (0);
#else
  if (!(SDDS_dataset->layout.fp)) {
    SDDS_SetError("Unable to sync file--file pointer is NULL (SDDS_SyncDataSet)");
    return (-1);
  }
  if (SDDS_dataset->layout.data_mode.fsync_data == 0)
    return (0);
  if (fsync(fileno(SDDS_dataset->layout.fp)) == 0)
    return (0);
  /*
    SDDS_SetError("Unable to sync file (SDDS_SyncDataSet)");
    return(-1);
  */
  /* This error should not be fatal */
  return (0);
#endif
}

/**
 * Processes a definition of a data parameter.
 *
 * @param SDDS_dataset Address of the SDDS_DATASET structure for the data set.
 * @param name A NULL-terminated character string giving the name of the parameter. Must be supplied.
 * @param symbol A NULL-terminated character string giving the symbol to be used to represent the parameter. If none is desired, pass NULL.
 * @param units A NULL-terminated character string giving the units of the parameter. If none is desired, pass NULL.
 * @param description A NULL-terminated character string giving a description of the parameter. If none is desired, pass NULL.
 * @param format_string A NULL-terminated character string giving a printf format string to be used to print the data for ASCII output. If NULL is passed, a reasonable default will be chosen.
 * @param type An integer value specifying the data type of the parameter. Must be one of the following:\n
 SDDS_LONGDOUBLE, SDDS_DOUBLE, SDDS_FLOAT, SDDS_LONG, SDDS_ULONG, SDDS_SHORT, SDDS_USHORT, SDDS_CHARACTER, or SDDS_STRING
 * @param fixed_value A pointer to a numerical variable giving the permanent value of the parameter.
 *
 * @return On success, returns the index of the newly-defined parameter. Returns -1 on failure and records an error message.
 *
 */
int32_t SDDS_DefineParameter1(SDDS_DATASET *SDDS_dataset, const char *name, const char *symbol, const char *units, const char *description, const char *format_string, int32_t type, void *fixed_value) {
  char buffer[SDDS_MAXLINE];
  if (!SDDS_IsValidName(name, "parameter"))
    return -1;
  if (!fixed_value || type == SDDS_STRING)
    return SDDS_DefineParameter(SDDS_dataset, name, symbol, units, description, format_string, type, fixed_value);
  if (type <= 0 || type > SDDS_NUM_TYPES) {
    SDDS_SetError("Unknown data type (SDDS_DefineParameter1)");
    return (-1);
  }
  buffer[SDDS_MAXLINE - 1] = 0;
  if (!SDDS_SprintTypedValue(fixed_value, 0, type, format_string, buffer, 0) || buffer[SDDS_MAXLINE - 1] != 0) {
    SDDS_SetError("Unable to define fixed value for parameter (SDDS_DefineParameter1)");
    return (-1);
  }
  return SDDS_DefineParameter(SDDS_dataset, name, symbol, units, description, format_string, type, buffer);
}

/**
 * Processes a definition of a data parameter.
 *
 * @param SDDS_dataset Address of the SDDS_DATASET structure for the data set.
 * @param name A NULL-terminated character string giving the name of the parameter. Must be supplied.
 * @param symbol A NULL-terminated character string giving the symbol to be used to represent the parameter. If none is desired, pass NULL.
 * @param units A NULL-terminated character string giving the units of the parameter. If none is desired, pass NULL.
 * @param description A NULL-terminated character string giving a description of the parameter. If none is desired, pass NULL.
 * @param format_string A NULL-terminated character string giving a printf format string to be used to print the data for ASCII output. If NULL is passed, a reasonable default will be chosen.
 * @param type An integer value specifying the data type of the parameter. Must be one of the following:\n
 *             SDDS_LONGDOUBLE, SDDS_DOUBLE, SDDS_FLOAT, SDDS_LONG, SDDS_ULONG, SDDS_SHORT, SDDS_USHORT, SDDS_CHARACTER, or SDDS_STRING
 * @param fixed_value A NULL-terminated character string giving the permanent value of the parameter. For data that is not of type SDDS_STRING, the string should be prepared using, for example, sprintf.
 *
 * @return On success, returns the index of the newly-defined parameter. Returns -1 on failure and records an error message.
 *
 */
int32_t SDDS_DefineParameter(SDDS_DATASET *SDDS_dataset, const char *name, const char *symbol, const char *units, const char *description, const char *format_string, int32_t type, char *fixed_value) {
  SDDS_LAYOUT *layout;
  PARAMETER_DEFINITION *definition;
  char s[SDDS_MAXLINE];
  SORTED_INDEX *new_indexed_parameter;
  int32_t index, duplicate;

  if (!SDDS_IsValidName(name, "parameter"))
    return -1;
  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_DefineParameter"))
    return (-1);
  if (!name) {
    SDDS_SetError("NULL name not allowed for parameter definition");
    return (-1);
  }
  layout = &SDDS_dataset->layout;
  if (!(layout->parameter_definition =
          SDDS_Realloc(layout->parameter_definition, sizeof(*layout->parameter_definition) * (layout->n_parameters + 1))) ||
      !(layout->parameter_index = SDDS_Realloc(layout->parameter_index, sizeof(*layout->parameter_index) * (layout->n_parameters + 1))) || !(new_indexed_parameter = (SORTED_INDEX *)SDDS_Malloc(sizeof(*new_indexed_parameter)))) {
    SDDS_SetError("Memory allocation failure (SDDS_DefineParameter)");
    return (-1);
  }
  if (!SDDS_CopyString(&new_indexed_parameter->name, name))
    return -1;
  index = binaryInsert((void **)layout->parameter_index, layout->n_parameters, new_indexed_parameter, SDDS_CompareIndexedNames, &duplicate);
  if (duplicate) {
    sprintf(s, "Parameter %s already exists (SDDS_DefineParameter)", name);
    SDDS_SetError(s);
    return (-1);
  }
  layout->parameter_index[index]->index = layout->n_parameters;

  if (!SDDS_ZeroMemory(definition = layout->parameter_definition + layout->n_parameters, sizeof(PARAMETER_DEFINITION))) {
    SDDS_SetError("Unable to define parameter--can't zero memory for parameter definition (SDDS_DefineParameter)");
    return (-1);
  }
  definition->name = new_indexed_parameter->name;
  if (symbol && !SDDS_CopyString(&definition->symbol, symbol)) {
    SDDS_SetError("Memory allocation failure (SDDS_DefineParameter)");
    return (-1);
  }
  if (units && !SDDS_CopyString(&definition->units, units)) {
    SDDS_SetError("Memory allocation failure (SDDS_DefineParameter)");
    return (-1);
  }
  if (description && !SDDS_CopyString(&definition->description, description)) {
    SDDS_SetError("Memory allocation failure (SDDS_DefineParameter)");
    return (-1);
  }
  if (type <= 0 || type > SDDS_NUM_TYPES) {
    SDDS_SetError("Unknown data type (SDDS_DefineParameter)");
    return (-1);
  }
  definition->type = type;
  if (format_string) {
    if (!SDDS_VerifyPrintfFormat(format_string, type)) {
      SDDS_SetError("Invalid format string (SDDS_DefineParameter)");
      return (-1);
    }
    if (!SDDS_CopyString(&definition->format_string, format_string)) {
      SDDS_SetError("Memory allocation failure (SDDS_DefineParameter)");
      return (-1);
    }
  }
  if (fixed_value && !SDDS_CopyString(&(definition->fixed_value), fixed_value)) {
    SDDS_SetError("Couldn't copy fixed_value string (SDDS_DefineParameter)");
    return (-1);
  }
  definition->definition_mode = SDDS_NORMAL_DEFINITION;
  if (type == SDDS_STRING)
    definition->memory_number = SDDS_CreateRpnMemory(name, 1);
  else
    definition->memory_number = SDDS_CreateRpnMemory(name, 0);
  layout->n_parameters += 1;
  return (layout->n_parameters - 1);
}

/**
 * Processes a definition of a data array.
 *
 * @param SDDS_dataset Address of the SDDS_DATASET structure for the data set.
 * @param name A NULL-terminated character string giving the name of the array. Must be supplied.
 * @param symbol A NULL-terminated character string giving the symbol to be used to represent the array. If none is desired, pass NULL.
 * @param units A NULL-terminated character string giving the units of the array. If none is desired, pass NULL.
 * @param description A NULL-terminated character string giving a description of the array. If none is desired, pass NULL.
 * @param format_string A NULL-terminated character string giving a printf format string to be used to print the data for ASCII output. If NULL is passed, a reasonable default will be chosen.
 * @param type An integer value specifying the data type of the array. Must be one of the following:\n
 *             SDDS_LONGDOUBLE, SDDS_DOUBLE, SDDS_FLOAT, SDDS_LONG, SDDS_ULONG, SDDS_SHORT, SDDS_USHORT, SDDS_CHARACTER, or SDDS_STRING
 * @param fixed_length An integer value giving the length of the field allotted to the array for ASCII output. Ignored if zero. If negative, the field length is the absolute value, but leading and trailing white-space are eliminated from data of type SDDS_STRING upon readin.
 * @param dimensions An integer value giving the number of dimensions of the array. Must be greater than 0.
 * @param group_name A NULL-terminated character string giving the name of the array group to which the array belongs. This mechanism allows the user to indicate that different arrays are related (e.g., parallel to one another).
 *
 * @return On success, returns the index of the newly-defined array. Returns -1 on failure and records an error message.
 *
 */
int32_t SDDS_DefineArray(SDDS_DATASET *SDDS_dataset, const char *name, const char *symbol, const char *units, const char *description, const char *format_string, int32_t type, int32_t field_length, int32_t dimensions, const char *group_name) {
  SDDS_LAYOUT *layout;
  ARRAY_DEFINITION *definition;
  char s[SDDS_MAXLINE];
  SORTED_INDEX *new_indexed_array;
  int32_t index, duplicate;

  if (!SDDS_IsValidName(name, "array"))
    return -1;
  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_DefineArray"))
    return (-1);
  if (!name) {
    SDDS_SetError("NULL name not allowed for array definition");
    return (-1);
  }
  layout = &SDDS_dataset->layout;
  if (!(layout->array_definition =
          SDDS_Realloc(layout->array_definition, sizeof(*layout->array_definition) * (layout->n_arrays + 1))) ||
      !(layout->array_index = SDDS_Realloc(layout->array_index, sizeof(*layout->array_index) * (layout->n_arrays + 1))) || !(new_indexed_array = (SORTED_INDEX *)SDDS_Malloc(sizeof(*new_indexed_array)))) {
    SDDS_SetError("Memory allocation failure (SDDS_DefineArray)");
    return (-1);
  }

  if (!SDDS_CopyString(&new_indexed_array->name, name))
    return -1;
  index = binaryInsert((void **)layout->array_index, layout->n_arrays, new_indexed_array, SDDS_CompareIndexedNames, &duplicate);
  if (duplicate) {
    sprintf(s, "Array %s already exists (SDDS_DefineArray)", name);
    SDDS_SetError(s);
    return (-1);
  }
  layout->array_index[index]->index = layout->n_arrays;

  if (!SDDS_ZeroMemory(definition = layout->array_definition + layout->n_arrays, sizeof(ARRAY_DEFINITION))) {
    SDDS_SetError("Unable to define array--can't zero memory for array definition (SDDS_DefineArray)");
    return (-1);
  }
  definition->name = new_indexed_array->name;
  if ((symbol && !SDDS_CopyString(&definition->symbol, symbol)) || (units && !SDDS_CopyString(&definition->units, units)) || (description && !SDDS_CopyString(&definition->description, description)) || (group_name && !SDDS_CopyString(&definition->group_name, group_name))) {
    SDDS_SetError("Memory allocation failure (SDDS_DefineArray)");
    return (-1);
  }
  if (type <= 0 || type > SDDS_NUM_TYPES) {
    SDDS_SetError("Unknown data type (SDDS_DefineArray)");
    return (-1);
  }
  definition->type = type;
  if (format_string) {
    if (!SDDS_VerifyPrintfFormat(format_string, type)) {
      SDDS_SetError("Invalid format string (SDDS_DefineArray)");
      return (-1);
    }
    if (!SDDS_CopyString(&definition->format_string, format_string)) {
      SDDS_SetError("Memory allocation failure (SDDS_DefineArray)");
      return (-1);
    }
  }
  if ((definition->field_length = field_length) < 0 && type != SDDS_STRING) {
    SDDS_SetError("Invalid field length (SDDS_DefineArray)");
    return (-1);
  }
  if ((definition->dimensions = dimensions) < 1) {
    SDDS_SetError("Invalid number of dimensions for array (SDDS_DefineArray)");
    return (-1);
  }
  layout->n_arrays += 1;
  return (layout->n_arrays - 1);
}

/**
 * Processes a definition of a data column.
 *
 * @param SDDS_dataset Address of the SDDS_DATASET structure for the data set.
 * @param name A NULL-terminated character string giving the name of the column. Must be supplied.
 * @param symbol: A NULL-terminated character string giving the symbol to be used to represent the column. If none is desired, pass NULL.
 * @param units A NULL-terminated character string giving the units of the column. If none is desired, pass NULL.
 * @param description A NULL-terminated character string giving a description of the column. If none is desired, pass NULL.
 * @param format_string A NULL-terminated character string giving a printf format string to be used to print the data for ASCII output. If NULL is passed, a reasonable default will be chosen.
 * @param type An integer value specifying the data type of the column. Must be one of the following:\n
 SDDS_LONGDOUBLE, SDDS_DOUBLE, SDDS_FLOAT, SDDS_LONG, SDDS_ULONG, SDDS_SHORT, SDDS_USHORT, SDDS_CHARACTER, or SDDS_STRING
 * @param field_length: An integer value given the length of the field allotted to the column for ASCII output. Ignored if zero. If negative, the field length is the absolute value, but leading and trailing white-space are eliminated from data of type SDDS_STRING upon readin.
 *
 * @return On success, returns the index of the newly-defined column. Returns -1 on failure and records an error message.
 *
 */
int32_t SDDS_DefineColumn(SDDS_DATASET *SDDS_dataset, const char *name, const char *symbol, const char *units, const char *description, const char *format_string, int32_t type, int32_t field_length) {
  SDDS_LAYOUT *layout;
  COLUMN_DEFINITION *definition;
  char s[SDDS_MAXLINE];
  SORTED_INDEX *new_indexed_column;
  int32_t index;
  int32_t duplicate;

  if (!SDDS_IsValidName(name, "column"))
    return -1;
  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_DefineColumn"))
    return (-1);
  if (!name) {
    SDDS_SetError("NULL name not allowed for column definition");
    return (-1);
  }
  layout = &SDDS_dataset->layout;
  if (!(layout->column_definition =
          SDDS_Realloc(layout->column_definition, sizeof(*layout->column_definition) * (layout->n_columns + 1))) ||
      !(layout->column_index = SDDS_Realloc(layout->column_index, sizeof(*layout->column_index) * (layout->n_columns + 1))) || !(new_indexed_column = (SORTED_INDEX *)SDDS_Malloc(sizeof(*new_indexed_column)))) {
    SDDS_SetError("Memory allocation failure (SDDS_DefineColumn)");
    return (-1);
  }
  if (!SDDS_CopyString(&new_indexed_column->name, name))
    return -1;
  index = binaryInsert((void **)layout->column_index, layout->n_columns, new_indexed_column, SDDS_CompareIndexedNames, &duplicate);
  if (duplicate) {
    sprintf(s, "Column %s already exists (SDDS_DefineColumn)", name);
    SDDS_SetError(s);
    return (-1);
  }
  layout->column_index[index]->index = layout->n_columns;
  if (!SDDS_ZeroMemory(definition = layout->column_definition + layout->n_columns, sizeof(COLUMN_DEFINITION))) {
    SDDS_SetError("Unable to define column--can't zero memory for column definition (SDDS_DefineColumn)");
    return (-1);
  }
  definition->name = new_indexed_column->name;
  if (symbol && !SDDS_CopyString(&definition->symbol, symbol)) {
    SDDS_SetError("Memory allocation failure (SDDS_DefineColumn)");
    return (-1);
  }
  if (units && !SDDS_CopyString(&definition->units, units)) {
    SDDS_SetError("Memory allocation failure (SDDS_DefineColumn)");
    return (-1);
  }
  if (description && !SDDS_CopyString(&definition->description, description)) {
    SDDS_SetError("Memory allocation failure (SDDS_DefineColumn)");
    return (-1);
  }
  if (type <= 0 || type > SDDS_NUM_TYPES) {
    SDDS_SetError("Unknown data type (SDDS_DefineColumn)");
    return (-1);
  }
  definition->type = type;
  if (format_string) {
    if (!SDDS_VerifyPrintfFormat(format_string, type)) {
      SDDS_SetError("Invalid format string (SDDS_DefineColumn)");
      return (-1);
    }
    if (!SDDS_CopyString(&definition->format_string, format_string)) {
      SDDS_SetError("Memory allocation failure (SDDS_DefineColumn)");
      return (-1);
    }
  }
  if ((definition->field_length = field_length) < 0 && type != SDDS_STRING) {
    SDDS_SetError("Invalid field length (SDDS_DefineColumn)");
    return (-1);
  }

  if (SDDS_dataset->n_rows_allocated) {
    if (!SDDS_dataset->data) {
      SDDS_SetError("data array NULL but rows have been allocated! (SDDS_DefineColumn)");
      return (-1);
    }
    /* data already present--must resize data and parameter memory */
    if (!(SDDS_dataset->data = SDDS_Realloc(SDDS_dataset->data, sizeof(*SDDS_dataset->data) * (layout->n_columns + 1))) || !(SDDS_dataset->data[layout->n_columns] = calloc(SDDS_dataset->n_rows_allocated, SDDS_type_size[type - 1]))) {
      SDDS_SetError("Memory allocation failure (SDDS_DefineColumn)");
      return (-1);
    }
  }

  /* not part of output: */
  definition->definition_mode = SDDS_NORMAL_DEFINITION;
  if (type == SDDS_STRING)
    definition->memory_number = SDDS_CreateRpnMemory(name, 1);
  else {
    definition->memory_number = SDDS_CreateRpnMemory(name, 0);
  }
  sprintf(s, "&%s", name);
  definition->pointer_number = SDDS_CreateRpnArray(s);

  layout->n_columns += 1;
  return (layout->n_columns - 1);
}

/**
 * Processes a definition of a data column.
 *
 * @param SDDS_dataset Address of the SDDS_DATASET structure for the data set.
 * @param name A NULL-terminated character string giving the name of the column.
 * @param unit A NULL-terminated character string giving the units of the column. If none is desired, pass NULL.
 * @param type An integer value specifying the data type of the column. Must be one of the following:\n
 *             SDDS_LONGDOUBLE, SDDS_DOUBLE, SDDS_FLOAT, SDDS_LONG, SDDS_ULONG, SDDS_SHORT, SDDS_USHORT, SDDS_CHARACTER, or SDDS_STRING
 *
 * @return 1 on success. On failure, returns 0 and records an error message.
 *
 */
int32_t SDDS_DefineSimpleColumn(SDDS_DATASET *SDDS_dataset, const char *name, const char *unit, int32_t type) {
  if (SDDS_DefineColumn(SDDS_dataset, name, NULL, unit, NULL, NULL, type, 0) < 0)
    return 0;
  return (1);
}

/**
 * Processes a definition of a data parameter.
 *
 * @param SDDS_dataset Address of the SDDS_DATASET structure for the data set.
 * @param name A NULL-terminated character string giving the name of the parameter.
 * @param unit A NULL-terminated character string giving the units of the parameter. If none is desired, pass NULL.
 * @param type An integer value specifying the data type of the parameter. Must be one of the following:\n
 *             SDDS_LONGDOUBLE, SDDS_DOUBLE, SDDS_FLOAT, SDDS_LONG, SDDS_ULONG, SDDS_SHORT, SDDS_USHORT, SDDS_CHARACTER, or SDDS_STRING
 *
 * @return 1 on success. On failure, returns 0 and records an error message.
 *
 */
int32_t SDDS_DefineSimpleParameter(SDDS_DATASET *SDDS_dataset, const char *name, const char *unit, int32_t type) {
  if (SDDS_DefineParameter(SDDS_dataset, name, NULL, unit, NULL, NULL, type, NULL) < 0)
    return 0;
  return (1);
}

/**
 * Processes multiple definitions of data columns of the same data type.
 *
 * @param SDDS_dataset Address of the SDDS_DATASET structure for the data set.
 * @param number The number of columns to define.
 * @param name An array of NULL-terminated character strings giving the names of the columns.
 * @param unit An array of NULL-terminated character strings giving the units of the columns.
 * @param type An integer value specifying the data type of all the columns. Must be one of the following:\n
 *             SDDS_LONGDOUBLE, SDDS_DOUBLE, SDDS_FLOAT, SDDS_LONG, SDDS_ULONG, SDDS_SHORT, SDDS_USHORT, SDDS_CHARACTER, or SDDS_STRING
 *
 * @return 1 on success. On failure, returns 0 and records an error message.
 *
 */
int32_t SDDS_DefineSimpleColumns(SDDS_DATASET *SDDS_dataset, int32_t number, char **name, char **unit, int32_t type) {
  int32_t i;
  if (!number)
    return (1);
  if (!name)
    return 0;
  for (i = 0; i < number; i++)
    if (SDDS_DefineColumn(SDDS_dataset, name[i], NULL, unit ? unit[i] : NULL, NULL, NULL, type, 0) < 0)
      return 0;
  return (1);
}

/**
 * Processes multiple definitions of data parameters of the same data type.
 *
 * @param SDDS_dataset Address of the SDDS_DATASET structure for the data set.
 * @param number The number of parameters to define.
 * @param name An array of NULL-terminated character strings giving the names of the parameters.
 * @param unit An array of NULL-terminated character strings giving the units of the parameters.
 * @param type An integer value specifying the data type of all the parameters. Must be one of the following:\n
 *             SDDS_LONGDOUBLE, SDDS_DOUBLE, SDDS_FLOAT, SDDS_LONG, SDDS_ULONG, SDDS_SHORT, SDDS_USHORT, SDDS_CHARACTER, or SDDS_STRING
 *
 * @return 1 on success. On failure, returns 0 and records an error message.
 *
 */
int32_t SDDS_DefineSimpleParameters(SDDS_DATASET *SDDS_dataset, int32_t number, char **name, char **unit, int32_t type) {
  int32_t i;
  if (!number)
    return (1);
  if (!name)
    return 0;
  for (i = 0; i < number; i++)
    if (SDDS_DefineParameter(SDDS_dataset, name[i], NULL, unit ? unit[i] : NULL, NULL, NULL, type, NULL) < 0)
      return 0;
  return (1);
}

static uint32_t nameValidityFlags = 0;
int32_t SDDS_SetNameValidityFlags(uint32_t flags) {
  uint32_t oldFlags;
  oldFlags = nameValidityFlags;
  nameValidityFlags = flags;
  return oldFlags;
}

/**
 * @brief Checks if a given name is valid for a specified class.
 *
 * This function checks if the provided name is valid for the specified class.
 *
 * @param name The name to be checked.
 * @param class The class name to which the name should belong. Only used for error reporting.
 * @return Returns 1 if the name is valid for the specified class, 0 otherwise.
 */
int32_t SDDS_IsValidName(const char *name, const char *class) {
  char *ptr;
  int32_t isValid = 1;
  char s[SDDS_MAXLINE];
  static char *validChars = "@:#+%-._$&/[]";
  static char *startChars = ".:";

  if (nameValidityFlags & SDDS_ALLOW_ANY_NAME)
    return 1;
  ptr = (char *)name;
  if (strlen(name) == 0)
    isValid = 0;
  else if (!(nameValidityFlags & SDDS_ALLOW_V15_NAME)) {
    /* post V1.5 allows only alpha and startChars members as first character */
    /* V1.5 allows alpha, digits, and any validChars members */
    if (!(isalpha(*ptr) || strchr(startChars, *ptr)))
      isValid = 0;
  }
  while (isValid && *ptr) {
    if (!(isalnum(*ptr) || strchr(validChars, *ptr)))
      isValid = 0;
    ptr++;
  }
  if (!isValid) {
    sprintf(s, "The following %s name is invalid: >%s<\n(sddsconvert may be used to change the name)\n", class, name);
    SDDS_SetError(s);
    return 0;
  }
  return 1;
}

/**
 * @brief Defines an associate for the SDDS dataset. NOT USED.
 *
 * This function defines an associate for the SDDS dataset with the specified name, filename, path, description, contents, and SDDS flag.
 *
 * @param SDDS_dataset The pointer to the SDDS dataset.
 * @param name The name of the associate.
 * @param filename The filename of the associate.
 * @param path The path of the associate.
 * @param description The description of the associate.
 * @param contents The contents of the associate.
 * @param sdds The SDDS flag indicating the associate type.
 * @return Returns index of SDDS Associate on success, or a negative value on failure.
 */
int32_t SDDS_DefineAssociate(SDDS_DATASET *SDDS_dataset, const char *name, const char *filename, const char *path, const char *description, const char *contents, int32_t sdds) {

#if RW_ASSOCIATES == 0
  return 0;
#else
  SDDS_LAYOUT *layout;
  ASSOCIATE_DEFINITION *definition;
  char s[SDDS_MAXLINE];
  if (!SDDS_IsValidName(name, "associate"))
    return -1;
  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_DefineAssociate"))
    return (-1);
  layout = &SDDS_dataset->layout;
  if (!(layout->associate_definition = SDDS_Realloc(layout->associate_definition, sizeof(*layout->associate_definition) * (layout->n_associates + 1)))) {
    SDDS_SetError("Memory allocation failure (SDDS_DefineAssociate)");
    return (-1);
  }
  if (!name) {
    SDDS_SetError("NULL name not allowed for associate file (SDDS_DefineAssociate)");
    return (-1);
  }
  if (!filename) {
    SDDS_SetError("NULL filename not allowed for associate file (SDDS_DefineAssociate)");
    return (-1);
  }
  if (SDDS_GetAssociateIndex(SDDS_dataset, name) >= 0) {
    sprintf(s, "Associate with name %s already exists (SDDS_DefineAssociate)", name);
    SDDS_SetError(s);
    return (-1);
  }
  if (!SDDS_ZeroMemory(definition = layout->associate_definition + layout->n_associates, sizeof(ASSOCIATE_DEFINITION))) {
    SDDS_SetError("Unable to define associate--can't zero memory for associate (SDDS_DefineAssociate)");
    return (-1);
  }

  if (!SDDS_CopyString(&definition->name, name)) {
    SDDS_SetError("Memory allocation failure (SDDS_DefineAssociate)");
    return (-1);
  }
  if (!SDDS_CopyString(&definition->filename, filename)) {
    SDDS_SetError("Memory allocation failure (SDDS_DefineAssociate)");
    return (-1);
  }
  if (path && !SDDS_CopyString(&definition->path, path)) {
    SDDS_SetError("Memory allocation failure (SDDS_DefineAssociate)");
    return (-1);
  }
  if (contents && !SDDS_CopyString(&definition->contents, contents)) {
    SDDS_SetError("Memory allocation failure (SDDS_DefineAssociate)");
    return (-1);
  }
  if (description && !SDDS_CopyString(&definition->description, description)) {
    SDDS_SetError("Memory allocation failure (SDDS_DefineAssociate)");
    return (-1);
  }
  definition->sdds = sdds;
  layout->n_associates += 1;
  return (layout->n_associates - 1);
#endif
}

/**
 * @brief Erases the data in the SDDS dataset.
 *
 * This function erases all the data in the given SDDS dataset, making it empty.
 *
 * @param SDDS_dataset The pointer to the SDDS dataset.
 * @return 1 on success,
 *         0 on error
 */
int32_t SDDS_EraseData(SDDS_DATASET *SDDS_dataset) {
  SDDS_LAYOUT *layout;
  int64_t i, j;
  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_EraseData"))
    return 0;
  layout = &SDDS_dataset->layout;
  if (SDDS_dataset->data) {
    for (i = 0; i < layout->n_columns; i++) {
      if (!SDDS_dataset->data[i])
        continue;
      if (layout->column_definition[i].type == SDDS_STRING) {
        for (j = 0; j < SDDS_dataset->n_rows; j++) {
          if (((char **)SDDS_dataset->data[i])[j]) {
            free(((char **)SDDS_dataset->data[i])[j]);
            ((char **)SDDS_dataset->data[i])[j] = NULL;
          }
        }
      }
    }
  }
  SDDS_dataset->n_rows = 0;

  if (SDDS_dataset->parameter) {
    for (i = 0; i < layout->n_parameters; i++) {
      if (!SDDS_dataset->parameter[i])
        continue;
      if (layout->parameter_definition[i].type == SDDS_STRING && *(char **)(SDDS_dataset->parameter[i])) {
        free(*(char **)(SDDS_dataset->parameter[i]));
        *(char **)SDDS_dataset->parameter[i] = NULL;
      }
    }
  }

  if (SDDS_dataset->array) {
    for (i = 0; i < layout->n_arrays; i++) {
      if (SDDS_dataset->array[i].definition->type == SDDS_STRING) {
        for (j = 0; j < SDDS_dataset->array[i].elements; j++) {
          if (((char **)SDDS_dataset->array[i].data)[j]) {
            free(((char **)SDDS_dataset->array[i].data)[j]);
            ((char **)SDDS_dataset->array[i].data)[j] = NULL;
          }
        }
      }
    }
  }

  return (1);
}

/**
 * @brief Sets the row count mode for the SDDS dataset.
 *
 * This function allows the user to set the row count mode for the SDDS dataset.
 * The row count mode determines how the row count is handled when writing data to the dataset.
 *
 * @param SDDS_dataset A pointer to the SDDS dataset structure.
 * @param mode The row count mode to be set. Must be one of the following:\n
 *            SDDS_VARIABLEROWCOUNT, SDDS_FIXEDROWCOUNT, or SDDS_NOROWCOUNT
 *
 * @return Returns 1 on success, 0 on failure.
 */
int32_t SDDS_SetRowCountMode(SDDS_DATASET *SDDS_dataset, uint32_t mode) {
  if (!SDDS_CheckDataset(SDDS_dataset, "SDDS_SetRowCountMode"))
    return 0;
  if (SDDS_dataset->layout.layout_written) {
    SDDS_SetError("Can't change row count mode after writing the layout, or for a file you are reading.");
    return 0;
  }
  if (mode & SDDS_VARIABLEROWCOUNT) {
    SDDS_dataset->layout.data_mode.fixed_row_count = 0;
    SDDS_dataset->layout.data_mode.no_row_counts = 0;
  } else if (mode & SDDS_FIXEDROWCOUNT) {
    SDDS_dataset->layout.data_mode.fixed_row_count = 1;
    SDDS_dataset->layout.data_mode.fixed_row_increment = 500;
    SDDS_dataset->layout.data_mode.no_row_counts = 0;
    SDDS_dataset->layout.data_mode.fsync_data = 0;
  } else if (mode & SDDS_NOROWCOUNT) {
    SDDS_dataset->layout.data_mode.fixed_row_count = 0;
    SDDS_dataset->layout.data_mode.no_row_counts = 1;
  } else {
    SDDS_SetError("Invalid row count mode (SDDS_SetRowCountMode).");
    return 0;
  }
  if (!SDDS_SaveLayout(SDDS_dataset))
    return 0;
  return 1;
}

/**
 * @brief Disables the file synchronization for the given SDDS dataset.
 *
 * This function disables the file synchronization for the specified SDDS dataset.
 * File synchronization ensures that file modifications are immediately written to disk.
 * SDDS_SyncDataSet will not call fsync if file synchronization is disabled.
 *
 * @param SDDS_dataset The SDDS dataset for which to disable file synchronization.
 */
void SDDS_DisableFSync(SDDS_DATASET *SDDS_dataset) {
  SDDS_dataset->layout.data_mode.fsync_data = 0;
}

/**
 * Enables file synchronization for the given SDDS dataset.
 *
 * @param SDDS_dataset The SDDS dataset for which file synchronization is to be enabled.
 */
void SDDS_EnableFSync(SDDS_DATASET *SDDS_dataset) {
  SDDS_dataset->layout.data_mode.fsync_data = 1;
}

/**
 * Performs a file synchronization operation on the SDDS dataset. NOT IMPLEMENTED.
 *
 * @param SDDS_dataset The SDDS dataset to perform the file synchronization on.
 * @return Returns 1 success, 0 on falure.
 */
int32_t SDDS_DoFSync(SDDS_DATASET *SDDS_dataset) {
#if defined(vxWorks) || defined(_WIN32) || defined(__APPLE__)
  return 1;
#else
  if (SDDS_dataset && SDDS_dataset->layout.fp)
    return fsync(fileno(SDDS_dataset->layout.fp)) == 0;
  return 0;
#endif
}
