#! /usr/bin/env python3

#/*************************************************************************\
#* Copyright (c) 2021 The University of Chicago, as Operator of Argonne
#* National Laboratory.
#* This file is distributed subject to a Software License Agreement found
#* in the file LICENSE that is included with this distribution. 
#\*************************************************************************/

#/* Program sddsicaoptics.py
# * Calculate beta function, phase advance, coupling from ICA spatial pattern data file (generated by sddsica)
# * 3/15/2021
# * Xiaobiao Huang
# *
# */

#./sddsicaoptics.py SV.A icaBeta.sdds -dataType=xy -icaSFile=SV.S -unwrap -verbose
#./sddsicaoptics.py SV.A icaBeta.sdds -dataType=xy -icaSFile=SV.S -unwrap -verbose -xModeIndex=0,1 -yModeIndex=2,3 -winXTune=0.1,0.25 -winYTune=0.2,0.3


import sys, sdds, time, math
import numpy as np



def parseArguments(args0):
    opts=[]
    args=[]
    flagopts=[]
    for arg in args0:
        if arg[0]=='-':
            if '=' in arg:
                    opts.append(arg.split('='))
            else:
                    flagopts.append(arg)
        else:
            args.append(arg)
    return opts, flagopts, args
    
class icaOPT:
	"""Data structure for optics and coupling data by ICA"""
	def __init__(self,numBPM):
          #initialize data storage variables
          self.numberBPM = numBPM
          self.Index = [i for i in range(numBPM)]
          self.nameBPMs = []
          self.betaScaledMode1x = [] #mode 1 on x-BPM
          self.betaScaledMode2y = [] #mode 2 on y-BPM
          self.avgBetaScaledMode1x = 0;
          self.avgBetaScaledMode2y = 0;
          self.angleMode1x = []
          self.angleMode2y = []
          self.betaScaledMode2xc = [] #mode 2 coupled to x-BPM
          self.betaScaledMode1yc = [] #mode 1 coupled to y-BPM
          self.avgBetaScaledMode2xc = 0;
          self.avgBetaScaledMode1yc = 0;
          self.angleMode2xc = []
          self.angleMode1yc = []
          self.ratioCouple1 = [];  #coupling ratio r1=amp1yc/amp1x
          self.ratioCouple2 = [];  #coupling ratio r2=amp2xc/amp2y
          self.avgCoupleRatio1 = 0;
          self.avgCoupleRatio2 = 0;
          self.nux=0;
          self.nuy=0;
          self.disperScaledX = [] #amplitude of synchrotron mode on x-BPM, (Dispersion)
          self.avgAmplitudeXModeZ = 0;
          self.disperScaledY = [] #amplitude of synchrotron mode on y-BPM, (vertical Dispersion)
          self.rmsDyScaled = 0;
          self.nuz=0;
          self.numberTurns=0; #number of turns
          
	def setNameBPM(self,nameBPM):
		self.nameBPMs = nameBPM
	def setNumberTurns(self,turnNum):
		self.numberTurns=turnNum
	def setModeZ(self,dispx,dispy):
		s1=0
		
		if len(dispx)==self.numberBPM:
			self.disperScaledX=dispx
			for i in range(self.numberBPM):
				s1+=dispx[i]
		else:
			print("Wrong number of BPMs in dispersion")
			return		
		s1 /=self.numberBPM
		#print(len(dispx),len(dispy),s1)
		self.avgAmplitudeXModeZ = s1
		if s1!=0:
			for i in range(self.numberBPM):
				self.disperScaledX[i]=dispx[i]/s1	
			#print(self.disperScaledX)
		if len(dispy)==self.numberBPM:
			self.disperScaledY = dispy
			for i in range(self.numberBPM):
				self.disperScaledY[i]=dispy[i]/s1	
			self.rmsDyScaled = np.std(np.array(self.disperScaledY))		
	
	def setModeInPlaneX(self,ampl1x,angle1x):
		self.betaScaledMode1x=ampl1x
		s1=0
		for i in range(self.numberBPM):
			s1+=ampl1x[i]
		s1 /=self.numberBPM
		self.avgBetaScaledMode1x = s1
		for i in range(self.numberBPM):
			self.betaScaledMode1x[i]=ampl1x[i]/s1
		self.angleMode1x=angle1x
	def setModeInPlaneY(self,ampl2y,angle2y):
		self.betaScaledMode2y=ampl2y
		s2=0
		for i in range(self.numberBPM):
			s2+=ampl2y[i]
		s2 /=self.numberBPM
		self.avgBetaScaledMode2y = s2
		for i in range(self.numberBPM):
			self.betaScaledMode2y[i]=ampl2y[i]/s2
		self.angleMode2y=angle2y
	def setModeInPlane(self,ampl1x,angle1x,ampl2y,angle2y):
		self.betaScaledMode1x=ampl1x
		self.betaScaledMode2y=ampl2y
		s1=0
		s2=0
		for i in range(self.numberBPM):
			s1+=ampl1x[i]
			s2+=ampl2y[i]
		s1 /=self.numberBPM
		s2 /=self.numberBPM
		self.avgBetaScaledMode1x = s1
		self.avgBetaScaledMode2y = s2
		for i in range(self.numberBPM):
			self.betaScaledMode1x[i]=ampl1x[i]/s1
			self.betaScaledMode2y[i]=ampl2y[i]/s2
		self.angleMode1x=angle1x
		self.angleMode2y=angle2y
	def setModeCoupled(self,ampl2xc,angle2xc,ampl1yc,angle1yc):
		self.betaScaledMode2xc=ampl2xc
		self.betaScaledMode1yc=ampl1yc
		s1=0
		s2=0
		for i in range(self.numberBPM):
			s1+=ampl2xc[i]
			s2+=ampl1yc[i]
		s1 /=self.numberBPM
		s2 /=self.numberBPM
		self.avgBetaScaledMode2xc = s1
		self.avgBetaScaledMode1yc = s2
		for i in range(self.numberBPM):
			self.betaScaledMode2xc[i]=ampl2xc[i]/s1
			self.betaScaledMode1yc[i]=ampl1yc[i]/s2
		self.ratioCouple1 = [0]*self.numberBPM
		self.ratioCouple2 = [0]*self.numberBPM
		s1=s2=0
		for i in range(self.numberBPM):	
			self.ratioCouple1[i]=math.sqrt(self.betaScaledMode1yc[i]/self.betaScaledMode1x[i]*self.avgBetaScaledMode1yc/self.avgBetaScaledMode1x)
			self.ratioCouple2[i]=math.sqrt(self.betaScaledMode2xc[i]/self.betaScaledMode2y[i]*self.avgBetaScaledMode2xc/self.avgBetaScaledMode2y)
			s1 += self.ratioCouple1[i]
			s2 += self.ratioCouple2[i]
		self.avgCoupleRatio1 = s1/self.numberBPM
		self.avgCoupleRatio2 = s2/self.numberBPM	

		self.angleMode2xc=angle2xc
		self.angleMode1yc=angle1yc
		
	def unwrapPhase(self):
		if len(self.angleMode1x)>0:
			self.angleMode1x = unwrap(self.angleMode1x);
		if len(self.angleMode2y)>0:
			self.angleMode2y = unwrap(self.angleMode2y);
		if len(self.angleMode2xc)>0:
			self.angleMode2xc = unwrap(self.angleMode2xc);
		if len(self.angleMode1yc)>0:
			self.angleMode1yc = unwrap(self.angleMode1yc);		
	def setTuneXY(self,nux,nuy):
		self.nux=nux
		self.nuy=nuy
	def setTuneZ(self,nuz):
		self.nuz=nuz
	def printData(self):
		for i in range(self.numberBPM):
			print(i,self.betaScaledMode1x[i],self.angleMode1x[i],self.betaScaledMode2y[i],self.angleMode2y[i])

def unwrap(p):
	PI = 3.14159265358979
	pa=np.unwrap(p,PI);
	return pa.tolist()

def interpFFT(x, win=[0,0.5]):
	"""Interpolated FFT
	"""
	afx=np.absolute(np.fft.fft(x))
	N = len(afx)
	i_start = int(np.round(win[0]*N))
	i_end = int(np.round(win[1]*N))
	if i_start<0:
		i_start=0
	if i_end>=N:
		i_end=N-1
	afxw = afx[i_start:i_end+1]
	
	im = np.argmax(afxw)
	#print(N, ',', len(afxw),' ', im)
	if im==0:
		im1=im+1
	elif im==len(afxw)-1:
		im1=im-1
	else:
		if afxw[im+1]>afxw[im-1]:
			im1 = im+1
		else:
			im1=im-1
	
	f = (i_start+im + (im1-im)*afxw[im1]/(afxw[im]+afxw[im1]))/N*1.0
	return f	

def loadSDDSdata(input):
	data=sdds.SDDS(0)
	data.load(input)
	return data

def writeOpticsData(output, mdata):
	x = sdds.SDDS(0)
	x.setDescription("text", "contents")
	paraNames = ["numberBPM","avgBetaScaledMode1x","avgBetaScaledMode2y","avgBetaScaledMode2xc","avgBetaScaledMode1yc","avgCoupleRatio1","avgCoupleRatio2","numberTurns","nux","nuy"]
	types = [x.SDDS_LONG,x.SDDS_DOUBLE,x.SDDS_DOUBLE,x.SDDS_DOUBLE,x.SDDS_DOUBLE,x.SDDS_DOUBLE,x.SDDS_DOUBLE,x.SDDS_LONG,x.SDDS_DOUBLE,x.SDDS_DOUBLE]
	dataVal=[mdata.numberBPM,mdata.avgBetaScaledMode1x,mdata.avgBetaScaledMode2y,mdata.avgBetaScaledMode2xc,mdata.avgBetaScaledMode1yc,mdata.avgCoupleRatio1, mdata.avgCoupleRatio2,mdata.numberTurns,mdata.nux,mdata.nuy]
	if mdata.nuz>0:
		paraNames.append("nuz")
		types.append(x.SDDS_DOUBLE)
		dataVal.append(mdata.nuz)
	if np.fabs(mdata.avgAmplitudeXModeZ)>0:
		paraNames.append("avgAmplitudeXModeZ")
		types.append(x.SDDS_DOUBLE)
		dataVal.append(mdata.avgAmplitudeXModeZ)
	if mdata.rmsDyScaled>0:
		paraNames.append("rmsDyScaled")
		types.append(x.SDDS_DOUBLE)
		dataVal.append(mdata.rmsDyScaled)		
	for i in range(len(paraNames)):
		x.defineSimpleParameter(paraNames[i], types[i])
		x.setParameterValueList(paraNames[i], [dataVal[i]])
	#print(x.parameterData)
	
	#defineColumn(self, name, symbol, units, description, formatString, type, fieldLength):
	#defineSimpleColumn(self, name, type):
	NBPM = len(mdata.Index)
	x.defineSimpleColumn("Index",  x.SDDS_LONG)
	x.setColumnValueList( "Index", mdata.Index,1)
	x.defineSimpleColumn("BPMName",  x.SDDS_STRING)
	x.setColumnValueList( "BPMName", mdata.nameBPMs,1)
	
	if len(mdata.betaScaledMode1x)==NBPM:
		x.defineSimpleColumn("betaScaledMode1x",  x.SDDS_DOUBLE)
		x.setColumnValueList( "betaScaledMode1x", mdata.betaScaledMode1x,1)
		x.defineSimpleColumn("angleMode1x",  x.SDDS_DOUBLE)
		x.setColumnValueList( "angleMode1x", mdata.angleMode1x,1)
	
	if len(mdata.betaScaledMode2y)==NBPM:
		x.defineSimpleColumn("betaScaledMode2y",  x.SDDS_DOUBLE)
		x.setColumnValueList( "betaScaledMode2y", mdata.betaScaledMode2y,1)
		x.defineSimpleColumn("angleMode2y",  x.SDDS_DOUBLE)
		x.setColumnValueList( "angleMode2y", mdata.angleMode2y,1)
	if len(mdata.betaScaledMode2xc)==NBPM:
		x.defineSimpleColumn("betaScaledMode2xc",  x.SDDS_DOUBLE)
		x.setColumnValueList( "betaScaledMode2xc", mdata.betaScaledMode2xc,1)
		x.defineSimpleColumn("angleMode2xc",  x.SDDS_DOUBLE)
		x.setColumnValueList( "angleMode2xc", mdata.angleMode2xc,1)
	if len(mdata.betaScaledMode1yc)==NBPM:
		x.defineSimpleColumn("betaScaledMode1yc",  x.SDDS_DOUBLE)
		x.setColumnValueList( "betaScaledMode1yc", mdata.betaScaledMode1yc,1)
		x.defineSimpleColumn("angleMode1yc",  x.SDDS_DOUBLE)
		x.setColumnValueList( "angleMode1yc", mdata.angleMode1yc,1)
	if len(mdata.ratioCouple1)==NBPM:
		x.defineSimpleColumn("ratioCouple1",  x.SDDS_DOUBLE)
		x.setColumnValueList( "ratioCouple1", mdata.ratioCouple1,1)
	if len(mdata.ratioCouple2)==NBPM:
		x.defineSimpleColumn("ratioCouple2",  x.SDDS_DOUBLE)
		x.setColumnValueList( "ratioCouple2", mdata.ratioCouple2,1)			
	if len(mdata.disperScaledX)==NBPM:
		x.defineSimpleColumn("disperScaledX",  x.SDDS_DOUBLE)
		x.setColumnValueList( "disperScaledX", mdata.disperScaledX,1)
	if len(mdata.disperScaledY)==NBPM:
		x.defineSimpleColumn("disperScaledY",  x.SDDS_DOUBLE)
		x.setColumnValueList( "disperScaledY", mdata.disperScaledY,1)		
				
	#print(len(mdata.Index),len(mdata.nameBPMs),len(mdata.betaScaledMode1x))
	#print(x.columnData)
	#print(x.columnName)
	x.save(output)          

def splitIntegerString(str):
	#print(str.split(','))
	a=str.split(',')
	b=[]
	for i in range(len(a)):
		b.append(int(a[i]))
	return b
def splitFloatString(str):
	#print(str.split(','))
	a=str.split(',')
	b=[]
	for i in range(len(a)):
		b.append(float(a[i]))
	return b	
#-----------------------------------------
#
option = {
  "datatype", "xmodeindex", "ymodeindex","zmodeindex", "unwrap", "pipe", "icasfile","winxtune","winytune"#,"column"
};

#[-pipe=[input][,output]]
USAGE="sddsicaoptics <inputfile> <outputfile> \n\
  -dataType=<x|y|xy> [-xModeIndex=<x1,x2> [-yModeIndex=<y1,y2>][-zModeIndex=num] \n\
  [-unwrap] [-icaSFILE=<filename>] [-verbose] [-winXTune] [-winYTune] [-winZTune]\n\
Program by Xiaobiao Huang. (3/22/2021)"; # ("__DATE__" "__TIME__", SVN revision: "SVN_VERSION")\n";

additional_help1 = "\n\
sddsicaoptics calculates linear optics and coupling parameters from sddsica output \n\n\
-dataType=x|y|xy           x, y, or xy, type of the BPM data, x-only, y-only, or xy stacked up\n\
-xModeIndex=x1,x2         two integers that indicate the pair of x-modes (nux) .\n\
-yModeIndex=y1,y2         two integers that indicate the pair of y-modes (nuy) .\n\
-zModeIndex=number         one integer that indicates the synchrotron mode (nuz) .\n\
-unwrap             if given, unwrap phase advances\n\
-icaSFile=filename           specify ICA s-file (temporal pattern) for additional processing \n\
-winXTune=low_x,high_x  low and high limits for horizontal fractional tune, default to 0, 0.5, respectively \n\
-winXTune=low_x,high_x  low and high limits for vertical fractional tune \n\
-winZTune=low_x,high_x  low and high limits for synchrotron tune \n\
-verbose			print out messages to stderr \n";
#-column=names             specify column names (w/ wildcards) for ICA modes.\n";

	
if len(sys.argv)<2:
	print(USAGE)
	print(additional_help1)
	print("input file not given")
	exit(1)

filename = sys.argv[1]
[opts, flagopts, args]=parseArguments(sys.argv)
if len(args)<3:
	print(args)
	print(USAGE)
	print(additional_help1)
	#print("input or output file not given")
	exit(1)
else:
	inputfile = args[1]
	output = args[2]

xModes=[]
yModes=[]
zMode =[]
icaSFile=''
winXTune=[0,0.5]
winYTune=[0,0.5]
winZTune=[0.001,0.1]
dataType='XY'
for i in range(len(opts)):
	#print(opts[i])
	if opts[i][0]=='-dataType':
		dataType=opts[i][1].upper()
		#print(dataType)
	elif opts[i][0]=='-xModeIndex':
		xModes=splitIntegerString(opts[i][1])
		#print(xModes)
	elif opts[i][0]=='-yModeIndex':
		yModes=splitIntegerString(opts[i][1])
		#print(yModes)
	elif opts[i][0]=='-zModeIndex':
		zMode=splitIntegerString(opts[i][1])
		#print(zModes)
	elif opts[i][0]=='-winXTune':
		winXTune=splitFloatString(opts[i][1])
		#print(winXTune)
	elif opts[i][0]=='-winYTune':
		winYTune=splitFloatString(opts[i][1])
		#print(winYTune)			
	elif opts[i][0]=='-winZTune':
		winZTune=splitFloatString(opts[i][1])
		#print(winZTune)			
	elif opts[i][0]=='-icaSFile':
		icaSFile=opts[i][1]
		#print(icaSFile)	
	else:
		print(opts[i]," is not supported")	
		
flag_unwrap = 0
flag_verbose = 0	
for i in range(len(flagopts)):
	if flagopts[i]=='-unwrap':
		flag_unwrap = 1
		#print(flagopts[i])
	elif flagopts[i]=='-verbose':
		flag_verbose = 1
		#print(flagopts[i])
	else:
		print(flagopts[i]," is not supported")		
		
#print(opts)
#print(flagopts)
#print(args)

#exit(0)


data=loadSDDSdata(inputfile)
if flag_verbose:
	print("Total # of columns in input file: ", len(data.columnName))

if xModes==[]:
	b1=[]
	for i in range(len(data.parameterName)):
		#print(data.parameterName[i],"\t",data.parameterData[i])
		if 'betaPair1A'==data.parameterName[i]:
			b1.append(data.parameterData[i][0]);
		if 'betaPair1B'==data.parameterName[i]:
			b1.append(data.parameterData[i][0]);
else:
	b1=xModes

if yModes==[]:
	b2=[]
	for i in range(len(data.parameterName)):
		#print(data.parameterName[i],"\t",data.parameterData[i])
		if 'betaPair2A'==data.parameterName[i]:
			b2.append(data.parameterData[i][0]);
		if 'betaPair2B'==data.parameterName[i]:
			b2.append(data.parameterData[i][0]);
else:
	b2=yModes

			
if flag_verbose:
	print("Mode 1 (x): ",b1)
	print("Mode 2 (y): ",b2)
	if len(zMode)>0:
		print("Mode Z: ",zMode[0])

#dataType='XY';
if dataType=='XY':
	NBPM=int(len(data.columnData[b1[0]+2][0])/2)
	Amp1=[data.columnData[b1[0]+2][0][i]**2+data.columnData[b1[1]+2][0][i]**2 for i in range(NBPM)]
	Angle1=[math.atan2(data.columnData[b1[0]+2][0][i],data.columnData[b1[1]+2][0][i]) for i in range(NBPM)]
	Amp2=[data.columnData[b2[0]+2][0][i+NBPM]**2+data.columnData[b2[1]+2][0][i+NBPM]**2 for i in range(NBPM)]
	Angle2=[math.atan2(data.columnData[b2[0]+2][0][i+NBPM],data.columnData[b2[1]+2][0][i+NBPM]) for i in range(NBPM)]
else:
	if len(b1)>0:
		NBPM=len(data.columnData[b1[0]+2][0])
		Amp1=[data.columnData[b1[0]+2][0][i]**2+data.columnData[b1[1]+2][0][i]**2 for i in range(NBPM)]
		Angle1=[math.atan2(data.columnData[b1[0]+2][0][i],data.columnData[b1[1]+2][0][i]) for i in range(NBPM)]
	else:
		Amp1=[]
		Angle1=[]	
	if len(b2)>0:
		NBPM=len(data.columnData[b2[0]+2][0])		
		Amp2=[data.columnData[b2[0]+2][0][i]**2+data.columnData[b2[1]+2][0][i]**2 for i in range(NBPM)]
		Angle2=[math.atan2(data.columnData[b2[0]+2][0][i],data.columnData[b2[1]+2][0][i]) for i in range(NBPM)]
	else:
		Amp2=[]
		Angle2=[]	

#mdata is a structure with all data to be written to the output SDDS file
mdata = icaOPT(NBPM);

#iBPMName = [match for match in data.parameterName if "BPMName" in match]
iBPMName = 1
if iBPMName<0:
	print("No BPM column is found")
	
if flag_verbose:
	print(iBPMName, data.columnName[iBPMName]);
mdata.setNameBPM(data.parameterData[iBPMName]) 
mdata.nameBPMs = data.columnData[iBPMName][0][0:NBPM];
#print(mdata.nameBPMs)

#mdata.setModeInPlane(Amp1,Angle1,Amp2,Angle2)
if len(Amp1)>0:
	mdata.setModeInPlaneX(Amp1,Angle1)
if len(Amp2)>0:
	mdata.setModeInPlaneY(Amp2,Angle2)
	
if dataType=='XY':
	Amp1c=[data.columnData[b1[0]+2][0][i+NBPM]**2+data.columnData[b1[1]+2][0][i+NBPM]**2 for i in range(NBPM)]
	Angle1c=[math.atan2(data.columnData[b1[0]+2][0][i+NBPM],data.columnData[b1[1]+2][0][i+NBPM]) for i in range(NBPM)]

	Amp2c=[data.columnData[b2[0]+2][0][i]**2+data.columnData[b2[1]+2][0][i]**2 for i in range(NBPM)]
	Angle2c=[math.atan2(data.columnData[b2[0]+2][0][i],data.columnData[b2[1]+2][0][i]) for i in range(NBPM)]
	mdata.setModeCoupled(Amp1c,Angle1c,Amp2c,Angle2c)
	if flag_verbose:
		print("average coupling parameter r1, r2:", mdata.avgCoupleRatio1, ", ", mdata.avgCoupleRatio2)

if len(zMode)>0:
	if dataType=='XY':
		dispx=[data.columnData[zMode[0]+2][0][i] for i in range(NBPM)]
		dispy=[data.columnData[zMode[0]+2][0][i+NBPM] for i in range(NBPM)]
		mdata.setModeZ(dispx,dispy)
	elif dataType=='X':
		dispx=[data.columnData[zMode[0]+2][0][i] for i in range(NBPM)]
		mdata.setModeZ(dispx,[])
	elif dataType=='Y':
		dispy=[data.columnData[zMode[0]+2][0][i] for i in range(NBPM)]
		mdata.setModeZ([],dispy)	
		
#Read ICA S-file if provided with file name
TOL_TUNE=0.001
if len(icaSFile)>0:
	try:
		sdata = loadSDDSdata(icaSFile)
	except:
		print("cannot process ICA S-file: ", icaSFile)	
	
	numTurn=0		
	if len(b1)>0:
		#print(b1[0]+1)
		s1a = np.array(sdata.columnData[b1[0]+1][0])
		s1b = np.array(sdata.columnData[b1[1]+1][0])
		numTurn=len(s1a)
		nux = interpFFT(s1a,winXTune)
		nuxb = interpFFT(s1b,winXTune)
		if np.fabs(nux-nuxb)>TOL_TUNE:
			print("Warning: tunes in x-mode pair are not matched: ",nux," ",nuxb)
				
			#print(nux)
	if len(b2)>0:
		#print(b2[0]+1)
		s2a = np.array(sdata.columnData[b2[0]+1][0])
		s2b = np.array(sdata.columnData[b2[1]+1][0])
		numTurn=len(s2a)
		nuy = interpFFT(s2a,winYTune)
		nuyb = interpFFT(s2b,winYTune)
		if np.fabs(nuy-nuyb)>TOL_TUNE:
			print("Warning: tunes in y-mode pair are not matched: ",nuy," ",nuyb)
			#print(nuy)	
	mdata.setTuneXY(nux,nuy)
	if flag_verbose:
		print("nux = ",nux, ", nuy = ",nuy)
	
	if len(zMode)>0:
		sz = np.array(sdata.columnData[zMode[0]+1][0])
		numTurn=len(sz)
		nuz = interpFFT(sz,winZTune)
		mdata.setTuneZ(nuz)
		if flag_verbose:
			print("nuz = ",nuz)	
	if numTurn==0:
		print("Error: incorrect number of turns ")
	else:
		mdata.setNumberTurns(numTurn)
		if flag_verbose:
			print("Number of Turns = ",numTurn)	
	
if flag_unwrap:
	mdata.unwrapPhase()

#print(Angle1c)
#print(mdata.angleMode2xc)


#if verbose:
#	mdata.printData()

#output = "icaBeta.sdds"
if flag_verbose:
	print("Writing output SDDS file")
writeOpticsData(output, mdata)
	

