#include <Python.h>
#include "SDDS.h"

#if PY_MAJOR_VERSION >= 3
#define IS_PY3K
#define PyString_Check PyUnicode_Check
#define PyString_AsString PyUnicode_AsUTF8
#define PyString_FromString PyUnicode_FromString
#define PyInt_Check PyLong_Check
#define PyInt_AsLong PyLong_AsLong
#define PyInt_AsUnsignedLong PyLong_AsUnsignedLong
#define PyInt_AsLongLong PyLong_AsLongLong
#define PyInt_AsUnsignedLongLong PyLong_AsUnsignedLongLong
#define PyExc_StandardError PyExc_Exception
#endif
#if PY_MAJOR_VERSION < 3
#define PyInt_AsUnsignedLongLong PyInt_AsUnsignedLongLongMask
#define PyInt_AsUnsignedLong PyInt_AsUnsignedLongMask
#define PyInt_AsLongLong PyInt_AsLong
#endif

SDDS_DATASET dataset_f[20];

/* SDDS_InitializeInput
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_InitializeInput( PyObject* self, PyObject* args ) 
{
  long fileIndex;
  char *filename;
  if (!PyArg_ParseTuple(args, "ls", &fileIndex, &filename)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_InitializeInput(&dataset_f[fileIndex], filename));
}

/* SDDS_InitializeAppend
   Returns: 0 on error
            1 on success */ 
static PyObject* sddsdata_InitializeAppend( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *filename;
  if (!PyArg_ParseTuple(args, "ls", &fileIndex, &filename)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_InitializeAppend(&dataset_f[fileIndex], filename));
}

/* SDDS_InitializeAppendToPage
   Returns: 0 on error
            rows on success */ 
static PyObject* sddsdata_InitializeAppendToPage( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *filename;
  long updateInterval, result;
  int64_t rowsPresent;
  if (!PyArg_ParseTuple(args, "lsl", &fileIndex, &filename, &updateInterval)) {
    return NULL;
  }
  result = SDDS_InitializeAppendToPage(&dataset_f[fileIndex], filename, updateInterval, &rowsPresent);
  if (result != 1)
    return NULL;
  return PyLong_FromLongLong((long long)rowsPresent);
}

/* SDDS_InitializeOutput
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_InitializeOutput( PyObject* self, PyObject* args )
{
  long fileIndex;
  long data_mode;
  long lines_per_row;
  char *description;
  char *contents;
  char *filename;
  if (!PyArg_ParseTuple(args, "lllsss", &fileIndex, &data_mode, &lines_per_row, &description, &contents, &filename)) {
    return NULL;
  }
  if (description) 
    if (strlen(description) == 0) 
      description = NULL;
  if (contents) 
    if (strlen(contents) == 0)
      contents = NULL;
  return PyLong_FromLong(SDDS_InitializeOutput(&dataset_f[fileIndex], data_mode, lines_per_row, description, contents, filename));
}

/* SDDS_SetColumnMajorOrder
   Returns: NULL */
static PyObject* sddsdata_SetColumnMajorOrder( PyObject* self, PyObject* args )
{
  long fileIndex;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  dataset_f[fileIndex].layout.data_mode.column_major = 1;
  return Py_None;
}

/* SDDS_SetFixedRowCountMode
   Returns: NULL */
static PyObject* sddsdata_SetFixedRowCountMode( PyObject* self, PyObject* args )
{
  long fileIndex;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  SDDS_SetRowCountMode(&dataset_f[fileIndex], SDDS_FIXEDROWCOUNT);
  return Py_None;
}

/* SDDS_SetRowMajorOrder
   Returns: NULL */
static PyObject* sddsdata_SetRowMajorOrder( PyObject* self, PyObject* args )
{
  long fileIndex;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  dataset_f[fileIndex].layout.data_mode.column_major = 0;
  return Py_None;
}

/* SDDS_EnableFSync
   Returns: NULL */
static PyObject* sddsdata_EnableFSync( PyObject* self, PyObject* args )
{
  long fileIndex;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  dataset_f[fileIndex].layout.data_mode.fsync_data = 1;
  return Py_None;
}

/* SDDS_DisableFSync
   Returns: NULL */
static PyObject* sddsdata_DisableFSync( PyObject* self, PyObject* args )
{
  long fileIndex;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  dataset_f[fileIndex].layout.data_mode.fsync_data = 0;
  return Py_None;
}

/* SDDS_Terminate
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_Terminate( PyObject* self, PyObject* args )
{
  long fileIndex;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_Terminate(&dataset_f[fileIndex]));
}

/* SDDS_SetTerminateMode
   Sets TerminateMode to:
   TERMINATE_DONT_FREE_ARRAY_STRINGS+TERMINATE_DONT_FREE_TABLE_STRINGS */
static PyObject* sddsdata_SetTerminateMode( PyObject* self, PyObject* args )
{
  SDDS_SetTerminateMode(TERMINATE_DONT_FREE_TABLE_STRINGS+TERMINATE_DONT_FREE_ARRAY_STRINGS);
  Py_INCREF(Py_None);
  return Py_None;
}

/* SDDS_DefineParameter
   Returns: -1 on error
            index on success */
static PyObject* sddsdata_DefineParameter( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *name;
  char *symbol;
  char *units;
  char *description;
  char *format_string;
  long type;
  PyObject *fixed_value;
  char *fixedStringValue;
  int32_t fixedLongValue;
  double fixedDoubleValue;
  char buffer[SDDS_MAXLINE];
  if (!PyArg_ParseTuple(args, "lssssslO", &fileIndex, &name , &symbol , &units , &description , &format_string , &type , &fixed_value)) {
    return NULL;
  }
  if (symbol)
    if (strlen(symbol) == 0) 
      symbol = NULL;
  if (units)
    if (strlen(units) == 0)
      units = NULL;
  if (description)
    if (strlen(description) == 0)
      description = NULL;
  if (format_string)
    if (strlen(format_string) == 0)
      format_string = NULL;

  if (PyString_Check(fixed_value)) {
    fixedStringValue = PyString_AsString(fixed_value);
    if (fixedStringValue)
      if (strlen(fixedStringValue) == 0)
	fixedStringValue = NULL;
    return PyLong_FromLong(SDDS_DefineParameter(&dataset_f[fileIndex], 
						name, symbol, units, 
						description, format_string, 
						type, fixedStringValue));
  } else {
    if (PyNumber_Check(fixed_value)) {
      if (PyLong_Check(fixed_value)) {
	fixedLongValue = (int32_t)PyLong_AsLong(fixed_value);
	SDDS_SprintTypedValue(&fixedLongValue, 0, SDDS_LONG , format_string, buffer, 0);
	return PyLong_FromLong(SDDS_DefineParameter(&dataset_f[fileIndex], 
						    name, symbol, units, 
						    description, format_string, 
						    type, buffer));
      } else if (PyInt_Check(fixed_value)) {
	fixedLongValue = (int32_t)PyInt_AsLong(fixed_value);
	SDDS_SprintTypedValue(&fixedLongValue, 0, SDDS_LONG , format_string, buffer, 0);
	return PyLong_FromLong(SDDS_DefineParameter(&dataset_f[fileIndex], 
						    name, symbol, units, 
						    description, format_string, 
						    type, buffer));
      } else if (PyFloat_Check(fixed_value)) {
	fixedDoubleValue = PyFloat_AsDouble(fixed_value);
	SDDS_SprintTypedValue(&fixedDoubleValue, 0, SDDS_DOUBLE , format_string, buffer, 0);
	return PyLong_FromLong(SDDS_DefineParameter(&dataset_f[fileIndex], 
						    name, symbol, units, 
						    description, format_string, 
						    type, buffer));
      }
    }
    return PyLong_FromLong(SDDS_DefineParameter1(&dataset_f[fileIndex], 
						   name, symbol, units, 
						   description, format_string, 
						   type, NULL));
  }
}

/* SDDS_DefineArray
   Returns: -1 on error
            index on success */
static PyObject* sddsdata_DefineArray( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *name;
  char *symbol;
  char *units;
  char *description;
  char *format_string;
  long type;
  long field_length;
  long dimensions;
  char *group_name;
  if (!PyArg_ParseTuple(args, "lsssssslll", &fileIndex, &name , &symbol, &units, &description, &format_string, &group_name, &type, &field_length, &dimensions)) {
    return NULL;
  }
  if (symbol) 
    if (strlen(symbol) == 0)
      symbol = NULL;
  if (units) 
    if (strlen(units) == 0)
      units = NULL;
  if (description) 
    if (strlen(description) == 0)
      description = NULL;   
  if (format_string) 
    if (strlen(format_string) == 0)
      format_string = NULL;
  if (group_name) 
    if (strlen(group_name) == 0)
      group_name = NULL;
  return PyLong_FromLong(SDDS_DefineArray(&dataset_f[fileIndex], name, symbol, units, description, format_string, type, field_length, dimensions, group_name));
}

/* SDDS_DefineColumn
   Returns: -1 on error
            index on success */
static PyObject* sddsdata_DefineColumn( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *name;
  char *symbol;
  char *units;
  char *description;
  char *format_string;
  long type;
  long field_length;
  if (!PyArg_ParseTuple(args, "lsssssll", &fileIndex, &name , &symbol, &units, &description, &format_string, &type, &field_length)) {
    return NULL;
  }
  if (symbol) 
    if (strlen(symbol) == 0)
      symbol = NULL;
  if (units) 
    if (strlen(units) == 0)
      units = NULL;
  if (description) 
    if (strlen(description) == 0)
      description = NULL;   
  if (format_string) 
    if (strlen(format_string) == 0)
      format_string = NULL;
  return PyLong_FromLong(SDDS_DefineColumn(&dataset_f[fileIndex], name, symbol, units, description, format_string, type, field_length));
}

/* SDDS_IsValidName
   Returns: 0 on invalid name
            1 on valid name */
static PyObject* sddsdata_IsValidName( PyObject* self, PyObject* args )
{
  char *name;
  if (!PyArg_ParseTuple(args, "s", &name)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_IsValidName(name, NULL));
}

/* SDDS_SetNameValidityFlags
   Sets nameValidityFlags to: SDDS_ALLOW_ANY_NAME
   Returns: previous value for nameValidityFlags */
static PyObject* sddsdata_SetNameValidityFlags( PyObject* self, PyObject* args )
{
  return PyLong_FromLong(SDDS_SetNameValidityFlags(SDDS_ALLOW_ANY_NAME));
}

/* SDDS_DefineSimpleColumn
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_DefineSimpleColumn( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *name;
  char *units;
  long type;
  if (!PyArg_ParseTuple(args, "lssl", &fileIndex, &name, &units, &type)) {
    return NULL;
  }
  if (units) 
    if (strlen(units) == 0)
      units = NULL;   
  return PyLong_FromLong(SDDS_DefineSimpleColumn(&dataset_f[fileIndex], name, units, type));
}

/* SDDS_DefineSimpleArray
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_DefineSimpleArray( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *name;
  char *units;
  long type;
  long dimensions;
  long results;
  if (!PyArg_ParseTuple(args, "lssll", &fileIndex, &name, &units, &type, &dimensions)) {
    return NULL;
  }
  if (units) 
    if (strlen(units) == 0)
      units = NULL;
  results = SDDS_DefineArray(&dataset_f[fileIndex], name, NULL, units, NULL, NULL, type, 0, dimensions, NULL);
  if (results == -1) {
    results = 0;
  } else {
    results = 1;
  }
  return PyLong_FromLong(results);
}

/* SDDS_DefineSimpleParameter
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_DefineSimpleParameter( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *name;
  char *units;
  long type;
  if (!PyArg_ParseTuple(args, "lssl", &fileIndex, &name, &units, &type)) {
    return NULL;
  }
  if (units) 
    if (strlen(units) == 0)
      units = NULL;   
  return PyLong_FromLong(SDDS_DefineSimpleParameter(&dataset_f[fileIndex], name, units, type));
}

/* SDDS_WriteLayout
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_WriteLayout( PyObject* self, PyObject* args )
{
  long fileIndex;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_WriteLayout(&dataset_f[fileIndex]));
}

/* SDDS_EraseData
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_EraseData( PyObject* self, PyObject* args )
{
  long fileIndex;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_EraseData(&dataset_f[fileIndex]));
}

/* SDDS_ProcessColumnString
   Argument: mode
     Valid values: 0, 1 (SDDS_NORMAL_DEFINITION, SDDS_WRITEONLY_DEFINITION)
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_ProcessColumnString( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *string;
  long mode;
  if (!PyArg_ParseTuple(args, "lsl", &fileIndex, &string, &mode)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_ProcessColumnString(&dataset_f[fileIndex], string, mode));
}

/* SDDS_ProcessArrayString
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_ProcessArrayString( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *string;
  if (!PyArg_ParseTuple(args, "ls", &fileIndex, &string)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_ProcessArrayString(&dataset_f[fileIndex], string));
}

/* SDDS_ProcessParameterString
   Argument: mode
     Valid values: 0, 1 (SDDS_NORMAL_DEFINITION, SDDS_WRITEONLY_DEFINITION)
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_ProcessParameterString( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *string;
  long mode;
  if (!PyArg_ParseTuple(args, "lsl", &fileIndex, &string, &mode)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_ProcessParameterString(&dataset_f[fileIndex], string, mode));
}

/* SDDS_InitializeCopy
   Argument: filemode
     Valid values: "r", "w", "rb", "wb", "m"
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_InitializeCopy( PyObject* self, PyObject* args )
{
  long fileIndexTarget;
  long fileIndexSource;
  char *filename;
  char *filemode;
  if (!PyArg_ParseTuple(args, "llss", &fileIndexTarget, &fileIndexSource, &filename, &filemode)) {
    return NULL;
  }
  if (filename)
    if (strlen(filename) == 0) 
      filename = NULL;
  return PyLong_FromLong(SDDS_InitializeCopy(&dataset_f[fileIndexTarget], &dataset_f[fileIndexSource], filename, filemode));
}

/* SDDS_CopyLayout
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_CopyLayout( PyObject* self, PyObject* args )
{
  long fileIndexTarget;
  long fileIndexSource;
  if (!PyArg_ParseTuple(args, "ll", &fileIndexTarget, &fileIndexSource)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_CopyLayout(&dataset_f[fileIndexTarget], &dataset_f[fileIndexSource]));
}

/* SDDS_AppendLayout
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_AppendLayout( PyObject* self, PyObject* args )
{
  long fileIndexTarget;
  long fileIndexSource;
  if (!PyArg_ParseTuple(args, "ll", &fileIndexTarget, &fileIndexSource)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_AppendLayout(&dataset_f[fileIndexTarget], &dataset_f[fileIndexSource], 0));
}

/* SDDS_CopyPage
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_CopyPage( PyObject* self, PyObject* args )
{
  long fileIndexTarget;
  long fileIndexSource;
  if (!PyArg_ParseTuple(args, "ll", &fileIndexTarget, &fileIndexSource)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_CopyPage(&dataset_f[fileIndexTarget], &dataset_f[fileIndexSource]));
}

/* SDDS_CopyParameters
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_CopyParameters( PyObject* self, PyObject* args )
{
  long fileIndexTarget;
  long fileIndexSource;
  if (!PyArg_ParseTuple(args, "ll", &fileIndexTarget, &fileIndexSource)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_CopyParameters(&dataset_f[fileIndexTarget], &dataset_f[fileIndexSource]));
}

/* SDDS_CopyArrays
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_CopyArrays( PyObject* self, PyObject* args )
{
  long fileIndexTarget;
  long fileIndexSource;
  if (!PyArg_ParseTuple(args, "ll", &fileIndexTarget, &fileIndexSource)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_CopyArrays(&dataset_f[fileIndexTarget], &dataset_f[fileIndexSource]));
}

/* SDDS_CopyColumns
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_CopyColumns( PyObject* self, PyObject* args )
{
  long fileIndexTarget;
  long fileIndexSource;
  if (!PyArg_ParseTuple(args, "ll", &fileIndexTarget, &fileIndexSource)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_CopyColumns(&dataset_f[fileIndexTarget], &dataset_f[fileIndexSource]));
}

/* SDDS_CopyRow
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_CopyRow( PyObject* self, PyObject* args )
{
  long fileIndexTarget;
  long target_row;
  long fileIndexSource;
  long source_row;
  if (!PyArg_ParseTuple(args, "llll", &fileIndexTarget, &target_row, &fileIndexSource, &source_row)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_CopyRow(&dataset_f[fileIndexTarget], target_row, &dataset_f[fileIndexSource], source_row));
}

/* SDDS_CopyRowDirect
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_CopyRowDirect( PyObject* self, PyObject* args )
{
  long fileIndexTarget;
  long target_row;
  long fileIndexSource;
  long source_row;
  if (!PyArg_ParseTuple(args, "llll", &fileIndexTarget, &target_row, &fileIndexSource, &source_row)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_CopyRowDirect(&dataset_f[fileIndexTarget], target_row, &dataset_f[fileIndexSource], source_row));
}

/* SDDS_CopyAdditionalRows
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_CopyAdditionalRows( PyObject* self, PyObject* args )
{
  long fileIndexTarget;
  long fileIndexSource;
  if (!PyArg_ParseTuple(args, "ll", &fileIndexTarget, &fileIndexSource)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_CopyAdditionalRows(&dataset_f[fileIndexTarget], &dataset_f[fileIndexSource]));
}

/* SDDS_DeferSavingLayout
   Argument: mode
     Valid values: 0, 1 (False, True) */
static PyObject* sddsdata_DeferSavingLayout( PyObject* self, PyObject* args )
{
  long fileIndex, mode;
  if (!PyArg_ParseTuple(args, "ll", &fileIndex, &mode)) {
    return NULL;
  }
  SDDS_DeferSavingLayout(&dataset_f[fileIndex],mode);
  Py_INCREF(Py_None);
  return Py_None;
}

/* SDDS_SaveLayout
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_SaveLayout( PyObject* self, PyObject* args )
{
  long fileIndex;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_SaveLayout(&dataset_f[fileIndex]));
}

/* SDDS_RestoreLayout
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_RestoreLayout( PyObject* self, PyObject* args )
{
  long fileIndex;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_RestoreLayout(&dataset_f[fileIndex]));
}

/* SDDS_StartPage
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_StartPage( PyObject* self, PyObject* args )
{
  long fileIndex;
  long expected_n_rows;
  if (!PyArg_ParseTuple(args, "ll", &fileIndex, &expected_n_rows)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_StartPage(&dataset_f[fileIndex], expected_n_rows));
}

/* SDDS_ClearPage
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_ClearPage( PyObject* self, PyObject* args )
{
  long fileIndex;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_ClearPage(&dataset_f[fileIndex]));
}

/* SDDS_LengthenTable
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_LengthenTable( PyObject* self, PyObject* args )
{
  long fileIndex;
  long n_additional_rows;
  if (!PyArg_ParseTuple(args, "ll", &fileIndex, &n_additional_rows)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_LengthenTable(&dataset_f[fileIndex], n_additional_rows));
}

/* SDDS_WritePage
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_WritePage( PyObject* self, PyObject* args )
{
  long fileIndex;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_WritePage(&dataset_f[fileIndex]));
}

/* SDDS_UpdatePage
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_UpdatePage( PyObject* self, PyObject* args )
{
  long fileIndex, mode;
  if (!PyArg_ParseTuple(args, "ll", &fileIndex, &mode)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_UpdatePage(&dataset_f[fileIndex], mode));
}

/* SDDS_InitializeHeaderlessInput
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_InitHeaderlessInput( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *filename;
  if (!PyArg_ParseTuple(args, "ls", &fileIndex, &filename)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_InitializeHeaderlessInput(&dataset_f[fileIndex], filename));
}

/* SDDS_ReadPage
   Returns: 0 on error
            -1 on EOF
            page number on success */
static PyObject* sddsdata_ReadPage( PyObject* self, PyObject* args )
{
  long fileIndex;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_ReadPage(&dataset_f[fileIndex]));
}

/* SDDS_ReadPageSparse
   Returns: 0 on error
            -1 on EOF
            page number on success */
static PyObject* sddsdata_ReadPageSparse( PyObject* self, PyObject* args )
{
  long fileIndex;
  long sparse_interval;
  long sparse_offset;
  if (!PyArg_ParseTuple(args, "lll", &fileIndex, &sparse_interval, &sparse_offset)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_ReadPageSparse(&dataset_f[fileIndex], 0, sparse_interval, sparse_offset));
}

/* SDDS_ReadPageLastRows
   Returns: 0 on error
            -1 on EOF
            page number on success */
static PyObject* sddsdata_ReadPageLastRows( PyObject* self, PyObject* args )
{
  long fileIndex;
  long last_rows;
  if (!PyArg_ParseTuple(args, "ll", &fileIndex, &last_rows)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_ReadPageLastRows(&dataset_f[fileIndex], last_rows));
}

/* SDDS_RowCount
   Returns number of rows */
static PyObject* sddsdata_RowCount( PyObject* self, PyObject* args )
{
  long fileIndex;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  return PyLong_FromLongLong((long long)SDDS_RowCount(&dataset_f[fileIndex]));
}

/* SDDS_SetColumnFlags
   Argument: column_flag_value
     Valid values: 0, 1
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_SetColumnFlags( PyObject* self, PyObject* args )
{
  long fileIndex;
  long column_flag_value;
  if (!PyArg_ParseTuple(args, "ll", &fileIndex, &column_flag_value)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_SetColumnFlags(&dataset_f[fileIndex], column_flag_value));
}

/* SDDS_SetRowFlags
   Argument: row_flag_value
     Valid values: 0, 1
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_SetRowFlags( PyObject* self, PyObject* args )
{
  long fileIndex;
  long row_flag_value;
  if (!PyArg_ParseTuple(args, "ll", &fileIndex, &row_flag_value)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_SetRowFlags(&dataset_f[fileIndex], row_flag_value));
}

/* SDDS_GetRowFlag
   Returns: -1 on error
            row flag on success */
static PyObject* sddsdata_GetRowFlag( PyObject* self, PyObject* args )
{
  long fileIndex;
  long row;
  if (!PyArg_ParseTuple(args, "ll", &fileIndex, &row)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_GetRowFlag(&dataset_f[fileIndex], row));
}

/* SDDS_DeleteColumn
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_DeleteColumn( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *column_name;
  if (!PyArg_ParseTuple(args, "ls", &fileIndex, &column_name)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_DeleteColumn(&dataset_f[fileIndex], column_name));
}

/* SDDS_DeleteParameter
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_DeleteParameter( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *parameter_name;
  if (!PyArg_ParseTuple(args, "ls", &fileIndex, &parameter_name)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_DeleteParameter(&dataset_f[fileIndex], parameter_name));
}

/* SDDS_DeleteUnsetColumns
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_DeleteUnsetColumns( PyObject* self, PyObject* args )
{
  long fileIndex;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_DeleteUnsetColumns(&dataset_f[fileIndex]));
}

/* SDDS_DeleteUnsetRows
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_DeleteUnsetRows( PyObject* self, PyObject* args )
{
  long fileIndex;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_DeleteUnsetRows(&dataset_f[fileIndex]));
}

/* SDDS_ColumnCount
   Returns: number of columns */
static PyObject* sddsdata_ColumnCount( PyObject* self, PyObject* args )
{
  long fileIndex;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_ColumnCount(&dataset_f[fileIndex]));
}

/* SDDS_ArrayCount
   Returns: number of arrays */
static PyObject* sddsdata_ArrayCount( PyObject* self, PyObject* args )
{
  long fileIndex;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_ArrayCount(&dataset_f[fileIndex]));
}

/* SDDS_ParameterCount
   Returns: number of parameters */
static PyObject* sddsdata_ParameterCount( PyObject* self, PyObject* args )
{
  long fileIndex;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_ParameterCount(&dataset_f[fileIndex]));
}

/* SDDS_GetDescription
   Modified from standard version
   Returns: NULL on error
            text and description on success */
static PyObject* sddsdata_GetDescription( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *text;
  char *contents;
  PyObject *v;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  if (SDDS_GetDescription(&dataset_f[fileIndex], &text, &contents) == 0) {
    return NULL;
  }
  if (!(v = PyList_New(2)))
    return NULL;
  if (text) {
    PyList_SetItem(v, 0, PyString_FromString(text));
    free(text);
  } else {
    PyList_SetItem(v, 0, PyString_FromString(""));
  }
  if (contents) {
    PyList_SetItem(v, 1, PyString_FromString(contents));
    free(contents);
  } else {
    PyList_SetItem(v, 1, PyString_FromString(""));
  }
  return v;
}

/* SDDS_GetDescription
   Modified from standard version
   Returns: NULL on error
            text on success */
static PyObject* sddsdata_GetDescriptionText( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *text;
  char *contents;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  if (SDDS_GetDescription(&dataset_f[fileIndex], &text, &contents) == 0) {
    return NULL;
  }
  if (text)
    return PyString_FromString(text);
  return PyString_FromString("");
}

/* SDDS_GetDescription
   Modified from standard version
   Returns: NULL on error
            contents on success */
static PyObject* sddsdata_GetDescriptionContents( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *text;
  char *contents;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  if (SDDS_GetDescription(&dataset_f[fileIndex], &text, &contents) == 0) {
    return NULL;
  }
  if (contents)
    return PyString_FromString(contents);
  return PyString_FromString("");
}

/* SDDS_NumberOfErrors
   Returns: number of errors */
static PyObject* sddsdata_NumberOfErrors( PyObject* self, PyObject* args ) 
{
  return PyLong_FromLong(SDDS_NumberOfErrors());
}

/* SDDS_ClearErrors 
   Sets number of errors to 0 */
static PyObject* sddsdata_ClearErrors( PyObject* self, PyObject* args ) 
{
  SDDS_ClearErrors();
  Py_INCREF(Py_None);
  return Py_None;
}

/* SDDS_SetError
   Sets error description */
static PyObject* sddsdata_SetError( PyObject* self, PyObject* args ) 
{
  char *error_text;
  if (!PyArg_ParseTuple(args, "s", &error_text)) {
    return NULL;
  }
  SDDS_SetError(error_text);
  Py_INCREF(Py_None);
  return Py_None;
}

/* SDDS_Bomb
   Displays message on stderr and exits */
static PyObject* sddsdata_Bomb( PyObject* self, PyObject* args ) 
{
  char *message;
  if (!PyArg_ParseTuple(args, "s", &message)) {
    return NULL;
  }
  SDDS_Bomb(message);
  Py_INCREF(Py_None);
  return Py_None;
}

/* SDDS_Warning
   Displays message on stderr */
static PyObject* sddsdata_Warning( PyObject* self, PyObject* args ) 
{
  char *message;
  if (!PyArg_ParseTuple(args, "s", &message)) {
    return NULL;
  }
  SDDS_Warning(message);
  Py_INCREF(Py_None);
  return Py_None;
}

/* SDDS_RegisterProgramName
   Sets registeredProgramName to be printed with some error messages */
static PyObject* sddsdata_RegisterProgramName( PyObject* self, PyObject* args ) 
{
  char *name;
  if (!PyArg_ParseTuple(args, "s", &name)) {
    return NULL;
  }
  SDDS_RegisterProgramName(name);
  Py_INCREF(Py_None);
  return Py_None;
}

/* SDDS_PrintErrors
   Argument: Mode
     Valid values: 0, 1, 2, 3 (0, SDDS_VERBOSE_PrintErrors, SDDS_EXIT_PrintErrors, 
                               SDDS_VERBOSE_PrintErrors|SDDS_EXIT_PrintErrors)
   Prints error messages that were set with SDDS_SetError */
static PyObject* sddsdata_PrintErrors( PyObject* self, PyObject* args ) 
{
  long mode;
  if (!PyArg_ParseTuple(args, "l", &mode)) {
    return NULL;
  }
  SDDS_PrintErrors(stderr, mode);
  Py_INCREF(Py_None);
  return Py_None;
}

/* SDDS_TransferColumnDefinition
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_TransferColumnDefinition( PyObject* self, PyObject* args )
{
  long fileIndexTarget;
  long fileIndexSource;
  char *name;
  char *newName;
  if (!PyArg_ParseTuple(args, "llss", &fileIndexTarget, &fileIndexSource, &name, &newName)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_TransferColumnDefinition(&dataset_f[fileIndexTarget], &dataset_f[fileIndexSource], name, newName));
}

/* SDDS_TransferArrayDefinition
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_TransferArrayDefinition( PyObject* self, PyObject* args )
{
  long fileIndexTarget;
  long fileIndexSource;
  char *name;
  char *newName;
  if (!PyArg_ParseTuple(args, "llss", &fileIndexTarget, &fileIndexSource, &name, &newName)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_TransferArrayDefinition(&dataset_f[fileIndexTarget], &dataset_f[fileIndexSource], name, newName));
}

/* SDDS_TransferParameterDefinition
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_TransferParameterDefinition( PyObject* self, PyObject* args )
{
  long fileIndexTarget;
  long fileIndexSource;
  char *name;
  char *newName;
  if (!PyArg_ParseTuple(args, "llss", &fileIndexTarget, &fileIndexSource, &name, &newName)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_TransferParameterDefinition(&dataset_f[fileIndexTarget], &dataset_f[fileIndexSource], name, newName));
}

/* SDDS_DefineColumnLikeParameter
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_DefineColumnLikeParameter( PyObject* self, PyObject* args )
{
  long fileIndexTarget;
  long fileIndexSource;
  char *name;
  char *newName;
  if (!PyArg_ParseTuple(args, "llss", &fileIndexTarget, &fileIndexSource, &name, &newName)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_DefineColumnLikeParameter(&dataset_f[fileIndexTarget], &dataset_f[fileIndexSource], name, newName));
}

/* SDDS_DefineParameterLikeColumn
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_DefineParameterLikeColumn( PyObject* self, PyObject* args )
{
  long fileIndexTarget;
  long fileIndexSource;
  char *name;
  char *newName;
  if (!PyArg_ParseTuple(args, "llss", &fileIndexTarget, &fileIndexSource, &name, &newName)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_DefineParameterLikeColumn(&dataset_f[fileIndexTarget], &dataset_f[fileIndexSource], name, newName));
}

/* SDDS_TransferAllColumnDefinitions
   Argument: mode
     Valid values: 0, 1, 2, 3 (0, SDDS_TRANSFER_KEEPOLD, SDDS_TRANSFER_OVERWRITE, 
                               SDDS_TRANSFER_KEEPOLD|SDDS_TRANSFER_OVERWRITE)
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_TransferAllColumnDefinitions( PyObject* self, PyObject* args )
{
  long fileIndexTarget;
  long fileIndexSource;
  long mode;
  if (!PyArg_ParseTuple(args, "lll", &fileIndexTarget, &fileIndexSource, &mode)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_TransferAllColumnDefinitions(&dataset_f[fileIndexTarget], &dataset_f[fileIndexSource], mode));
}

/* SDDS_TransferAllArrayDefinitions
   Argument: mode
     Valid values: 0, 1, 2, 3 (0, SDDS_TRANSFER_KEEPOLD, SDDS_TRANSFER_OVERWRITE, 
                               SDDS_TRANSFER_KEEPOLD|SDDS_TRANSFER_OVERWRITE)
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_TransferAllArrayDefinitions( PyObject* self, PyObject* args )
{
  long fileIndexTarget;
  long fileIndexSource;
  long mode;
  if (!PyArg_ParseTuple(args, "lll", &fileIndexTarget, &fileIndexSource, &mode)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_TransferAllArrayDefinitions(&dataset_f[fileIndexTarget], &dataset_f[fileIndexSource], mode));
}

/* SDDS_TransferAllParameterDefinitions
   Argument: mode
     Valid values: 0, 1, 2, 3 (0, SDDS_TRANSFER_KEEPOLD, SDDS_TRANSFER_OVERWRITE, 
                               SDDS_TRANSFER_KEEPOLD|SDDS_TRANSFER_OVERWRITE)
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_TransferAllParameterDefinitions( PyObject* self, PyObject* args )
{
  long fileIndexTarget;
  long fileIndexSource;
  long mode;
  if (!PyArg_ParseTuple(args, "lll", &fileIndexTarget, &fileIndexSource, &mode)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_TransferAllParameterDefinitions(&dataset_f[fileIndexTarget], &dataset_f[fileIndexSource], mode));
}

/* SDDS_GetColumnIndex
   Returns: -1 on error
            index on success */
static PyObject* sddsdata_GetColumnIndex( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *name;
  if (!PyArg_ParseTuple(args, "ls", &fileIndex, &name)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_GetColumnIndex(&dataset_f[fileIndex], name));
}

/* SDDS_GetArrayIndex
   Returns: -1 on error
            index on success */
static PyObject* sddsdata_GetArrayIndex( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *name;
  if (!PyArg_ParseTuple(args, "ls", &fileIndex, &name)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_GetArrayIndex(&dataset_f[fileIndex], name));
}

/* SDDS_GetParameterIndex
   Returns: -1 on error
            index on success */
static PyObject* sddsdata_GetParameterIndex( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *name;
  if (!PyArg_ParseTuple(args, "ls", &fileIndex, &name)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_GetParameterIndex(&dataset_f[fileIndex], name));
}

/* SDDS_GetColumnType
   Returns: 0 on error
            type on success */
static PyObject* sddsdata_GetColumnType( PyObject* self, PyObject* args )
{
  long fileIndex;
  long index;
  if (!PyArg_ParseTuple(args, "ll", &fileIndex, &index)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_GetColumnType(&dataset_f[fileIndex], index));
}

/* SDDS_GetArrayType
   Returns: 0 on error
            type on success */
static PyObject* sddsdata_GetArrayType( PyObject* self, PyObject* args )
{
  long fileIndex;
  long index;
  if (!PyArg_ParseTuple(args, "ll", &fileIndex, &index)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_GetArrayType(&dataset_f[fileIndex], index));
}

/* SDDS_GetNamedColumnType
   Returns: 0 on error
            type on success */
static PyObject* sddsdata_GetNamedColumnType( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *name;
  if (!PyArg_ParseTuple(args, "ls", &fileIndex, &name)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_GetNamedColumnType(&dataset_f[fileIndex], name));
}

/* SDDS_GetNamedArrayType
   Returns: 0 on error
            type on success */
static PyObject* sddsdata_GetNamedArrayType( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *name;
  if (!PyArg_ParseTuple(args, "ls", &fileIndex, &name)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_GetNamedArrayType(&dataset_f[fileIndex], name));
}

/* SDDS_GetColumnDefinition
   Returns: NULL on error
            definition on success */
static PyObject* sddsdata_GetColumnDefinition( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *name;
  COLUMN_DEFINITION *pardef;
  PyObject *v;
  if (!PyArg_ParseTuple(args, "ls", &fileIndex, &name)) {
    return NULL;
  }
  if ((pardef = SDDS_GetColumnDefinition(&dataset_f[fileIndex], name)) == NULL)
    return NULL;
  if (!(v = PyList_New(6)))
    return NULL;

  if (pardef->symbol)
    PyList_SetItem(v, 0, PyString_FromString(pardef->symbol));
  else
    PyList_SetItem(v, 0, PyString_FromString(""));
  if (pardef->units)
    PyList_SetItem(v, 1, PyString_FromString(pardef->units));
  else
    PyList_SetItem(v, 1, PyString_FromString(""));
  if (pardef->description)
    PyList_SetItem(v, 2, PyString_FromString(pardef->description));
  else
    PyList_SetItem(v, 2, PyString_FromString(""));
  if (pardef->format_string)
    PyList_SetItem(v, 3, PyString_FromString(pardef->format_string));
  else
    PyList_SetItem(v, 3, PyString_FromString(""));
  PyList_SetItem(v, 4, PyLong_FromLong(pardef->type));
  PyList_SetItem(v, 5, PyLong_FromLong(pardef->field_length));

  SDDS_FreeColumnDefinition(pardef);

  return v;
}

/* SDDS_GetArrayDefinition
   Returns: NULL on error
            definition on success */
static PyObject* sddsdata_GetArrayDefinition( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *name;
  ARRAY_DEFINITION *pardef;
  PyObject *v;
  if (!PyArg_ParseTuple(args, "ls", &fileIndex, &name)) {
    return NULL;
  }
  if ((pardef = SDDS_GetArrayDefinition(&dataset_f[fileIndex], name)) == NULL)
    return NULL;
  if (!(v = PyList_New(8)))
    return NULL;

  if (pardef->symbol)
    PyList_SetItem(v, 0, PyString_FromString(pardef->symbol));
  else
    PyList_SetItem(v, 0, PyString_FromString(""));
  if (pardef->units)
    PyList_SetItem(v, 1, PyString_FromString(pardef->units));
  else
    PyList_SetItem(v, 1, PyString_FromString(""));
  if (pardef->description)
    PyList_SetItem(v, 2, PyString_FromString(pardef->description));
  else
    PyList_SetItem(v, 2, PyString_FromString(""));
  if (pardef->format_string)
    PyList_SetItem(v, 3, PyString_FromString(pardef->format_string));
  else
    PyList_SetItem(v, 3, PyString_FromString(""));
  if (pardef->group_name)
    PyList_SetItem(v, 4, PyString_FromString(pardef->group_name));
  else
    PyList_SetItem(v, 4, PyString_FromString(""));
  PyList_SetItem(v, 5, PyLong_FromLong(pardef->type));
  PyList_SetItem(v, 6, PyLong_FromLong(pardef->field_length));
  PyList_SetItem(v, 7, PyLong_FromLong(pardef->dimensions));

  SDDS_FreeArrayDefinition(pardef);

  return v;
}

/* SDDS_GetParameterType
   Returns: 0 on error
            type on success */
static PyObject* sddsdata_GetParameterType( PyObject* self, PyObject* args )
{
  long fileIndex;
  long index;
  if (!PyArg_ParseTuple(args, "ll", &fileIndex, &index)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_GetParameterType(&dataset_f[fileIndex], index));
}

/* SDDS_GetNamedParameterType
   Returns: 0 on error
            type on success */
static PyObject* sddsdata_GetNamedParameterType( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *name;
  if (!PyArg_ParseTuple(args, "ls", &fileIndex, &name)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_GetNamedParameterType(&dataset_f[fileIndex], name));  
}

/* SDDS_GetParameterDefinition
   Returns: NULL on error
            definition on success */
static PyObject* sddsdata_GetParameterDefinition( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *name;
  PARAMETER_DEFINITION *pardef;
  PyObject *v;
  if (!PyArg_ParseTuple(args, "ls", &fileIndex, &name)) {
    return NULL;
  }
  if ((pardef = SDDS_GetParameterDefinition(&dataset_f[fileIndex], name)) == NULL)
    return NULL;
  if (!(v = PyList_New(6)))
    return NULL;

  if (pardef->symbol)
    PyList_SetItem(v, 0, PyString_FromString(pardef->symbol));
  else
    PyList_SetItem(v, 0, PyString_FromString(""));
  if (pardef->units)
    PyList_SetItem(v, 1, PyString_FromString(pardef->units));
  else
    PyList_SetItem(v, 1, PyString_FromString(""));
  if (pardef->description)
    PyList_SetItem(v, 2, PyString_FromString(pardef->description));
  else
    PyList_SetItem(v, 2, PyString_FromString(""));
  if (pardef->format_string)
    PyList_SetItem(v, 3, PyString_FromString(pardef->format_string));
  else
    PyList_SetItem(v, 3, PyString_FromString(""));
  PyList_SetItem(v, 4, PyLong_FromLong(pardef->type));
  if (pardef->fixed_value)
    PyList_SetItem(v, 5, PyString_FromString(pardef->fixed_value));
  else
    PyList_SetItem(v, 5, PyString_FromString(""));

  SDDS_FreeParameterDefinition(pardef);

  return v;
}

/* SDDS_GetTypeSize
   Returns: -1 on error
            size on success */
static PyObject* sddsdata_GetTypeSize( PyObject* self, PyObject* args )
{
  long type;
  if (!PyArg_ParseTuple(args, "l", &type)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_GetTypeSize(type));
}

/* SDDS_GetTypeName
   Returns: NULL on error
            name on success */
static PyObject* sddsdata_GetTypeName( PyObject* self, PyObject* args )
{
  long type;
  char *name;
  char msgbuf[256];
  if (!PyArg_ParseTuple(args, "l", &type)) {
    return NULL;
  }
  name = SDDS_GetTypeName(type);
  if (name)
    return PyString_FromString(name);
  sprintf(msgbuf, "sdds.GetTypeName: %ld is an invalid SDDS data type", type);
  PyErr_SetString(PyExc_StandardError, msgbuf);
  return NULL;
}

/* SDDS_IdentifyType
   Argument: typeName
     Valid values: double, float, long, short, string, character
   Returns: 0 on error
            type on success */ 
static PyObject* sddsdata_IdentifyType( PyObject* self, PyObject* args )
{
  char *typeName;
  if (!PyArg_ParseTuple(args, "s", &typeName)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_IdentifyType(typeName));
}

/* SDDS_CheckColumn
   Returns: 0 (SDDS_CHECK_OKAY)
            1 (SDDS_CHECK_NONEXISTENT)
            2 (SDDS_CHECK_WRONGTYPE)
            3 (SDDS_CHECK_WRONGUNITS) */
static PyObject* sddsdata_CheckColumn( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *name;
  char *units;
  long type;
  if (!PyArg_ParseTuple(args, "lssl", &fileIndex, &name, &units, &type)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_CheckColumn(&dataset_f[fileIndex], name, units, type, stderr));
}

/* SDDS_CheckArray
   Returns: 0 (SDDS_CHECK_OKAY)
            1 (SDDS_CHECK_NONEXISTENT)
            2 (SDDS_CHECK_WRONGTYPE)
            3 (SDDS_CHECK_WRONGUNITS) */
static PyObject* sddsdata_CheckArray( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *name;
  char *units;
  long type;
  if (!PyArg_ParseTuple(args, "lssl", &fileIndex, &name, &units, &type)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_CheckArray(&dataset_f[fileIndex], name, units, type, stderr));
}

/* SDDS_CheckParameter
   Returns: 0 (SDDS_CHECK_OKAY)
            1 (SDDS_CHECK_NONEXISTENT)
            2 (SDDS_CHECK_WRONGTYPE)
            3 (SDDS_CHECK_WRONGUNITS) */
static PyObject* sddsdata_CheckParameter( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *name;
  char *units;
  long type;
  if (!PyArg_ParseTuple(args, "lssl", &fileIndex, &name, &units, &type)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_CheckParameter(&dataset_f[fileIndex], name, units, type, stderr));
}

/* SDDS_HasWhitespace
   Returns: 0 (false)
            1 (true) */
static PyObject* sddsdata_HasWhitespace( PyObject* self, PyObject* args )
{
  char *string;
  if (!PyArg_ParseTuple(args, "s", &string)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_HasWhitespace(string));
}

/* SDDS_StringIsBlank
   Returns: 0 (false)
            1 (true) */
static PyObject* sddsdata_StringIsBlank( PyObject* self, PyObject* args )
{
  char *s;
  if (!PyArg_ParseTuple(args, "s", &s)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_StringIsBlank(s));
}

/* SDDS_ApplyFactorToParameter
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_ApplyFactorToParameter( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *name;
  double factor;
  if (!PyArg_ParseTuple(args, "lsd", &fileIndex, &name, &factor)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_ApplyFactorToParameter(&dataset_f[fileIndex], name, factor));
}

/* SDDS_ApplyFactorToColumn
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_ApplyFactorToColumn( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *name;
  double factor;
  if (!PyArg_ParseTuple(args, "lsd", &fileIndex, &name, &factor)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_ApplyFactorToColumn(&dataset_f[fileIndex], name, factor));
}

/* SDDS_DeleteParameterFixedValues
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_DeleteParameterFixedValues( PyObject* self, PyObject* args )
{
  long fileIndex;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_DeleteParameterFixedValues(&dataset_f[fileIndex]));
}

/* SDDS_SetDataMode
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_SetDataMode( PyObject* self, PyObject* args )
{
  long fileIndex;
  long newmode;
  if (!PyArg_ParseTuple(args, "ll", &fileIndex, &newmode)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_SetDataMode(&dataset_f[fileIndex], newmode));
}

/* SDDS_CheckDataset
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_CheckDataset( PyObject* self, PyObject* args )
{
  long fileIndex;
  char *caller;
  if (!PyArg_ParseTuple(args, "ls", &fileIndex, &caller)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_CheckDataset(&dataset_f[fileIndex], caller));
}

/* SDDS_SetAutoCheckMode
   Argument: newMode
     Valid values: 0, 1 (0, TABULAR_DATA_CHECKS)
   Returns: previous mode */
static PyObject* sddsdata_SetAutoCheckMode( PyObject* self, PyObject* args )
{
  long newMode;
  if (!PyArg_ParseTuple(args, "l", &newMode)) {
    return NULL;
  }
  return PyLong_FromLong(SDDS_SetAutoCheckMode(newMode));
}

/* SDDS_GetColumnNames
   Modified from standard version
   Returns: NULL on error
            name on success */
static PyObject* sddsdata_GetColumnNameFromIndex( PyObject* self, PyObject* args )
{
  long fileIndex;
  long index;
  char **data;
  int32_t number;
  PyObject *v;
  long i;
  if (!PyArg_ParseTuple(args, "ll", &fileIndex, &index)) {
    return NULL;
  }
  data = SDDS_GetColumnNames(&dataset_f[fileIndex], &number);
  if (!(data))
    return NULL;
  if ((index >= 0) && (index < number)) {
    v = PyString_FromString(data[index]);
    for (i=0;i<number;i++)
      free(data[i]);
    free(data);
    return v;
  }
  for (i=0;i<number;i++)
    free(data[i]);
  free(data);
  return NULL;
}

/* SDDS_GetColumnNames
   Modified from standard version
   Returns: NULL on error
            names on success */
static PyObject* sddsdata_GetColumnNames( PyObject* self, PyObject* args )
{
  long fileIndex;
  long index;
  char **data;
  int32_t number;
  PyObject *v;
  long i;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  data = SDDS_GetColumnNames(&dataset_f[fileIndex], &number);
  if (!(data))
    return NULL;
  if (!(v = PyList_New(number)))
    return NULL;
  for (index = 0; index < number; index++)
    PyList_SetItem(v, index, PyString_FromString(data[index]));
  for (i=0;i<number;i++)
    free(data[i]);
  free(data);
  return v;
}

/* SDDS_GetArrayNames
   Modified from standard version
   Returns: NULL on error
            name on success */
static PyObject* sddsdata_GetArrayNameFromIndex( PyObject* self, PyObject* args )
{
  long fileIndex;
  long index;
  char **data;
  int32_t number;
  PyObject *v;
  long i;
  if (!PyArg_ParseTuple(args, "ll", &fileIndex, &index)) {
    return NULL;
  }
  data = SDDS_GetArrayNames(&dataset_f[fileIndex], &number);
  if (!(data))
    return NULL;
  if ((index >= 0) && (index < number)) {
    v = PyString_FromString(data[index]);
    for (i=0;i<number;i++)
      free(data[i]);
    free(data);
    return v;
  }
  for (i=0;i<number;i++)
    free(data[i]);
  free(data);
  return NULL;
}

/* SDDS_GetArrayNames
   Modified from standard version
   Returns: NULL on error
            names on success */
static PyObject* sddsdata_GetArrayNames( PyObject* self, PyObject* args )
{
  long fileIndex;
  long index;
  char **data;
  int32_t number;
  PyObject *v;
  long i;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  data = SDDS_GetArrayNames(&dataset_f[fileIndex], &number);
  if (!(data))
    return NULL;
  if (!(v = PyList_New(number)))
    return NULL;
  for (index = 0; index < number; index++)
    PyList_SetItem(v, index, PyString_FromString(data[index]));
  for (i=0;i<number;i++)
    free(data[i]);
  free(data);
  return v;
}

/* SDDS_GetParameterNames
   Modified from standard version
   Returns: NULL on error
            name on success */
static PyObject* sddsdata_GetParameterNameFromIndex( PyObject* self, PyObject* args )
{
  long fileIndex;
  long index;
  char **data;
  int32_t number;
  PyObject *v;
  long i;
  if (!PyArg_ParseTuple(args, "ll", &fileIndex, &index)) {
    return NULL;
  }
  data = SDDS_GetParameterNames(&dataset_f[fileIndex], &number);
  if (!(data))
    return NULL;
  if ((index >= 0) && (index < number)) {
    v = PyString_FromString(data[index]);
    for (i=0;i<number;i++)
      free(data[i]);
    free(data);
    return v;
  }
  for (i=0;i<number;i++)
    free(data[i]);
  free(data);
  return NULL;
}

/* SDDS_GetParameterNames
   Modified from standard version
   Returns: NULL on error
            names on success */
static PyObject* sddsdata_GetParameterNames( PyObject* self, PyObject* args )
{
  long fileIndex;
  long index;
  char **data;
  int32_t number;
  PyObject *v;
  long i;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  data = SDDS_GetParameterNames(&dataset_f[fileIndex], &number);
  if (!(data))
    return NULL;
  if (!(v = PyList_New(number)))
    return NULL;
  for (index = 0; index < number; index++)
    PyList_SetItem(v, index, PyString_FromString(data[index]));
  for (i=0;i<number;i++)
    free(data[i]);
  free(data);
  return v;
}

/* SDDS_SetParameter
   Modified from standard version
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_SetParameter( PyObject* self, PyObject* args )
{
  long fileIndex;
  PyObject *indexOrName;
  PyObject *v;
  long index;
  long type;
  if (!PyArg_ParseTuple(args, "lOO", &fileIndex, &indexOrName, &v)) {
    return NULL;
  }
  if (PyString_Check(indexOrName)) {
    index = SDDS_GetParameterIndex(&dataset_f[fileIndex], PyString_AsString(indexOrName));
  } else if (PyNumber_Check(indexOrName)) {
    if (PyInt_Check(indexOrName))
      index = PyInt_AsLong(indexOrName);
    else if (PyLong_Check(indexOrName))
      index = PyLong_AsLong(indexOrName);
    else
      return NULL;
  } else
    return NULL;
  if ((type = SDDS_GetParameterType(&dataset_f[fileIndex], index)) == 0)
    return NULL;
  switch (type) {
  case SDDS_SHORT:
  case SDDS_USHORT:
  case SDDS_LONG:
    if (PyLong_Check(v)) {
      return PyLong_FromLong(SDDS_SetParameters(&dataset_f[fileIndex],SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE,index,PyLong_AsLong(v),-1));
    } else {
      return PyLong_FromLong(SDDS_SetParameters(&dataset_f[fileIndex],SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE,index,PyInt_AsLong(v),-1));
    }
  case SDDS_ULONG:
    if (PyLong_Check(v)) {
      return PyLong_FromLong(SDDS_SetParameters(&dataset_f[fileIndex],SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE,index,PyLong_AsUnsignedLong(v),-1));
    } else {
      return PyLong_FromLong(SDDS_SetParameters(&dataset_f[fileIndex],SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE,index,PyInt_AsUnsignedLong(v),-1));
    }
  case SDDS_LONG64:
    if (PyLong_Check(v)) {
      return PyLong_FromLong(SDDS_SetParameters(&dataset_f[fileIndex],SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE,index,PyLong_AsLongLong(v),-1));
    } else {
      return PyLong_FromLong(SDDS_SetParameters(&dataset_f[fileIndex],SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE,index,PyInt_AsLongLong(v),-1));
    }
  case SDDS_ULONG64:
    if (PyLong_Check(v)) {
      return PyLong_FromLong(SDDS_SetParameters(&dataset_f[fileIndex],SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE,index,PyLong_AsUnsignedLongLong(v),-1));
    } else {
      return PyLong_FromLong(SDDS_SetParameters(&dataset_f[fileIndex],SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE,index,PyInt_AsUnsignedLongLong(v),-1));
    }
  case SDDS_FLOAT:
  case SDDS_DOUBLE:
    return PyLong_FromLong(SDDS_SetParameters(&dataset_f[fileIndex],SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE,index,PyFloat_AsDouble(v),-1));
  case SDDS_CHARACTER:
    return PyLong_FromLong(SDDS_SetParameters(&dataset_f[fileIndex],SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE,index,(char)(PyString_AsString(v)[0]),-1));
  case SDDS_STRING:
    return PyLong_FromLong(SDDS_SetParameters(&dataset_f[fileIndex],SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE,index,PyString_AsString(v),-1));
  }
  return PyLong_FromLong(0);
}

/* SDDS_SetArray
   Modified from standard version
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_SetArray( PyObject* self, PyObject* args )
{
  long fileIndex;
  long elements;
  PyObject *indexOrName;
  PyObject *v;
  PyObject *dim;
  long type;
  long i;
  void *data = NULL;
  long index;
  long result=0;
  PyObject *temp;
  ARRAY_DEFINITION *arraydef;
  char **names=NULL;
  int32_t number;
  int32_t dimensions;
  int32_t *dimension=NULL;

  if (!PyArg_ParseTuple(args, "lOOO", &fileIndex, &indexOrName, &v, &dim)) {
    return NULL;
  }
  if (PyString_Check(indexOrName)) {
    index = SDDS_GetArrayIndex(&dataset_f[fileIndex], PyString_AsString(indexOrName));
    if ((arraydef = SDDS_GetArrayDefinition(&dataset_f[fileIndex], PyString_AsString(indexOrName))) == NULL)
      return NULL;
  } else if (PyNumber_Check(indexOrName)) {
    if (PyInt_Check(indexOrName))
      index = PyInt_AsLong(indexOrName);
    else if (PyLong_Check(indexOrName))
      index = PyLong_AsLong(indexOrName);
    else
      return NULL;
    names = SDDS_GetArrayNames(&dataset_f[fileIndex], &number);
    if (!(names))
      return NULL;
    if (number <= index)
      return NULL;
    if ((arraydef = SDDS_GetArrayDefinition(&dataset_f[fileIndex], names[index])) == NULL)
      return NULL;
  } else
    return NULL;
  type = arraydef->type;
  dimensions = arraydef->dimensions;
  
  if (dimensions != PyList_Size(dim))
    return NULL;
  dimension = malloc(sizeof(int32_t)*dimensions);

  for (i=0;i<dimensions;i++) {
    temp = PyList_GetItem(dim, i);
    if (PyInt_Check(temp))
      dimension[i] = PyInt_AsLong(temp);
    else if (PyLong_Check(temp))
      dimension[i] = PyLong_AsLong(temp);
    else
      return NULL;
  }

  elements = (long)PyList_Size(v);
  switch (type) {
  case SDDS_SHORT:
    data = malloc(sizeof(short)*elements);
    for (i=0;i<elements;i++) {
      temp = PyList_GetItem(v, i);
      if (PyLong_Check(temp))
        ((short*)data)[i] = (short)PyLong_AsLong(temp);
      else
        ((short*)data)[i] = (short)PyInt_AsLong(temp);
      }
    break;
  case SDDS_USHORT:
    data = malloc(sizeof(unsigned short)*elements);
    for (i=0;i<elements;i++) {
      temp = PyList_GetItem(v, i);
      if (PyLong_Check(temp))
      	((unsigned short*)data)[i] = (unsigned short)PyLong_AsLong(temp);
      else
        ((unsigned short*)data)[i] = (unsigned short)PyInt_AsLong(temp);
      }
    break;
  case SDDS_LONG:
    data = malloc(sizeof(int32_t)*elements);
    for (i=0;i<elements;i++) {
      temp = PyList_GetItem(v, i);
      if (PyLong_Check(temp))
        ((int32_t*)data)[i] = (int32_t)PyLong_AsLong(temp);
      else
        ((int32_t*)data)[i] = (int32_t)PyInt_AsLong(temp);
    }
    break;
  case SDDS_ULONG:
    data = malloc(sizeof(uint32_t)*elements);
    for (i=0;i<elements;i++) {
      temp = PyList_GetItem(v, i);
      if (PyLong_Check(temp))
        ((uint32_t*)data)[i] = (uint32_t)PyLong_AsUnsignedLong(temp);
      else
        ((uint32_t*)data)[i] = (uint32_t)PyInt_AsUnsignedLong(temp);
    }
    break;
  case SDDS_LONG64:
    data = malloc(sizeof(int64_t)*elements);
    for (i=0;i<elements;i++) {
      temp = PyList_GetItem(v, i);
      if (PyLong_Check(temp))
        ((int64_t*)data)[i] = (int64_t)PyLong_AsLongLong(temp);
      else
        ((int64_t*)data)[i] = (int64_t)PyInt_AsLongLong(temp);
    }
    break;
  case SDDS_ULONG64:
    data = malloc(sizeof(uint64_t)*elements);
    for (i=0;i<elements;i++) {
      temp = PyList_GetItem(v, i);
      if (PyLong_Check(temp))
        ((uint64_t*)data)[i] = (uint64_t)PyLong_AsUnsignedLongLong(temp);
      else
        ((uint64_t*)data)[i] = (uint64_t)PyInt_AsUnsignedLongLong(temp);
    }
    break;
  case SDDS_FLOAT:
    data = malloc(sizeof(float)*elements);
    for (i=0;i<elements;i++)
      ((float*)data)[i] = (float)PyFloat_AsDouble(PyList_GetItem(v, i));
    break;
  case SDDS_DOUBLE:
    data = malloc(sizeof(double)*elements);
    for (i=0;i<elements;i++)
      ((double*)data)[i] = (double)PyFloat_AsDouble(PyList_GetItem(v, i));
    break;
  case SDDS_CHARACTER:
    data = malloc(sizeof(char)*elements);
    for (i=0;i<elements;i++)
      ((char*)data)[i] = (char)(PyString_AsString(PyList_GetItem(v, i))[0]);
    break;
  case SDDS_STRING:
    data = malloc(sizeof(char*)*elements);
    for (i=0;i<elements;i++)
      SDDS_CopyString(((char**)data)+i, PyString_AsString(PyList_GetItem(v, i)));
    break;
  }
  if (PyString_Check(indexOrName)) {
    result = SDDS_SetArray(&dataset_f[fileIndex],PyString_AsString(indexOrName),SDDS_CONTIGUOUS_DATA,data,dimension);
  } else {
    result = SDDS_SetArray(&dataset_f[fileIndex],names[index],SDDS_CONTIGUOUS_DATA,data,dimension);
    for (i=0;i<number;i++)
      free(names[i]);
    free(names);
  }
  switch (type) {
  case SDDS_SHORT:
    free((short*)data);
    break;
  case SDDS_USHORT:
    free((unsigned short*)data);
    break;
  case SDDS_LONG:
    free((int32_t*)data);
    break;
  case SDDS_ULONG:
    free((uint32_t*)data);
    break;
  case SDDS_LONG64:
    free((int64_t*)data);
    break;
  case SDDS_ULONG64:
    free((uint64_t*)data);
    break;
  case SDDS_FLOAT:
    free((float*)data);
    break;
  case SDDS_DOUBLE:
    free((double*)data);
    break;
  case SDDS_CHARACTER:
    free((char*)data);
    break;
  case SDDS_STRING:
    for (i=0;i<elements;i++)
      free(((char**)data)[i]);
    free((char**)data);
    break;
  }
  if (dimension) free(dimension);
  return PyLong_FromLong(result);
}

/* SDDS_SetColumn
   Modified from standard version
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_SetColumn( PyObject* self, PyObject* args )
{
  long fileIndex;
  long rows;
  PyObject *indexOrName;
  PyObject *v;
  long type;
  long i;
  void *data = NULL;
  long index;
  long result=0;
  PyObject *temp;
  if (!PyArg_ParseTuple(args, "lOO", &fileIndex, &indexOrName, &v)) {
    return NULL;
  }
  if (PyString_Check(indexOrName)) {
    index = SDDS_GetColumnIndex(&dataset_f[fileIndex], PyString_AsString(indexOrName));
  } else if (PyNumber_Check(indexOrName)) {
    if (PyInt_Check(indexOrName))
      index = PyInt_AsLong(indexOrName);
    else if (PyLong_Check(indexOrName))
      index = PyLong_AsLong(indexOrName);
    else
      return NULL;
  } else
    return NULL;
  if ((type = SDDS_GetColumnType(&dataset_f[fileIndex], index)) == 0)
    return NULL;
  rows = (long)PyList_Size(v);
  switch (type) {
  case SDDS_SHORT:
    data = malloc(sizeof(short)*rows);
    for (i=0;i<rows;i++) {
      temp = PyList_GetItem(v, i);
      if (PyLong_Check(temp))
        ((short*)data)[i] = (short)PyLong_AsLong(temp);
      else
        ((short*)data)[i] = (short)PyInt_AsLong(temp);
      }
    break;
  case SDDS_USHORT:
    data = malloc(sizeof(unsigned short)*rows);
    for (i=0;i<rows;i++) {
      temp = PyList_GetItem(v, i);
      if (PyLong_Check(temp))
        ((unsigned short*)data)[i] = (unsigned short)PyLong_AsLong(temp);
      else
        ((unsigned short*)data)[i] = (unsigned short)PyInt_AsLong(temp);
      }
    break;
  case SDDS_LONG:
    data = malloc(sizeof(int32_t)*rows);
    for (i=0;i<rows;i++) {
      temp = PyList_GetItem(v, i);
      if (PyLong_Check(temp))
        ((int32_t*)data)[i] = (int32_t)PyLong_AsLong(temp);
      else
        ((int32_t*)data)[i] = (int32_t)PyInt_AsLong(temp);
    }
    break;
  case SDDS_ULONG:
    data = malloc(sizeof(uint32_t)*rows);
    for (i=0;i<rows;i++) {
      temp = PyList_GetItem(v, i);
      if (PyLong_Check(temp))
        ((uint32_t*)data)[i] = (uint32_t)PyLong_AsUnsignedLong(temp);
      else
        ((uint32_t*)data)[i] = (uint32_t)PyInt_AsUnsignedLong(temp);
    }
    break;
  case SDDS_LONG64:
    data = malloc(sizeof(int64_t)*rows);
    for (i=0;i<rows;i++) {
      temp = PyList_GetItem(v, i);
      if (PyLong_Check(temp))
        ((int64_t*)data)[i] = (int64_t)PyLong_AsLongLong(temp);
      else
        ((int64_t*)data)[i] = (int64_t)PyInt_AsLongLong(temp);
    }
    break;
  case SDDS_ULONG64:
    data = malloc(sizeof(uint64_t)*rows);
    for (i=0;i<rows;i++) {
      temp = PyList_GetItem(v, i);
      if (PyLong_Check(temp))
        ((uint64_t*)data)[i] = (uint64_t)PyLong_AsUnsignedLongLong(temp);
      else
        ((uint64_t*)data)[i] = (uint64_t)PyInt_AsUnsignedLongLong(temp);
    }
    break;
  case SDDS_FLOAT:
    data = malloc(sizeof(float)*rows);
    for (i=0;i<rows;i++)
      ((float*)data)[i] = (float)PyFloat_AsDouble(PyList_GetItem(v, i));
    break;
  case SDDS_DOUBLE:
    data = malloc(sizeof(double)*rows);
    for (i=0;i<rows;i++)
      ((double*)data)[i] = (double)PyFloat_AsDouble(PyList_GetItem(v, i));
    break;
  case SDDS_CHARACTER:
    data = malloc(sizeof(char)*rows);
    for (i=0;i<rows;i++)
      ((char*)data)[i] = (char)(PyString_AsString(PyList_GetItem(v, i))[0]);
    break;
  case SDDS_STRING:
    data = malloc(sizeof(char*)*rows);
    for (i=0;i<rows;i++)
      SDDS_CopyString(((char**)data)+i, PyString_AsString(PyList_GetItem(v, i)));
    break;
  }
  result = SDDS_SetColumn(&dataset_f[fileIndex],SDDS_SET_BY_INDEX,data,rows,index,NULL);
  switch (type) {
  case SDDS_SHORT:
    free((short*)data);
    break;
  case SDDS_USHORT:
    free((unsigned short*)data);
    break;
  case SDDS_LONG:
    free((int32_t*)data);
    break;
  case SDDS_ULONG:
    free((uint32_t*)data);
    break;
  case SDDS_LONG64:
    free((int64_t*)data);
    break;
  case SDDS_ULONG64:
    free((uint64_t*)data);
    break;
  case SDDS_FLOAT:
    free((float*)data);
    break;
  case SDDS_DOUBLE:
    free((double*)data);
    break;
  case SDDS_CHARACTER:
    free((char*)data);
    break;
  case SDDS_STRING:
    for (i=0;i<rows;i++)
      free(((char**)data)[i]);
    free((char**)data);
    break;
  }
  return PyLong_FromLong(result);
}

/* SDDS_SetRowValues
   Modified from standard version
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_SetRowValues( PyObject* self, PyObject* args )
{
  long fileIndex;
  long row, elements;
  PyObject *v;
  long type;
  long i;
  long index;
  PyObject *temp;
  if (!PyArg_ParseTuple(args, "llO", &fileIndex, &row, &v)) {
    return NULL;
  }
  elements = (long)PyList_Size(v);
  for (i=0; i<elements; i=i+2) {
    index = SDDS_GetColumnIndex(&dataset_f[fileIndex], PyString_AsString(PyList_GetItem(v, i)));
    temp = PyList_GetItem(v, i+1);
    if ((type = SDDS_GetColumnType(&dataset_f[fileIndex], index)) == 0)
      return NULL;
    switch (type) {
    case SDDS_SHORT:
    case SDDS_USHORT:
    case SDDS_LONG:
      if (PyLong_Check(temp)) {
        if (SDDS_SetRowValues(&dataset_f[fileIndex],SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE,row,index,PyLong_AsLong(temp),-1) == 0)
          return NULL;
      } else {
        if (SDDS_SetRowValues(&dataset_f[fileIndex],SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE,row,index,PyInt_AsLong(temp),-1) == 0)
          return NULL;
      }
      break;
    case SDDS_ULONG:
      if (PyLong_Check(temp)) {
        if (SDDS_SetRowValues(&dataset_f[fileIndex],SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE,row,index,PyLong_AsUnsignedLong(temp),-1) == 0)
          return NULL;
      } else {
        if (SDDS_SetRowValues(&dataset_f[fileIndex],SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE,row,index,PyInt_AsUnsignedLong(temp),-1) == 0)
          return NULL;
      }
      break;
    case SDDS_LONG64:
      if (PyLong_Check(temp)) {
        if (SDDS_SetRowValues(&dataset_f[fileIndex],SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE,row,index,PyLong_AsLongLong(temp),-1) == 0)
          return NULL;
      } else {
        if (SDDS_SetRowValues(&dataset_f[fileIndex],SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE,row,index,PyInt_AsLongLong(temp),-1) == 0)
          return NULL;
      }
      break;
    case SDDS_ULONG64:
      if (PyLong_Check(temp)) {
        if (SDDS_SetRowValues(&dataset_f[fileIndex],SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE,row,index,PyLong_AsUnsignedLongLong(temp),-1) == 0)
          return NULL;
      } else {
        if (SDDS_SetRowValues(&dataset_f[fileIndex],SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE,row,index,PyInt_AsUnsignedLongLong(temp),-1) == 0)
          return NULL;
      }
      break;
    case SDDS_FLOAT:
    case SDDS_DOUBLE:
      if (SDDS_SetRowValues(&dataset_f[fileIndex],SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE,row,index,PyFloat_AsDouble(temp),-1) == 0)
          return NULL;
      break;
    case SDDS_CHARACTER:
      if (SDDS_SetRowValues(&dataset_f[fileIndex],SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE,row,index,(char)(PyString_AsString(temp)[0]),-1) == 0)
          return NULL;
      break;
    case SDDS_STRING:
      if (SDDS_SetRowValues(&dataset_f[fileIndex],SDDS_SET_BY_INDEX|SDDS_PASS_BY_VALUE,row,index,PyString_AsString(temp),-1) == 0)
          return NULL;
      break;
    }
  }
  return PyLong_FromLong(1);
}

/* SDDS_GetColumn 
   Modified from standard version
   Returns: NULL on error
            data on success */
static PyObject* sddsdata_GetColumn( PyObject* self, PyObject* args )
{
  long fileIndex;
  PyObject *indexOrName;
  long originalType;
  int64_t rows, i;
  void *columnValue;
  char *name;
  char **data;
  int32_t number;
  char buffer[40];
  long index;
  PyObject *v;
  if (!PyArg_ParseTuple(args, "lO", &fileIndex, &indexOrName)) {
    return NULL;
  }
  if (PyString_Check(indexOrName)) {
    index = SDDS_GetColumnIndex(&dataset_f[fileIndex], PyString_AsString(indexOrName));
  } else if (PyNumber_Check(indexOrName)) {
    if (PyInt_Check(indexOrName))
      index = PyInt_AsLong(indexOrName);
    else if (PyLong_Check(indexOrName))
      index = PyLong_AsLong(indexOrName);
    else
      return NULL;
  } else
    return NULL;
  
  data = SDDS_GetColumnNames(&dataset_f[fileIndex], &number);
  if (!(data))
    return NULL;
  if ((index >= 0) && (index < number)) {
    name = data[index];
  } else {
    return NULL;
  }

  rows = SDDS_RowCount(&(dataset_f[fileIndex]));
  if (rows < 0) {
    Py_INCREF(Py_None);
    return Py_None;
  }

  originalType = SDDS_GetColumnType(&(dataset_f[fileIndex]), index);
    
  columnValue = SDDS_GetColumn(&(dataset_f[fileIndex]), name);

  for (i=0;i<number;i++)
    free(data[i]);
  free(data);

  if (!(columnValue))
    return NULL;

  if (!(v = PyList_New(rows)))
    return NULL;
  
  switch (originalType) {
  case SDDS_SHORT:
    for (i=0;i < rows;i++) {
      PyList_SetItem(v, i, PyLong_FromLong((long)(((short*)columnValue)[i])));
    }
    free((short*)columnValue);
    break;
  case SDDS_USHORT:
    for (i=0;i < rows;i++) {
      PyList_SetItem(v, i, PyLong_FromLong((long)(((unsigned short*)columnValue)[i])));
    }
    free((unsigned short*)columnValue);
    break;
  case SDDS_LONG:
    for (i=0;i < rows;i++) {
      PyList_SetItem(v, i, PyLong_FromLong((long)(((int32_t*)columnValue)[i])));
    }
    free((int32_t*)columnValue);
    break;
  case SDDS_ULONG:
    for (i=0;i < rows;i++) {
      PyList_SetItem(v, i, PyLong_FromUnsignedLong((unsigned long)(((uint32_t*)columnValue)[i])));
    }
    free((uint32_t*)columnValue);
    break;
  case SDDS_LONG64:
    for (i=0;i < rows;i++) {
      PyList_SetItem(v, i, PyLong_FromLongLong((long long)(((int64_t*)columnValue)[i])));
    }
    free((int64_t*)columnValue);
    break;
  case SDDS_ULONG64:
    for (i=0;i < rows;i++) {
      PyList_SetItem(v, i, PyLong_FromUnsignedLongLong((unsigned long long)(((uint64_t*)columnValue)[i])));
    }
    free((uint64_t*)columnValue);
    break;
  case SDDS_FLOAT:
    for (i=0;i < rows;i++) {
      sprintf(buffer, "%.6E", ((float*)columnValue)[i]);
      PyList_SetItem(v, i, PyFloat_FromDouble(atof(buffer)));
    }
    free((float*)columnValue);
    break;
  case SDDS_DOUBLE:
    for (i=0;i < rows;i++) {
      PyList_SetItem(v, i, PyFloat_FromDouble(((double*)columnValue)[i]));
    }
    free((double*)columnValue);
    break;
  case SDDS_CHARACTER:
    for (i=0;i < rows;i++) {
      sprintf(buffer, "%c", ((char*)columnValue)[i]);
      PyList_SetItem(v, i, PyString_FromString(buffer));
    }
    free((char*)columnValue);
    break;
  case SDDS_STRING:
    for (i=0;i < rows;i++) {
      PyList_SetItem(v, i, PyString_FromString(((char**)columnValue)[i]));
    }
    for (i=0;i < rows;i++)
      free(((char**)columnValue)[i]);
    free((char**)columnValue);
    break;
  default:
    return NULL;
  }
  return v;
}

/* SDDS_GetArray 
   Modified from standard version
   Returns: NULL on error
            data on success */
static PyObject* sddsdata_GetArray( PyObject* self, PyObject* args )
{
  long fileIndex;
  PyObject *indexOrName;
  long originalType, i, elements;
  SDDS_ARRAY *arrayValue=NULL;
  char *name;
  char **data;
  int32_t number;
  char buffer[40];
  long index;
  PyObject *v;
  if (!PyArg_ParseTuple(args, "lO", &fileIndex, &indexOrName)) {
    return NULL;
  }
  if (PyString_Check(indexOrName)) {
    index = SDDS_GetArrayIndex(&dataset_f[fileIndex], PyString_AsString(indexOrName));
  } else if (PyNumber_Check(indexOrName)) {
    if (PyInt_Check(indexOrName))
      index = PyInt_AsLong(indexOrName);
    else if (PyLong_Check(indexOrName))
      index = PyLong_AsLong(indexOrName);
    else
      return NULL;
  } else
    return NULL;
  
  data = SDDS_GetArrayNames(&dataset_f[fileIndex], &number);
  if (!(data))
    return NULL;
  if ((index >= 0) && (index < number)) {
    name = data[index];
  } else {
    return NULL;
  }

  arrayValue = SDDS_GetArray(&(dataset_f[fileIndex]), name, NULL);

  for (i=0;i<number;i++)
    free(data[i]);
  free(data);

  originalType = arrayValue->definition->type;

  if (!(arrayValue))
    return NULL;

  elements = arrayValue->elements;

  /*
  if (!(v = PyList_New(arrayValue->definition->dimensions)))
    return NULL;
  for (i=0;i<arrayValue>dimensions;i++) {
    if (!(v[i] = PyList_New(arrayValue->dimension[i])))
      return NULL;
  }
  */
  if (!(v = PyList_New(arrayValue->elements)))
    return NULL;

  switch (originalType) {
  case SDDS_SHORT:
    for (i=0;i < elements;i++) {
      PyList_SetItem(v, i, PyLong_FromLong((long)(((short*)arrayValue->data)[i])));
    }
    break;
  case SDDS_USHORT:
    for (i=0;i < elements;i++) {
      PyList_SetItem(v, i, PyLong_FromLong((long)(((unsigned short*)arrayValue->data)[i])));
    }
    break;
  case SDDS_LONG:
    for (i=0;i < elements;i++) {
      PyList_SetItem(v, i, PyLong_FromLong((long)(((int32_t*)arrayValue->data)[i])));
    }
    break;
  case SDDS_ULONG:
    for (i=0;i < elements;i++) {
      PyList_SetItem(v, i, PyLong_FromUnsignedLong((unsigned long)(((uint32_t*)arrayValue->data)[i])));
    }
    break;
  case SDDS_LONG64:
    for (i=0;i < elements;i++) {
      PyList_SetItem(v, i, PyLong_FromLongLong((long long)(((int64_t*)arrayValue->data)[i])));
    }
    break;
  case SDDS_ULONG64:
    for (i=0;i < elements;i++) {
      PyList_SetItem(v, i, PyLong_FromUnsignedLongLong((unsigned long long)(((uint64_t*)arrayValue->data)[i])));
    }
    break;
  case SDDS_FLOAT:
    for (i=0;i < elements;i++) {
      sprintf(buffer, "%.6E", ((float*)arrayValue->data)[i]);
      PyList_SetItem(v, i, PyFloat_FromDouble(atof(buffer)));
    }
    break;
  case SDDS_DOUBLE:
    for (i=0;i < elements;i++) {
      PyList_SetItem(v, i, PyFloat_FromDouble(((double*)arrayValue->data)[i]));
    }
    break;
  case SDDS_CHARACTER:
    for (i=0;i < elements;i++) {
      sprintf(buffer, "%c", ((char*)arrayValue->data)[i]);
      PyList_SetItem(v, i, PyString_FromString(buffer));
    }
    break;
  case SDDS_STRING:
    for (i=0;i < elements;i++) {
      PyList_SetItem(v, i, PyString_FromString(((char**)arrayValue->data)[i]));
    }
    break;
  default:
    return NULL;
  }
  SDDS_FreeArray(arrayValue);
  return v;
}

/* SDDS_GetArrayDimensions 
   Modified from standard version
   Returns: NULL on error
            dimensions */
static PyObject* sddsdata_GetArrayDimensions( PyObject* self, PyObject* args )
{
  long fileIndex;
  PyObject *indexOrName;
  long i;
  SDDS_ARRAY *arrayValue=NULL;
  char *name;
  char **data;
  int32_t number;
  long index;
  PyObject *v;
  if (!PyArg_ParseTuple(args, "lO", &fileIndex, &indexOrName)) {
    return NULL;
  }
  if (PyString_Check(indexOrName)) {
    index = SDDS_GetArrayIndex(&dataset_f[fileIndex], PyString_AsString(indexOrName));
  } else if (PyNumber_Check(indexOrName)) {
    if (PyInt_Check(indexOrName))
      index = PyInt_AsLong(indexOrName);
    else if (PyLong_Check(indexOrName))
      index = PyLong_AsLong(indexOrName);
    else
      return NULL;
  } else
    return NULL;
  
  data = SDDS_GetArrayNames(&dataset_f[fileIndex], &number);
  if (!(data))
    return NULL;
  if ((index >= 0) && (index < number)) {
    name = data[index];
  } else {
    return NULL;
  }

  arrayValue = SDDS_GetArray(&(dataset_f[fileIndex]), name, NULL);

  for (i=0;i<number;i++)
    free(data[i]);
  free(data);

  if (!(arrayValue))
    return NULL;
  
  if (!(v = PyList_New(arrayValue->definition->dimensions)))
    return NULL;
  for (i=0;i<arrayValue->definition->dimensions;i++) {
    PyList_SetItem(v, i, PyLong_FromLong(arrayValue->dimension[i]));
  }
  SDDS_FreeArray(arrayValue);
  return v;
}

/* SDDS_GetParameter
   Modified from standard version
   Returns: NULL on error
            value on success */
static PyObject* sddsdata_GetParameter( PyObject* self, PyObject* args ) 
{
  long fileIndex;
  PyObject *indexOrName;
  long originalType;
  void *parameterValue;
  char *name;
  char **data;
  int32_t number;
  char buffer[40];
  long index;
  long i;
  PyObject *v=NULL;
  if (!PyArg_ParseTuple(args, "lO", &fileIndex, &indexOrName)) {
    return NULL;
  }
  if (PyString_Check(indexOrName)) {
    index = SDDS_GetParameterIndex(&dataset_f[fileIndex], PyString_AsString(indexOrName));
  } else if (PyNumber_Check(indexOrName)) {
    if (PyInt_Check(indexOrName))
      index = PyInt_AsLong(indexOrName);
    else if (PyLong_Check(indexOrName))
      index = PyLong_AsLong(indexOrName);
    else
      return NULL;
  } else
    return NULL;

  data = SDDS_GetParameterNames(&dataset_f[fileIndex], &number);
  if (!(data))
    return NULL;
  if ((index >= 0) && (index < number)) {
    name = data[index];
  } else {
    return NULL;
  }
  if (!name) {
    return NULL;
  }

  parameterValue = SDDS_GetParameter(&(dataset_f[fileIndex]), name, NULL);

  for (i=0;i<number;i++)
    free(data[i]);
  free(data);

  if (!(parameterValue)) {
    return NULL;
  }
  originalType = SDDS_GetParameterType(&dataset_f[fileIndex], index);

  switch (originalType) {
  case SDDS_SHORT:
    v = PyLong_FromLong((long)(((short*)parameterValue)[0]));
    free((short*)parameterValue);
    break;
  case SDDS_USHORT:
    v = PyLong_FromLong((long)(((unsigned short*)parameterValue)[0]));
    free((unsigned short*)parameterValue);
    break;
  case SDDS_LONG:
    v = PyLong_FromLong((long)(((int32_t*)parameterValue)[0]));
    free((int32_t*)parameterValue);
    break;
  case SDDS_ULONG:
    v = PyLong_FromUnsignedLong((unsigned long)(((uint32_t*)parameterValue)[0]));
    free((uint32_t*)parameterValue);
    break;
  case SDDS_LONG64:
    v = PyLong_FromLongLong((long long)(((int64_t*)parameterValue)[0]));
    free((int64_t*)parameterValue);
    break;
  case SDDS_ULONG64:
    v = PyLong_FromUnsignedLongLong((unsigned long long)(((uint64_t*)parameterValue)[0]));
    free((uint64_t*)parameterValue);
    break;
  case SDDS_FLOAT:
    sprintf(buffer, "%.6E", ((float*)parameterValue)[0]);
    v = PyFloat_FromDouble(atof(buffer));
    free((float*)parameterValue);
    break;
  case SDDS_DOUBLE:
    v = PyFloat_FromDouble(((double*)parameterValue)[0]);
    free((double*)parameterValue);
    break;
  case SDDS_CHARACTER:
    sprintf(buffer, "%c", ((char*)parameterValue)[0]);
    v = PyString_FromString(buffer);
    free((char*)parameterValue);
    break;
  case SDDS_STRING:
    v = PyString_FromString(((char**)parameterValue)[0]);
    free(((char**)parameterValue)[0]);
    free((char**)parameterValue);
    break;
  }
  return v;
}

/* SDDS_InitializeInput
   Returns: 0 on error
            1 on success */
static PyObject* sddsdata_GetMode( PyObject* self, PyObject* args ) 
{
  long fileIndex;
  if (!PyArg_ParseTuple(args, "l", &fileIndex)) {
    return NULL;
  }
  return PyLong_FromLong(dataset_f[fileIndex].layout.data_mode.mode);
}

struct module_state {
    PyObject *error;
};

#if PY_MAJOR_VERSION >= 3
#define GETSTATE(m) ((struct module_state*)PyModule_GetState(m))
#else
#define GETSTATE(m) (&_state)
static struct module_state _state;
#endif

static PyMethodDef sddsdata_methods[] = {
  { "InitializeInput", sddsdata_InitializeInput, METH_VARARGS },
  { "InitializeAppend", sddsdata_InitializeAppend, METH_VARARGS },
  { "InitializeAppendToPage", sddsdata_InitializeAppendToPage, METH_VARARGS },
  { "InitializeOutput", sddsdata_InitializeOutput, METH_VARARGS },
  { "SetColumnMajorOrder", sddsdata_SetColumnMajorOrder, METH_VARARGS },
  { "SetRowMajorOrder", sddsdata_SetRowMajorOrder, METH_VARARGS },
  { "SetFixedRowCountMode", sddsdata_SetFixedRowCountMode, METH_VARARGS },
  { "EnableFSync", sddsdata_EnableFSync, METH_VARARGS },
  { "DisableFSync", sddsdata_DisableFSync, METH_VARARGS },
  { "Terminate", sddsdata_Terminate, METH_VARARGS },
  { "SetTerminateMode", sddsdata_SetTerminateMode, METH_VARARGS },
  { "DefineParameter", sddsdata_DefineParameter, METH_VARARGS },
  { "DefineArray", sddsdata_DefineArray, METH_VARARGS },
  { "DefineColumn", sddsdata_DefineColumn, METH_VARARGS },
  { "IsValidName", sddsdata_IsValidName, METH_VARARGS },
  { "SetNameValidityFlags", sddsdata_SetNameValidityFlags, METH_VARARGS },
  { "DefineSimpleColumn", sddsdata_DefineSimpleColumn, METH_VARARGS },
  { "DefineSimpleArray", sddsdata_DefineSimpleArray, METH_VARARGS },
  { "DefineSimpleParameter", sddsdata_DefineSimpleParameter, METH_VARARGS },
  { "WriteLayout", sddsdata_WriteLayout, METH_VARARGS },
  { "EraseData", sddsdata_EraseData, METH_VARARGS },
  { "ProcessColumnString", sddsdata_ProcessColumnString, METH_VARARGS },
  { "ProcessArrayString", sddsdata_ProcessArrayString, METH_VARARGS },
  { "ProcessParameterString", sddsdata_ProcessParameterString, METH_VARARGS },
  { "InitializeCopy", sddsdata_InitializeCopy, METH_VARARGS },
  { "CopyLayout", sddsdata_CopyLayout, METH_VARARGS },
  { "AppendLayout", sddsdata_AppendLayout, METH_VARARGS },
  { "CopyPage", sddsdata_CopyPage, METH_VARARGS },
  { "CopyParameters", sddsdata_CopyParameters, METH_VARARGS },
  { "CopyArrays", sddsdata_CopyArrays, METH_VARARGS },
  { "CopyColumns", sddsdata_CopyColumns, METH_VARARGS },
  { "CopyRow", sddsdata_CopyRow, METH_VARARGS },
  { "CopyRowDirect", sddsdata_CopyRowDirect, METH_VARARGS },
  { "CopyAdditionalRows", sddsdata_CopyAdditionalRows, METH_VARARGS },
  { "DeferSavingLayout", sddsdata_DeferSavingLayout, METH_VARARGS },
  { "SaveLayout", sddsdata_SaveLayout, METH_VARARGS },
  { "RestoreLayout", sddsdata_RestoreLayout, METH_VARARGS },
  { "StartPage", sddsdata_StartPage, METH_VARARGS },
  { "ClearPage", sddsdata_ClearPage, METH_VARARGS },
  { "LengthenTable", sddsdata_LengthenTable, METH_VARARGS },
  { "WritePage", sddsdata_WritePage, METH_VARARGS },
  { "UpdatePage", sddsdata_UpdatePage, METH_VARARGS },
  { "InitHeaderlessInput", sddsdata_InitHeaderlessInput, METH_VARARGS },
  { "ReadPage", sddsdata_ReadPage, METH_VARARGS },
  { "ReadPageSparse", sddsdata_ReadPageSparse, METH_VARARGS },
  { "ReadPageLastRows", sddsdata_ReadPageLastRows, METH_VARARGS },
  { "RowCount", sddsdata_RowCount, METH_VARARGS },
  { "SetColumnFlags", sddsdata_SetColumnFlags, METH_VARARGS },
  { "SetRowFlags", sddsdata_SetRowFlags, METH_VARARGS },
  { "GetRowFlag", sddsdata_GetRowFlag, METH_VARARGS },
  { "DeleteColumn", sddsdata_DeleteColumn, METH_VARARGS },
  { "DeleteParameter", sddsdata_DeleteParameter, METH_VARARGS },
  { "DeleteUnsetColumns", sddsdata_DeleteUnsetColumns, METH_VARARGS },
  { "DeleteUnsetRows", sddsdata_DeleteUnsetRows, METH_VARARGS },
  { "ColumnCount", sddsdata_ColumnCount, METH_VARARGS },
  { "ArrayCount", sddsdata_ArrayCount, METH_VARARGS },
  { "ParameterCount", sddsdata_ParameterCount, METH_VARARGS },
  { "GetDescription", sddsdata_GetDescription, METH_VARARGS },
  { "GetDescriptionText", sddsdata_GetDescriptionText, METH_VARARGS },
  { "GetDescriptionContents", sddsdata_GetDescriptionContents, METH_VARARGS },
  { "NumberOfErrors", sddsdata_NumberOfErrors, METH_VARARGS },
  { "ClearErrors", sddsdata_ClearErrors, METH_VARARGS },
  { "SetError", sddsdata_SetError, METH_VARARGS },
  { "Bomb", sddsdata_Bomb, METH_VARARGS },
  { "Warning", sddsdata_Warning, METH_VARARGS },
  { "RegisterProgramName", sddsdata_RegisterProgramName, METH_VARARGS },
  { "PrintErrors", sddsdata_PrintErrors, METH_VARARGS },
  { "TransferColumnDefinition", sddsdata_TransferColumnDefinition, METH_VARARGS },
  { "TransferArrayDefinition", sddsdata_TransferArrayDefinition, METH_VARARGS },
  { "TransferParameterDefinition", sddsdata_TransferParameterDefinition, METH_VARARGS },
  { "DefineColumnLikeParameter", sddsdata_DefineColumnLikeParameter, METH_VARARGS },
  { "DefineParameterLikeColumn", sddsdata_DefineParameterLikeColumn, METH_VARARGS },
  { "TransferAllColumnDefinitions", sddsdata_TransferAllColumnDefinitions, METH_VARARGS },
  { "TransferAllArrayDefinitions", sddsdata_TransferAllArrayDefinitions, METH_VARARGS },
  { "TransferAllParameterDefinitions", sddsdata_TransferAllParameterDefinitions, METH_VARARGS },
  { "GetColumnIndex", sddsdata_GetColumnIndex, METH_VARARGS },
  { "GetArrayIndex", sddsdata_GetArrayIndex, METH_VARARGS },
  { "GetParameterIndex", sddsdata_GetParameterIndex, METH_VARARGS },
  { "GetColumnType", sddsdata_GetColumnType, METH_VARARGS },
  { "GetArrayType", sddsdata_GetArrayType, METH_VARARGS },
  { "GetNamedColumnType", sddsdata_GetNamedColumnType, METH_VARARGS },
  { "GetNamedArrayType", sddsdata_GetNamedArrayType, METH_VARARGS },
  { "GetColumnDefinition", sddsdata_GetColumnDefinition, METH_VARARGS },
  { "GetArrayDefinition", sddsdata_GetArrayDefinition, METH_VARARGS },
  { "GetParameterType", sddsdata_GetParameterType, METH_VARARGS },
  { "GetNamedParameterType", sddsdata_GetNamedParameterType, METH_VARARGS },
  { "GetParameterDefinition", sddsdata_GetParameterDefinition, METH_VARARGS },
  { "GetTypeSize", sddsdata_GetTypeSize, METH_VARARGS },
  { "GetTypeName", sddsdata_GetTypeName, METH_VARARGS },
  { "IdentifyType", sddsdata_IdentifyType, METH_VARARGS },
  { "CheckColumn", sddsdata_CheckColumn, METH_VARARGS },
  { "CheckArray", sddsdata_CheckArray, METH_VARARGS },
  { "CheckParameter", sddsdata_CheckParameter, METH_VARARGS },
  { "HasWhitespace", sddsdata_HasWhitespace, METH_VARARGS },
  { "StringIsBlank", sddsdata_StringIsBlank, METH_VARARGS },
  { "ApplyFactorToParameter", sddsdata_ApplyFactorToParameter, METH_VARARGS },
  { "ApplyFactorToColumn", sddsdata_ApplyFactorToColumn, METH_VARARGS },
  { "DeleteParameterFixedValues", sddsdata_DeleteParameterFixedValues, METH_VARARGS },
  { "SetDataMode", sddsdata_SetDataMode, METH_VARARGS },
  { "CheckDataset", sddsdata_CheckDataset, METH_VARARGS },
  { "SetAutoCheckMode", sddsdata_SetAutoCheckMode, METH_VARARGS },
  { "GetColumnNameFromIndex", sddsdata_GetColumnNameFromIndex, METH_VARARGS },
  { "GetColumnNames", sddsdata_GetColumnNames, METH_VARARGS },
  { "GetArrayNameFromIndex", sddsdata_GetArrayNameFromIndex, METH_VARARGS },
  { "GetArrayNames", sddsdata_GetArrayNames, METH_VARARGS },
  { "GetParameterNameFromIndex", sddsdata_GetParameterNameFromIndex, METH_VARARGS },
  { "GetParameterNames", sddsdata_GetParameterNames, METH_VARARGS },
  { "SetParameter", sddsdata_SetParameter, METH_VARARGS },
  { "SetColumn", sddsdata_SetColumn, METH_VARARGS },
  { "SetArray", sddsdata_SetArray, METH_VARARGS },
  { "SetRowValues", sddsdata_SetRowValues, METH_VARARGS },
  { "GetColumn", sddsdata_GetColumn, METH_VARARGS },
  { "GetArray", sddsdata_GetArray, METH_VARARGS },
  { "GetArrayDimensions", sddsdata_GetArrayDimensions, METH_VARARGS },
  { "GetParameter", sddsdata_GetParameter, METH_VARARGS },
  { "GetMode", sddsdata_GetMode, METH_VARARGS },
  { NULL, NULL }
};

#if PY_MAJOR_VERSION >= 3
 #if defined(_WIN32)
  __declspec(dllexport) PyObject* PyInit_sddsdata(void);
 #endif
#else
 #if defined(_WIN32)
  __declspec(dllexport) void initsddsdata(void);
 #endif
#endif

#if PY_MAJOR_VERSION >= 3
static int sddsdata_traverse(PyObject *m, visitproc visit, void *arg) {
    Py_VISIT(GETSTATE(m)->error);
    return 0;
}
static int sddsdata_clear(PyObject *m) {
    Py_CLEAR(GETSTATE(m)->error);
    return 0;
}
static struct PyModuleDef moduledef = {
        PyModuleDef_HEAD_INIT,
        "sddsdata",
        NULL,
        sizeof(struct module_state),
        sddsdata_methods,
        NULL,
        sddsdata_traverse,
        sddsdata_clear,
        NULL
};
#define INITERROR return NULL
#else
#define INITERROR return
#endif

#if PY_MAJOR_VERSION >= 3
//PyObject * PyInit_sddsdata(void)
PyMODINIT_FUNC PyInit_sddsdata(void)
#else
void initsddsdata(void)
#endif
{
  struct module_state *st;
#if PY_MAJOR_VERSION >= 3
    PyObject *module = PyModule_Create(&moduledef);
#else
    PyObject *module = Py_InitModule( "sddsdata", sddsdata_methods );
#endif
    if (module == NULL)
        INITERROR;

    st = GETSTATE(module);

    st->error = PyErr_NewException("sddsedir.Error", NULL, NULL);
    if (st->error == NULL) {
        Py_DECREF(module);
        INITERROR;
    }

#if PY_MAJOR_VERSION >= 3
    return module;
#endif

}
